#ifndef _DPF_CLIENT
#define _DPF_CLIENT

#include "DPF_common.h"

typedef enum _DPF_client_status {
    FRESH,
    // client: TCP client socket status
    INIT_CONNECTION, CONNECT_DEST_AGENT_00, CONNECT_DEST_AGENT_01,
    CONNECT_DEST_AGENT_02, CONNECT_DEST_AGENT_03, CONNECT_DEST_AGENT_1,
    CONNECT_DEST_AGENT_2, CONTACT_DEST_AGENT_1, CONTACT_DEST_AGENT_2,
    ACTIVE_CONNECTING_00, ACTIVE_CONNECTING_01, ACTIVE_CONNECTING_02,
    ACTIVE_CONNECTING_03, ACTIVE_CONNECTING_1, ACTIVE_CONNECTING_2,
    PASSIVE_CONNECTING_1, PASSIVE_CONNECTING_2,
    RELAY_CONNECTING_00, RELAY_CONNECTING_01, RELAY_CONNECTING_02, // UDP only
    RELAY_CONNECTING_03, RELAY_CONNECTING_1, RELAY_CONNECTING_2, // UDP only
    FINISHING, // UDP only
    CONNECTION_FAILED, CONNECTION_SUCC,
    // client: TCP server socket status
    CONNECT_INAGENT_0, CONNECT_INAGENT_1, REGISTERING_1, REGISTERING_2, LOCAL_BOUND,
    BIND_FAIL, HALF_BOUND, FULL_BOUND, LISTENING_1, LISTEN_RECOVERING, LISTENING_2,
    LOCAL_LISTEN, HALF_LISTEN, LISTEN_FAIL, FULL_LISTEN,
    // client: TCP hidden
    CONTACTING
} DPF_client_status;

typedef enum _DPF_app_status {
    CREATED,
    BOUND,
    LISTENING,
    CONNECTED
} DPF_app_status;

/****************************************************************************/
/*                            ClientSock Class                              */
/* This is the base class of DPF_TCP_sock and DPF_UDP_sock classes        */
/****************************************************************************/
class ClientSock {
    public:
        int fd;
        int hfd; // UDP socket to send heartbeat and DEL command
        int refs; // how many owners share this information
        bool forked; // if this socket is shared by multiple processes
        DPF_client_status status;
        time_t entered;
        DPF_app_status app_status; // application's understanding of this socket
        bool nonblock; // Since DPF makes underlying sockets nonblocking, we need a
                       // tag that indicates application's understanding in terms of
                       // blocking
        bool bound; // represents if the application has called 'bind' for the socket
                    // Since DPF needs to bind the socket regardless the app calls
                    // 'bind' or not, we need to keep this status to have the same
                    // understanding as that of the app. For example, if app calls
                    // 'connect' and DPF fails to do that, it needs to recover the
                    // previous status of the socket. Note that "app_status" cannot
                    // replace this field because UDP socket can traverse back and forth
        ExAddr *myaddr; // the address that the socket is bound to. myaddr->local
                        // holds the address that the socket is locally bound to.
                        // On the other hand, myaddr->proxy holds the official
                        // address. If the machine is inside private network, then
                        // myaddr->proxy will holds the address of the proxy socket
                        // created at inagent
        int error;
        DPF_channel *inagent;
        DPF_channel *peer_agent;

        /* Methods */
        ClientSock(int fd);
        ~ClientSock();
        void tostr(const char *prefix, char *str);
        // Those two functions are for serialization and unserialization
        int stripe(char *outbuf);
        int unstripe(char *inbuf);
};

/****************************************************************************/
/*                            DPF_TCP_sock Class                           */
/****************************************************************************/
class DPF_TCP_sock : public ClientSock {
    private:
    public:
        int listener;
        ExAddr *peer; // the address of the peer that this TCP socket is connected to
                      // or is trying to connect to
                      // peer->proxy contains the official address,
                      // i.e., application's understanding on the peer's address.
                      // peer->agent contains the address of peer's inagent.
                      // peer->local contains the address to which packets are needed
                      // to sent actually.
        /* connection queue */
        DPF_TCP_sock *chead; // = NULL;
        DPF_TCP_sock *ctail; // = NULL;

        DPF_TCP_sock(int owner);
        ~DPF_TCP_sock();
        int make_fresh();
        void tostr(const char *prefix, char *str);
        // Those two functions are for serialization and unserialization
        // @args:
        //   - deep: if true, connection queue must be (un)striped
        int stripe(char *outbuf, bool deep);
        int unstripe(char *inbuf, bool deep);
};

/****************************************************************************/
/*                  DPF_UDP_sock and supporting classes                    */
/****************************************************************************/
class UDP_msg {
    private:
    public:
        struct sockaddr_in from; // the sinful address of the sender of this msg
        char *data; // data is stored here
        unsigned short size; // the size of data stored in 'data'
        UDP_msg *next;

        UDP_msg(struct sockaddr_in *src, char *msg, unsigned short msglen);
        ~UDP_msg();
        void tostr(const char *prefix, char *str);
        // Those two functions are for serialization and unserialization
        bool read_str(char *str);
        bool write_str(char *str);
};
class UDP_peer : public ExAddr {
    public:
        time_t lastContact;
        UDP_peer *next;
        UDP_peer(struct sockaddr_in *local,
                struct sockaddr_in *proxy,
                struct sockaddr_in *agent);
        void tostr(const char *prefix, char *str);
        int stripe(char *outbuf);
        int unstripe(char *inbuf);
};
#define MAX_UDP_RBUF 20 // max number of msgs in receiving buffer
class DPF_UDP_sock : public ClientSock {
    private:
        void rm_rbuf();
    public:
        UDP_peer *peer; // this variable represents the (current) peer endpoint. If the
                      // socket is connected, this represents the peer. If it is
                      // not connected, this will contain the address of peer endpoint
                      // that the socket is currently trying to open a channel. For
                      // example, when the app calls 'sendto', DPF needs to open a
                      // channel by exchanging several msgs. During this channel
                      // opening up, this var contains the target address. When the
                      // channel is succefully opened up, this information about the
                      // peer is moved to 'peerCache' and later used to communicate
                      // with the peer.
                      // peer->proxy contains the official address,
                      // i.e., application's understanding on the peer's address.
                      // peer->agent contains the address of peer's inagent.
                      // peer->local contains the address to which packets are needed
                      // to sent actually.
        UDP_peer *peerCache; // contains peers' information that the socket has recently
                           // contacted.
        /* receiving buffer */
        // the receiving buffer can contain upto MAX_UDP_RBUF msgs. Since TCP/IP does
        // not have 'peek' service, DPF needs to receive UDP msgs and checks if they
        // are DUMMY, which is used for opening UDP pinholes at firewalls and NAT. If
        // the msg is not DUMMY, then it needs to queue the msg to pass to the app
        // when it calls recv or recvfrom
        struct {
            unsigned short no_msg; // no of msgs get queued
            UDP_msg *head;
            UDP_msg *tail;
        } rbuf;

        /* common routines */
        DPF_UDP_sock(int owner);
        ~DPF_UDP_sock();
        int make_fresh();
        // move peer to peerCache
        void move_peer_2_cache();
        // If the peer is not already in peerCache, then add to peerCache.
        // And then update lastContact with the current time
        void add_2_cache(struct sockaddr_in *from);
        // queue the packet to 'recv buffer'
        // if receive buffer is full, then the oldest msg will be dropped silently
        void queue(char *data, int size, struct sockaddr_in *from);
        // get a msg from the head of 'recv buffer'
        // Note that caller must make sure that the queue is not empty
        // Note that the buffer passed from the caller must be big enough to
        // hold the msg being returned
        // @args
        //    - from: the address of the sender is returned via this arg
        //    - data: actual msg is returned via this arg
        // @ret: the size of msg being returned.
        unsigned short dequeue(struct sockaddr_in *from, void *data);
        void rm_peerCache();
        UDP_peer * find_peer(struct sockaddr_in *sin);
        // Remove peers who have not been recently heard from
        void flush_peers(time_t current);
        void tostr(const char *prefix, char *str);
        // Those two functions are for serialization and unserialization
        int stripe(char *outbuf);
        int unstripe(char *inbuf);
};

/****************************************************************************/
/*                  DPF_UDP_sock and supporting classes                    */
/****************************************************************************/
struct DPF_fdInfo {
    DPF_fd_type type;
    ClientSock *info;
};

// @arg:
//   - peer_ip: ip address (in network order) of the peer
//   - DPF: return true if DPF mechanism must be used to contact the peer.
//   - inagent: The address of the inagent of the peer is returned via this
//              argument. If 'DPF' is false, then this arg must be ignored
// @ret: true, if no error
//       false, if error
bool _DPF_get_route( uint32_t peer_ip, bool *DPF, struct sockaddr_in *inagent);

/* functions in data.C */
// Register DPF_TCP_sock or DPF_UDP_sock to _DPF_fd[fd] as a type 'type'
void _register_sock(int fd, DPF_fd_type type, ClientSock *info);
void _queue_connection(DPF_TCP_sock *listener, DPF_TCP_sock *hidden);
int _replace_fd(int target, int withfd);
// close fd, adjust _DPF_maxfd, clear _DPF_fd[fd] slot, and delete _DPF_fd[fd].info
// @ret: close(fd) result
int _close_fd(int fd);
// close fd, adjust _DPF_maxfd, and clear _DPF_fd[fd] slot w/o deleting
// _DPF_fd[fd]
// @ret: close(fd) result
int _close_duped_fd(int fd);
// adjust _DPF_maxfd. Caller shall make sure that new _DPF_maxfd is less than
// 'upper'
void _setmaxfd(int upper);
int _copy_sock(int fd, int newfd);
void _cleanup(void);

/* functions asynchronous event handling */
void _DPF_add_fds(int *maxfd, fd_set *rdfds, fd_set *wrfds);
void _DPF_handle_event(int maxfd, fd_set *rdfds, fd_set *wrfds);
void _DPF_heartbeat(time_t current);
void _DPF_wait_and_handle_event();

/* state transition machines */
void TCP_CLIENT_state_machine(DPF_TCP_sock *sock, time_t current);
void TCP_SERVER_state_machine(int fd, time_t current);
void TCP_SERV_MNG_state_machine(DPF_TCP_sock *sock, time_t current);
void UDP_state_machine(DPF_UDP_sock *sock, time_t current);

#endif
