#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "DPF_client.h"

extern DPF_fdInfo _DPF_fd[];
extern char *_client_status_str[];
extern uint32_t _DPF_myIP;
extern ExAddr *_DPF_inagent;
extern char *_DPF_route;

void
TCP_CLIENT_state_machine(DPF_TCP_sock *sock, time_t current)
{
    DPF_channel_result hsk_result;
    DPF_CMD_opcode opcode;
    struct sockaddr_in addr1, addr2;
    char byte;
    int error, result;
    socklen_t len;
    socklen_t addrlen;
    struct sockaddr_in sin;
    bool use_DPF;
    struct sockaddr_in inagent;

    time_t cur = time(NULL);

    while(1) {
        LOG(D_FULL, "\t[TCP_CLIENT, fd=%d] %s (current=%lu)\n",
            sock->fd, _client_status_str[sock->status], current);
        switch(sock->status) {
            case INIT_CONNECTION:
                // Condition to enter here:
                //    - from DPF_connect
                // Action:
                //    - set sock->myaddr if necessary
                //    - set sock->peer->local or agent based on routing table lookup
                // Next:
                //    - ACTIVE_CONNECTING_1, if no routing table or direct connection
                //    - CONNECT_DEST_AGENT_1, if DPF must be used for this connection
                // 1. bind the socket if necessary
                if (!(sock->myaddr)) {
                    memset(&sin, 0, sizeof(sin));
                    sin.sin_family = AF_INET;
                    sin.sin_addr.s_addr = _DPF_myIP;
                    if (_local_bind(sock->fd, &sin, sizeof(sin)) < 0) {
                        LOG(D_BASIC, "_local_bind failed: %s\n", strerror(errno));
                        sock->error = errno;
                        sock->status = CONNECTION_FAILED;
                        break;
                    }
                    addrlen = sizeof(sin);
                    if (getsockname(sock->fd, (struct sockaddr *) &sin, &addrlen) < 0) {
                        LOG(D_BASIC, "getsockname failed: %s\n", strerror(errno));
                        sock->error = errno;
                        sock->status = CONNECTION_FAILED;
                        break;
                    }
                    sock->myaddr = new ExAddr(&sin, NULL, NULL);
                    LOG(D_FULL, "\tLocal: %s\n", _sin2str(sock->myaddr->local));
                }
                // 2. see if DPF mechanism must be used
                if (!_DPF_route) {
                    sock->peer->set_local(sock->peer->proxy);
                    sock->status = ACTIVE_CONNECTING_1;
                    sock->entered = cur;
                    break;
                }
                if (!_DPF_get_route(sock->peer->proxy->sin_addr.s_addr, &use_DPF, &inagent)) {
                    ABORT("routing table read failed");
                }
                // 3. go to the next stage
                if (use_DPF) {
                    sock->peer->set_agent(&inagent);
                    sock->status = CONNECT_DEST_AGENT_1;
                } else {
                    sock->peer->set_local(sock->peer->proxy);
                    sock->status = ACTIVE_CONNECTING_1;
                }
                break;
            case CONNECT_DEST_AGENT_1:
                // Condition to enter here:
                //    - from INIT_CONNECTION
                // action: connect to peer's inagent
                sock->peer_agent = new DPF_channel();
                _register_sock(sock->peer_agent->get_fd(), TCP_PEER_IN_MNGER, sock);
                hsk_result = sock->peer_agent->connect((struct sockaddr *)sock->peer->agent,
                        sizeof(struct sockaddr_in));
                switch(hsk_result) {
                    case FAILED:
                        sock->status = CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        sock->status = CONNECT_DEST_AGENT_2;
                        sock->entered = cur;
                        return;
                    case SUCC:
                        sock->status = CONTACT_DEST_AGENT_1;
                        sock->entered = cur;
                        break;
                }
                break;
            case CONNECT_DEST_AGENT_2:
                // Condition to enter here:
                //    - sock->peer_agent->fd is read/write ready. Determined by
                //      sock->peer_agent->set_fd()
                //    - timeout occurred
                // Action: security handshake with peer's InAgent
                hsk_result = sock->peer_agent->finish_connect(current);
                switch(hsk_result) {
                    case FAILED:
                        sock->status = CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        return;
                    case SUCC:
                        sock->status = CONTACT_DEST_AGENT_1;
                        sock->entered = cur;
                        break;
                }
                break;
            case CONTACT_DEST_AGENT_1:
                // Condition to enter here:
                //    - from CONNECT_DEST_AGENT_1
                //    - from CONNECT_DEST_AGENT_2
                //    - sock->peer_agent->fd becomes write ready
                //    - timeout
                // Action: send TCP_CONNECT
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = CONNECTION_FAILED;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                if (!(sock->peer_agent->can_send())) {
                    if (!(sock->peer_agent->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        sock->status == CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    }
                } else {
                    result = sock->peer_agent->send_cmd(TCP_CONNECT, sock->myaddr->local,
                            sock->peer->proxy, 0); 
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "send(TCP_CONNECT) failed: %s\n", strerror(errno));
                        sock->status = CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    }
                }
                sock->status = CONTACT_DEST_AGENT_2;
                sock->entered = cur;
                return;
            case CONTACT_DEST_AGENT_2:
                // Pre: TCP_CONNECT was sent to dest agent
                // Condition to enter here:
                //    - sock->peer_agent->fd is read ready
                //    - timeout
                // Action: receive response to TCP_CONNECT
                // Next:
                //    - ACTIVE_CONNECTING_1, if ACTIVE_CONNECT is received
                if (current != 0) { // called by DPF_heartbeat
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = CONNECTION_FAILED;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                result = sock->peer_agent->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv(response to CONNECT) failed: %s\n", strerror(errno));
                    sock->status = CONNECTION_FAILED;
                    sock->error = errno;
                    break;
                }
                switch(opcode) {
                    case ACTIVE_CONNECT:
                        addr1.sin_family = AF_INET;
                        sock->peer->set_local(&addr1);
                        sock->status = ACTIVE_CONNECTING_1;
                        break;
                    case NAK:
                        sock->status = CONNECTION_FAILED;
                        sock->error = byte;
                        break;
                    default:
                        ABORT("unexpected msg received\n");
                }
                break;
            case ACTIVE_CONNECTING_1:
                // Condition to enter:
                //    - from INIT_CONNECTION
                //    - from CONTACT_DEST_AGENT_2
                if (sock->peer->connect(sock->fd) < 0) {
                    if (errno != EINPROGRESS) {
                        LOG(D_BASIC, "connecting to dest failed: %s\n", strerror(errno));
                        sock->status = CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    }
                    sock->status = ACTIVE_CONNECTING_2;
                    sock->entered = cur;
                    return;
                }
                sock->status = CONNECTION_SUCC;
                break;
            case ACTIVE_CONNECTING_2:
                // Pre: connection to the peer returned EINPROGRESS
                // Condition to enter here:
                //    - when sock->fd is write ready
                //    - timeout
                // Action: getsockopt
                if (current != 0) { // called by DPF_heartbeat
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = CONNECTION_FAILED;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                len = sizeof(error);
                if (getsockopt(sock->fd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
                    LOG(D_BASIC, "\t\tgetsockopt failed:%s\n", strerror(errno));
                    sock->status = CONNECTION_FAILED;
                    sock->error = errno;
                    break;
                }
                if (error) {
                    if (error == EINPROGRESS) return;
                    sock->status = CONNECTION_FAILED;
                    sock->error = error;
                } else {
                    sock->status = CONNECTION_SUCC;
                }
                break;
            case CONNECTION_FAILED:
                // Condition to enter here
                //    - something went wrong at any previous stage:
            case CONNECTION_SUCC:
                if (sock->peer_agent) {
                    (void) _close_duped_fd(sock->peer_agent->get_fd());
                    delete sock->peer_agent;
                    sock->peer_agent = NULL;
                }
                // Since sock->peer is set by the caller, it must be freed by the
                // caller. Though sock->myaddr may have been created inside this
                // state machine, we leave it undeleted since we may use this binding
                // later.
                return;
            default:
                ABORT("tcp_client_state_machine got to unknown state\n");
                return;
        } // end of switch(sock->status)
    } // end of while(1)
}
