#include <time.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "DPF_common.h"
#include "DPF_client.h"

extern DPF_fdInfo _DPF_fd[];
extern char *_client_status_str[];
extern ExAddr *_DPF_inagent;

// State-transition machine for TCP_UNCLASSIFIED and TCP_SERVER
// @args:
//    - current: the current time. Should be zero when this routine is called not by
//               timeout routine
void
TCP_SERV_MNG_state_machine(DPF_TCP_sock *sock, time_t current)
{
    DPF_channel_result hsk_result;
    DPF_CMD_opcode opcode;
    struct sockaddr_in addr1, addr2;
    char byte;
    int result;

    time_t cur = time(NULL);

    while(true) {
        LOG(D_FULL, "[TCP_SERV_MNG, fd=%d, mfd=%d] %s (%lu)\n", sock->fd, sock->inagent->get_fd(),
                _client_status_str[sock->status], current);
        switch(sock->status) {
            case CONNECT_INAGENT_0:
                // Condition to enter here:
                //    - from DPF_bind -> tcp_bind
                // action: connect to the inagent
                hsk_result = sock->inagent->connect((struct sockaddr *)_DPF_inagent->local,
                        sizeof(struct sockaddr_in));
                switch(hsk_result) {
                    case FAILED:
                        if (sock->app_status == LISTENING) {
                            sock->status = HALF_LISTEN;
                        } else {
                            sock->status = BIND_FAIL;
                        }
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        sock->status = CONNECT_INAGENT_1;
                        sock->entered = cur;
                        return;
                    case SUCC:
                        sock->status = REGISTERING_1;
                        sock->entered = cur;
                        break;
                }
                break;
            case CONNECT_INAGENT_1:
                // Condition to enter here:
                //    - sock->inagent->fd is read/write ready. Determined by
                //      sock->inagent->set_fd()
                //    - timeout occurred
                hsk_result = sock->inagent->finish_connect(current);
                switch(hsk_result) {
                    case FAILED:
                        sock->status = BIND_FAIL;
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        return;
                    case SUCC:
                        sock->status = REGISTERING_1;
                        sock->entered = cur;
                        break;
                }
                break;
            case REGISTERING_1:
                // Condition to enter here:
                //    - from CONNECT_INAGENT_0
                //    - from CONNECT_INAGENT_1
                //    - sock->inagent->fd write ready
                //    - timeout
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = BIND_FAIL;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                if (!(sock->inagent->can_send())) {
                    if (!(sock->inagent->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        sock->status == BIND_FAIL;
                        sock->error = errno;
                        break;
                    }
                } else {
                    result = sock->inagent->send_cmd(TCP_BIND, sock->myaddr->local,
                            sock->myaddr->proxy, 0);
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "send(TCP_BIND) failed: %s\n", strerror(errno));
                        sock->status = BIND_FAIL;
                        sock->error = errno;
                        break;
                    }
                }
                sock->status = REGISTERING_2;
                sock->entered = cur;
                return;
            case REGISTERING_2:
                // Pre: TCP_BIND request was sent
                // Condition to enter here:
                //    - sock->inagent->fd read ready
                //    - timeout
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = BIND_FAIL;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                result = sock->inagent->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv(TCP_BIND_OK) failed: %s\n", strerror(errno));
                    sock->status = BIND_FAIL;
                    sock->error = errno;
                    break;
                }
                if (opcode == TCP_BIND_OK) {
                    addr1.sin_family = AF_INET;
                    sock->myaddr->set_proxy(&addr1);
                    if (_DPF_inagent->connect(sock->hfd) < 0) {
                        LOG(D_BASIC,
                            "Can't connect sock->hfd to inagent's heartbeat address: %s\n",
                            strerror(errno));
                        sock->status = BIND_FAIL;
                        sock->error = errno;
                        break;
                    }
                    sock->status = FULL_BOUND;
                    return;
                } else if (opcode == NAK) {
                    sock->status = BIND_FAIL;
                    sock->error = byte;
                    break;
                }
                ABORT("invalid opcode received");
                return;
            case BIND_FAIL:
                return;
            case LISTENING_1:
                // Condition to enter here:
                //    - previous state = FULL_BOUND & DPF_listen called
                //    - sock->inagent->fd write ready
                //    - timeout
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = LISTEN_FAIL;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                if (!(sock->inagent->can_send())) {
                    if (!(sock->inagent->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        sock->status == LISTEN_FAIL;
                        sock->error = errno;
                        break;
                    }
                    sock->status = LISTENING_2;
                    sock->entered = cur;
                    return;
                }
                result = sock->inagent->send_cmd(LISTEN, sock->myaddr->local, NULL, 0);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "send(LISTEN) failed: %s\n", strerror(errno));
                    sock->status = LISTEN_FAIL;
                    sock->error = errno;
                    break;
                }
                sock->status = LISTENING_2;
                sock->entered = cur;
                return;
            case LISTENING_2:
                // Pre: LISTEN request was sent in LISTENING_1
                // Condition to enter here:
                //    - sock->inagent->fd read ready
                //    - heartbeat timer went off
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = LISTEN_FAIL;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                result = sock->inagent->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv(LISTEN_OK) failed: %s\n", strerror(errno));
                    sock->status = LISTEN_FAIL;
                    sock->error = errno;
                    break;
                }
                if (opcode == LISTEN_OK) {
                    sock->status = FULL_LISTEN;
                    break;
                } else if (opcode == NAK) {
                    sock->status = LISTEN_FAIL;
                    sock->error = byte;
                    break;
                }
                ABORT("invalid opcode received");
                return;
            case FULL_LISTEN:
                // Pre: the socket is registered as a listening socket to inagent
                // Condition to enter here
                //    - from LISTENING_2
                // Action:
                //    - connect sock->hfd to inagent's heartbeat address
                if (_DPF_inagent->connect(sock->hfd) < 0) {
                    LOG(D_BASIC,
                        "Can't connect sock->hfd to inagent's heartbeat address: %s\n",
                        strerror(errno));
                    sock->status = LISTEN_FAIL;
                    sock->error = errno;
                    break;
                }
                return;
            case LISTEN_FAIL:
                // Something went wrong
                return;
            default:
                ABORT("TCP_MANAGER_machine called for invalid status");
                break;
        } // switch(sock->status)
    } // while(true)
}
