#include <stdio.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "DPF_common.h"
#include "DPF_client.h"

extern DPF_fdInfo _DPF_fd[];
extern char *_client_status_str[];
extern uint32_t _DPF_myIP;
extern ExAddr *_DPF_inagent;
extern char *_DPF_route;

void
UDP_state_machine(DPF_UDP_sock *sock, time_t current)
{
    DPF_channel_result hsk_result;
    DPF_CMD_opcode opcode;
    struct sockaddr_in addr1, addr2;
    char byte;
    int result;
    socklen_t addrlen;
    struct sockaddr_in sin;
    bool use_DPF;
    struct sockaddr_in inagent;

    time_t cur = time(NULL);

    while (true) {
        LOG(D_FULL, "[UDP_machine, fd=%d] %s (%lu)\n", sock->fd,
                _client_status_str[sock->status], current);
        switch(sock->status) {
            case CONNECT_INAGENT_0:
                // Condition to enter here:
                //    - from DPF_bind -> udp_bind
                // action: connect to the inagent
                hsk_result = sock->inagent->connect((struct sockaddr *)_DPF_inagent->local,
                        sizeof(struct sockaddr_in));
                switch(hsk_result) {
                    case FAILED:
                        sock->status = HALF_BOUND;
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        sock->status = CONNECT_INAGENT_1;
                        sock->entered = cur;
                        return;
                    case SUCC:
                        sock->status = REGISTERING_1;
                        sock->entered = cur;
                        break;
                    default:
                        ABORT("Security handshake can't be done in a single swipe");
                }
                break;
            case CONNECT_INAGENT_1:
                // Condition to enter here:
                //    - sock->inagent->fd is read/write ready. Determined by
                //      sock->inagent->set_fd()
                //    - timeout occurred
                hsk_result = sock->inagent->finish_connect(current);
                switch(hsk_result) {
                    case FAILED:
                        sock->status = HALF_BOUND;
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        return;
                    case SUCC:
                        sock->status = REGISTERING_1;
                        sock->entered = cur;
                        break;
                }
                break;
            case REGISTERING_1:
                // Condition to enter here:
                //    - from REGISTERING_1
                //    - sock->inagent->fd write ready
                //    - timeout
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = BIND_FAIL;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                if (!(sock->inagent->can_send())) {
                    if (!(sock->inagent->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        sock->status == BIND_FAIL;
                        sock->error = errno;
                        break;
                    }
                } else {
                    result = sock->inagent->send_cmd(UDP_BIND, sock->myaddr->local,
                            sock->myaddr->proxy, 0);
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "send(UDP_BIND) failed: %s\n", strerror(errno));
                        sock->status = BIND_FAIL;
                        sock->error = errno;
                        break;
                    }
                }
                sock->status = REGISTERING_2;
                sock->entered = cur;
                return;
            case REGISTERING_2:
                // Pre: UDP_BIND request was sent
                // Condition to enter here:
                //    - sock->inagent->fd read ready
                //    - timeout
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = BIND_FAIL;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                result = sock->inagent->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv(UDP_BIND_OK) failed: %s\n", strerror(errno));
                    sock->status = BIND_FAIL;
                    sock->error = errno;
                    break;
                }
                if (opcode == UDP_BIND_OK) {
                    addr1.sin_family = AF_INET;
                    sock->myaddr->set_proxy(&addr1);
                    sock->status = FULL_BOUND;
                    break;
                } else if (opcode == NAK) {
                    sock->error = byte;
                    sock->status = BIND_FAIL;
                    break;
                }
                ABORT("invalid opcode received");
                return;
            case BIND_FAIL:
                return;
            case FULL_BOUND:
                // Pre: the socket is registered to inagent
                // Condition to enter:
                //    - connect sock->hfd to inagent's heartbeat address
                if (_DPF_inagent->connect(sock->hfd) < 0) {
                    LOG(D_BASIC,
                        "Can't connect sock->hfd to inagent's heartbeat address: %s\n",
                        strerror(errno));
                    sock->status = BIND_FAIL;
                    break;
                }
                return;
            case INIT_CONNECTION:
                // Condition to enter here:
                //    - from DPF_connect or sendto
                // Action:
                //    - set sock->myaddr if necessary
                //    - set sock->peer->local or agent based on routing table lookup
                // Next:
                //    - CONNECTION_SUCC, if no routing table or direct connection can be used
                //    - CONNECT_DEST_AGENT_1, if DPF must be used for this connection
                // 1. bind the socket if necessary.
                if (!(sock->myaddr)) {
                    memset(&sin, 0, sizeof(sin));
                    sin.sin_family = AF_INET;
                    sin.sin_addr.s_addr = _DPF_myIP;
                    if (_local_bind(sock->fd, &sin, sizeof(sin)) < 0) {
                        LOG(D_BASIC, "_local_bind failed: %s\n", strerror(errno));
                        sock->error = errno;
                        sock->status = CONNECTION_FAILED;
                        break;
                    }
                    addrlen = sizeof(sin);
                    if (getsockname(sock->fd, (struct sockaddr *) &sin, &addrlen) < 0) {
                        LOG(D_BASIC, "getsockname failed: %s\n", strerror(errno));
                        sock->error = errno;
                        sock->status = CONNECTION_FAILED;
                        break;
                    }
                    sock->myaddr = new ExAddr(&sin, NULL, NULL);
                    LOG(D_FULL, "\tLocal: %s\n", _sin2str(sock->myaddr->local));
                }
                // 2. read routing table. sendto or DPF_connect sets
                //    sock->peer->proxy with sinaddr passed from the application
                if (!_DPF_route) {
                    sock->peer->set_local(sock->peer->proxy);
                    sock->status = CONNECTION_SUCC;
                    break;
                }
                if (!_DPF_get_route(sock->peer->proxy->sin_addr.s_addr, &use_DPF, &inagent)) {
                    ABORT("routing table read failed");
                }
                // 3. go to the next stage
                if (use_DPF) {
                    sock->peer->set_agent(&inagent);
                    sock->status = CONNECT_DEST_AGENT_1;
                    sock->entered = cur;
                } else {
                    sock->peer->set_local(sock->peer->proxy);
                    sock->status = CONNECTION_SUCC;
                }
                break;
            case CONNECT_DEST_AGENT_1:
                // Condition to enter here:
                //    - from INIT_CONNECTION
                // action: connect to peer's inagent
                sock->peer_agent = new DPF_channel();
                _register_sock(sock->peer_agent->get_fd(), UDP_PEER_IN_MNGER, sock);
                hsk_result = sock->peer_agent->connect((struct sockaddr *)sock->peer->agent,
                        sizeof(struct sockaddr_in));
                switch(hsk_result) {
                    case FAILED:
                        sock->status = CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        sock->status = CONNECT_DEST_AGENT_2;
                        sock->entered = cur;
                        return;
                    case SUCC:
                        sock->status = CONTACT_DEST_AGENT_1;
                        sock->entered = cur;
                        break;
                    default:
                        ABORT("Security handshake can't be done in a single swipe");
                }
                break;
            case CONNECT_DEST_AGENT_2:
                // Condition to enter here:
                //    - sock->peer_agent->fd is read/write ready. Determined by
                //      sock->peer_agent->set_fd()
                //    - timeout occurred
                // Action: security handshake with peer's InAgent
                hsk_result = sock->peer_agent->finish_connect(current);
                switch(hsk_result) {
                    case FAILED:
                        sock->status = CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    case BLOCKED:
                        return;
                    case SUCC:
                        sock->status = CONTACT_DEST_AGENT_1;
                        sock->entered = cur;
                        break;
                }
                break;
            case CONTACT_DEST_AGENT_1:
                // Pre: security handshake with the dest agent done.
                // Condition to enter here:
                //    - from CONNECT_DEST_AGENT_1
                //    - from CONNECT_DEST_AGENT_2
                //    - sock->peer_agent->fd becomes write ready
                //    - timeout
                // Action: send UDP_CONNECT
                if (current != 0) {
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = CONNECTION_FAILED;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                if (!(sock->peer_agent->can_send())) {
                    if (!(sock->peer_agent->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        sock->status == CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    }
                } else {
                    result = sock->peer_agent->send_cmd(UDP_CONNECT, sock->myaddr->local,
                            sock->peer->proxy, 0);
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "send(UDP_CONNECT) failed: %s\n", strerror(errno));
                        sock->status = CONNECTION_FAILED;
                        sock->error = errno;
                        break;
                    }
                }
                sock->status = CONTACT_DEST_AGENT_2;
                sock->entered = cur;
                return;
            case CONTACT_DEST_AGENT_2:
                // Pre: UDP_CONNECT was sent to dest agent
                // Condition to enter here:
                //    - sock->peer_agent->fd is read ready
                //    - timeout
                // Action: receive response to UDP_CONNECT
                // Next:
                //    - CONNECTION_SUCC, if ACTIVE_INIT
                if (current != 0) { // called by DPF_heartbeat
                    if (current - sock->entered > DPF_MAX_STAY) {
                        sock->status = CONNECTION_FAILED;
                        sock->error = ETIMEDOUT;
                        break;
                    }
                    return;
                }
                result = sock->peer_agent->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv(response to CONNECT) failed: %s\n", strerror(errno));
                    sock->status = CONNECTION_FAILED;
                    sock->error = errno;
                    break;
                }
                switch(opcode) {
                    case ACTIVE_INIT:
                        addr1.sin_family = AF_INET;
                        sock->peer->set_local(&addr1);
                        sock->status = CONNECTION_SUCC;
                        break;
                    case NAK:
                        sock->status = CONNECTION_FAILED;
                        sock->error = byte;
                        break;
                    default:
                        ABORT("unexpected msg received\n");
                }
                break;
            case CONNECTION_FAILED:
            case CONNECTION_SUCC:
                // Cleanup for sock->proxy, sock->myaddr, and resuming sock->status
                // will be done by DPF_connect or DPF_sendto
                if (sock->peer_agent) {
                    _close_duped_fd(sock->peer_agent->get_fd());
                    delete sock->peer_agent;
                    sock->peer_agent = NULL;
                }
                return;
            default:
                ABORT("UDP_machine is called for invalid status");
        } // switch
    } // while
}
