#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>

#include "DPF_client.h"

UDP_msg::UDP_msg(struct sockaddr_in *src, char *msg, unsigned short msglen)
{
    memcpy(&(this->from), src, sizeof(struct sockaddr_in));
    this->data = (char *)malloc(msglen);
    if (this->data == NULL) {
        ABORT("new UDP_msg: malloc failed: %s", strerror(errno));
    }
    memcpy(this->data, msg, msglen);
    this->size = msglen;
    this->next = NULL;
}

UDP_msg::~UDP_msg()
{
    if (this->data) free(this->data);
}

void
UDP_msg::tostr(const char *prefix, char *str)
{
    sprintf(str, "%s<msg: %d byte[from %s]> ",
            prefix, this->size, _sin2str(&(this->from)));
}

UDP_peer::UDP_peer(struct sockaddr_in *local,
                   struct sockaddr_in *proxy,
                   struct sockaddr_in *agent)
    : ExAddr(local, proxy, agent)
{
    lastContact = time(NULL);
    next = NULL;
}

void
UDP_peer::tostr(const char *prefix, char *str)
{
    ExAddr::tostr(prefix, str);
    sprintf(&str[strlen(str)], "lastContact=%s", ctime(&lastContact));
}

int
UDP_peer::stripe(char *buf)
{
    // the length of the serialization
    // lastContact(4 bytes)
    // serialization of ExAddr

    char str[50];
    uint32_t nlong;
    uint16_t nshort;
    int index = 2;
    // lastContact(4 bytes)
    nlong = htonl(lastContact);
    memcpy(&buf[index], &nlong, 4);
    sprintf(str, "[%d]lastContact=%lu", index, lastContact);
    index += 4;
    // serialization of ExAddr
    sprintf(&str[strlen(str)], "[%d]ExAddr_serialization", index);
    int ex_len = ExAddr::stripe(&buf[index]);
    index += ex_len;

    // the length of the serialization
    nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of UDP_peer: [0]len=%d%s\n", index-2, str);

    return index;
}

int
UDP_peer::unstripe(char *buf)
{
    // the length of the serialization
    // lastContact(4 bytes)
    // serialization of ExAddr

    char str[50];
    uint32_t nlong;
    uint16_t nshort;
    int index = 0;
    // the length of the serialization
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // lastContact(4 bytes)
    memcpy(&nlong, &buf[index], 4);
    lastContact = ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]lastContact=%lu", index, lastContact);
    index += 4;
    // serialization of ExAddr
    sprintf(&str[strlen(str)], "[%d]ExAddr_serialization", index);
    int ex_len = ExAddr::unstripe(&buf[index]);
    index += ex_len;

    LOG(D_FULL, "Unserialization of UDP_peer: %s\n", str);

    if (length != index-2) {
        ABORT("length does not match with the actual size");
    }

    return index;
}

DPF_UDP_sock::DPF_UDP_sock(int owner)
    : ClientSock(owner)
{
    peer = NULL;
    peerCache = NULL;
    rbuf.no_msg = 0;
    rbuf.head = rbuf.tail = NULL;
}

DPF_UDP_sock::~DPF_UDP_sock()
{
    if (peer) { delete peer; }
    if (peerCache) { rm_peerCache(); }
    if (rbuf.no_msg != 0) { rm_rbuf(); }
}

int
DPF_UDP_sock::make_fresh()
{
    if (myaddr) {
        delete myaddr;
        myaddr = NULL;
    }
    if (inagent) {
        _close_duped_fd(inagent->get_fd());
        delete inagent;
        inagent = NULL;
    }
    if (peer_agent) {
        _close_duped_fd(peer_agent->get_fd());
        delete peer_agent;
        peer_agent = NULL;
    }

    // make the fd fresh
    int tmp_sock = _udp_nonblock_socket();
    if (tmp_sock < 0) {
        LOG(D_BASIC, "make_fresh:temporary socket creation failed\n");
        return -1;
    }
    if (_replace_fd(fd, tmp_sock) < 0) {
        LOG(D_BASIC, "can't replace DPF_UDP_sock->fd with temp sock\n");
        return -1;
    }
    return 0;
}

void
DPF_UDP_sock::rm_peerCache()
{
    UDP_peer *peer = peerCache;
    while(peer) {
        UDP_peer *ptr = peer;
        peer = peer->next;
        delete ptr;
    }
    peerCache = NULL;
}

void
DPF_UDP_sock::rm_rbuf()
{
    UDP_msg *ptr, *msg = rbuf.head;
    while (msg) {
        ptr = msg;
        msg = msg->next;
        delete ptr;
    }
}

// Move peer to peerCache
void
DPF_UDP_sock::move_peer_2_cache()
{
    // see if the peer is already in the cache
    for (UDP_peer *ptr = peerCache; ptr; ptr = ptr->next) {
        if (ptr->same_proxy(peer)) {
            // found. No need to move
            delete peer;
            peer = NULL;
            return;
        }
    }
    // Not found. We need to move the peer
    peer->next = peerCache;
    peerCache = peer;
    peer = NULL;
    return;
}

// If the peer is not already in peerCache, then add to peerCache. And then update
// lastContact with the current time
void
DPF_UDP_sock::add_2_cache(struct sockaddr_in *from)
{
    UDP_peer *peer = find_peer(from);
    if (!peer) {
        peer = new UDP_peer(from, from, NULL);
        peer->next = peerCache;
        peerCache = peer;
    }
    peer->lastContact = time(NULL);
    return;
}

// find the peer in the peerCache whose proxy field is the same as 'sin'
UDP_peer *
DPF_UDP_sock::find_peer(struct sockaddr_in *sin)
{
    for (UDP_peer *ptr = peerCache; ptr; ptr = ptr->next) {
        if (ptr->same_proxy(sin)) {
            // found. No need to move
            return ptr;
        }
    }
    return NULL;
}

// queue the packet to 'recv buffer'
// if receive buffer is full, then the oldest msg will be dropped silently
void
DPF_UDP_sock::queue(char *data, int size, struct sockaddr_in *from)
{
    UDP_msg *msg;
    if (rbuf.no_msg >= MAX_UDP_RBUF) {
        msg = rbuf.head;
        rbuf.head = msg->next;
        delete msg;
        rbuf.no_msg--;
    }
    
    if (status == CONNECTION_SUCC) {
        msg = new UDP_msg(peer->local, data, size);
    } else {
        msg = new UDP_msg(from, data, size);
    }
    if (rbuf.tail == NULL) {
        rbuf.head = rbuf.tail = msg;
    } else {
        rbuf.tail->next = msg;
        rbuf.tail = msg;
    }
    rbuf.no_msg++;

    if (status != CONNECTION_SUCC) {
        return add_2_cache(from);
    }
}

// get a msg from the head of 'recv buffer'
// Note that caller must make sure that the queue is not empty
// Note that the buffer passed from the caller must be big enough to
// hold the msg being returned
// @args
//    - from: the address of the sender is returned via this arg
//    - data: actual msg is returned via this arg
// @ret: the size of msg being returned.
unsigned short
DPF_UDP_sock::dequeue(struct sockaddr_in *from, void *data)
{
    UDP_msg *msg = rbuf.head;
    unsigned short size = msg->size;
    if (from) {
        memcpy(from, &(msg->from), sizeof(struct sockaddr_in));
    }
    memcpy(data, msg->data, size);
    rbuf.head = msg->next;
    if (!rbuf.head) {
        rbuf.tail = NULL;
    }
    rbuf.no_msg--;
    delete msg;

    return size;
}

// Remove peers who have not been recently heard from
void
DPF_UDP_sock::flush_peers(time_t current)
{
    UDP_peer *ptr, *pptr = NULL, *peer = peerCache;
    char str[200];
    while(peer) {
        if (current - peer->lastContact > DPF_FLUSH_RATE) {
            ptr = peer;
            peer = peer->next;
            // remove the peer from the cache
            if (!pptr) { // first 
                peerCache = peer;
            } else {
                pptr->next = peer;
            }
            ptr->tostr("", str);
            LOG(D_FULL, "Deleting a client not heard from for a while: %s\n", str);
            delete ptr;
        } else {
            pptr = peer;
            peer = peer->next;
        }
    }
}

// dump DPF_UDP_sock into "str"
// Each new line should start with "prefix"
void
DPF_UDP_sock::tostr(const char *prefix, char *str)
{
    ClientSock::tostr(prefix, str);
    strcat(str, " peer=");
    if (peer) {
        peer->tostr("", &str[strlen(str)]);
    } else {
        strcat(str, "NULL");
    }
    // I need to append tostr of peerCache and rbuf later
}

int
DPF_UDP_sock::stripe(char *buf)
{
    // length of the serialization excluding this field(2 bytes)
    // UDP_peer *peer
    // UDP_peer *peerCache
        // the number of entries in the cache(2 bytes)
        // the concatenation of serializations of entries in the cache
    // ClientSock serialization
    // receiving buffer is not get inherited to child process

    char str[3000];
    uint32_t nlong;
    uint16_t nshort;
    int index = 2;
    // UDP_peer *peer
    sprintf(str, "[%d]peer_serialization", index);
    if (peer) {
        int peer_len = peer->stripe(&buf[index]);
        index += peer_len;
    } else {
        nshort = htons(0);
        memcpy(&buf[index], &nshort, 2);
        index += 2;
    }
    // UDP_peer *peerCache
        // the number of entries in the cache(2 bytes)
        // the concatenation of serializations of entries in the cache
    int cache_len;
    int saved_idx = index;
    sprintf(&str[strlen(str)], "[%d]peerCache_serialization", index);
    index += 2;
    UDP_peer *entry;
    for (entry = peerCache, cache_len = 0; entry; entry = entry->next, cache_len++) {
        sprintf(&str[strlen(str)], "[%d]entry", index);
        int entry_len = entry->stripe(&buf[index]);
        index += entry_len;
    }
    nshort = htons(cache_len);
    memcpy(&buf[saved_idx], &nshort, 2);
    // receiving buffer is not get inherited to child process
    if (rbuf.no_msg != 0) {
        LOG(D_FULL, "%d msgs discarded\n", rbuf.no_msg);
    }
    // the serialization of ClientSock
    sprintf(&str[strlen(str)], "[%d]ClientSock_serialization", index);
    int ClientSock_len = ClientSock::stripe(&buf[index]);
    index += ClientSock_len;

    // length of the serialization excluding this field(2 bytes)
    nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of DPF_UDP_sock: [0]len=%d%s\n", index-2, str);

    return index;
}

int
DPF_UDP_sock::unstripe(char *buf)
{
    // length of the serialization excluding this field(2 bytes)
    // UDP_peer *peer
    // UDP_peer *peerCache
        // the number of entries in the cache(2 bytes)
        // the concatenation of serializations of entries in the cache
    // the serialization of ClientSock
    // receiving buffer is not get inherited to child process

    char str[3000];
    uint32_t nlong;
    uint16_t nshort;
    int index = 0;
    // length of the serialization excluding this field(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // UDP_peer *peer
    memcpy(&nshort, &buf[index], 2);
    unsigned short peer_len = ntohs(nshort);
    sprintf(&str[strlen(str)], "[%d]peer_serialization", index);
    if (peer_len > 0) {
        if (!peer) {
            peer = new UDP_peer(NULL, NULL, NULL);
        }
        index += peer->unstripe(&buf[index]);
    } else {
        index += 2;
    }
    // UDP_peer *peerCache
        // the number of entries in the cache(2 bytes)
        // the concatenation of serializations of entries in the cache
    memcpy(&nshort, &buf[index], 2);
    unsigned short cache_len = ntohs(nshort);
    sprintf(&str[strlen(str)], "[%d]peerCache_serialization(%d entries)", index, cache_len);
    index += 2;
    for (; cache_len > 0; cache_len--) {
        UDP_peer *entry = new UDP_peer(NULL, NULL, NULL);
        sprintf(&str[strlen(str)], "[%d]entry", index);
        int entry_len = entry->unstripe(&buf[index]);
        index += entry_len;
        // put the entry to peerCache
        entry->next = peerCache;
        peerCache = entry;
    }
    // the serialization of ClientSock
    sprintf(&str[strlen(str)], "[%d]ClientSock_serialization", index);
    int ClientSock_len = ClientSock::unstripe(&buf[index]);
    index += ClientSock_len;

    LOG(D_FULL, "Unserialization of DPF_UDP_sock: %s\n", str);

    if (length != index-2) {
        ABORT("length does not match with the actual size of the serialization");
    }

    return index;
}
