#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];
extern char *_app_status_str[];
extern ExAddr *_DPF_inagent;
extern uint32_t _DPF_myIP;

// calculate proxy addr in the following way:
// - ip
//   : INADDR_ANY, if the socket is locally bound to a private ip, or
//   : the ip that the socket is actually bound to, i.e. the ip of 'local'
// - port
//   : the value that the app wants to bind to, or
//   : the port of local_addr, if public machine and no port specified
// @args:
//   - proxy: calculated proxy address is returned via this arg
//   - appaddr: the address that the application passed for binding to
//   - local: the address that the socket is locally bound
static void
calc_proxy(struct sockaddr_in *proxy, struct sockaddr_in *appaddr, struct sockaddr_in *local)
{
    proxy->sin_family = AF_INET;
    // set ip
    if (_priv_net(ntohl(local->sin_addr.s_addr))) {
        // If the socket is bound to a private ip, the application must have called
        // bind with ip = INADDR_ANY. So proxy->ip must be INADDR_ANY and proxy->port
        // be the same as appaddr->port. If the application has specified a port
        // number, then inagent will bind the socket to that port number.
        proxy->sin_addr.s_addr = htonl(INADDR_ANY);
        proxy->sin_port = appaddr->sin_port;
    } else {
        // If the socket is bound to a public ip, then no proxy socket will be
        // created. So the proxy address must be the same as 'local'
        proxy->sin_addr.s_addr = local->sin_addr.s_addr;
        proxy->sin_port = local->sin_port;
    }
    return;
}

static int
tcp_bind(int fd, struct sockaddr_in *my_addr, socklen_t addrlen)
{
    DPF_TCP_sock *sock = (DPF_TCP_sock *)_DPF_fd[fd].info;

    // status check
    if (sock->app_status != CREATED) {
        LOG(D_BASIC, "GCB_bind: fd[%d] is in invalid status: %s\n",
                fd, _app_status_str[sock->app_status]);
        errno = EINVAL;
        return -1;
    }

    // unbind if the socket has been bound for some purpose and the application wants to
    // bind it to a specific ip:port
    if (sock->myaddr &&
        (my_addr->sin_addr.s_addr != htonl(INADDR_ANY) || my_addr->sin_port != 0)) {
        if ((sock->make_fresh()) < 0) {
            LOG(D_BASIC, "socket is bound for other reason. Can't unbind it\n");
            return -1;
        }
    }

    // bind fd locally if necessary
    if (sock->myaddr == NULL) {
        struct sockaddr_in local;
        local.sin_family = AF_INET;
        if (my_addr->sin_addr.s_addr == htonl(INADDR_ANY)) {
            local.sin_addr.s_addr = _DPF_myIP;
        } else {
            local.sin_addr.s_addr = my_addr->sin_addr.s_addr;
        }
        local.sin_port = my_addr->sin_port;
        if (_local_bind(fd, &local, sizeof(local)) < 0) {
            LOG(D_BASIC, "_local_bind failed: %s\n", strerror(errno));
            sock->error = errno;
            return -1;
        }
        socklen_t len = sizeof(local);
        if (getsockname(fd, (struct sockaddr *)&local, &len) < 0) {
            LOG(D_BASIC, "DPF_bind: getsockname failed: %s\n", strerror(errno));
            sock->error = errno;
            return -1;
        }
        sock->myaddr = new ExAddr(&local, NULL, NULL);
        if (!(sock->myaddr)) { ABORT("new failed"); }
    }
    LOG(D_FULL, "\tLocal: %s\n", _sin2str(sock->myaddr->local));

    // If this machine does not have inagent, i.e., inbound connections to this
    // machine are allowed, or private ip is passed from the application, then we
    // have done
    //if (!_DPF_inagent || _priv_net(ntohl(my_addr->sin_addr.s_addr))) {
    if (!_DPF_inagent) {
        //LOG(D_BASIC, "No inagent is defined or private ip passed with bind call\n");
        LOG(D_BASIC, "No inagent is defined\n");
        LOG(D_BASIC, "This socket will not get registered to inagent\n");
        LOG(D_FULL, "\tOfficial: %s\n", _sin2str(sock->myaddr->local));
        sock->app_status = BOUND;
        sock->status = LOCAL_BOUND;
        sock->bound = true;
        return 0;
    }

    /* Now we need to register the socket to the inagent */
    // calculate (partial) proxy address
    struct sockaddr_in proxy;
    calc_proxy(&proxy, my_addr, sock->myaddr->local);
    sock->myaddr->set_proxy(&proxy);
    // create a channel that will be connected to the inagent
    sock->inagent = new DPF_channel();
    // register fd of the channel just created
    _register_sock(sock->inagent->get_fd(), TCP_IN_MNGER, sock);
    // create sock->hfd
    sock->hfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock->hfd < 0) {
        LOG(D_BASIC, "can't create a UDP socket: %s\n",strerror(errno));
        return -1;
    }
    // connect fd to the inagent, send BIND, recv BIND_OK
    sock->status = CONNECT_INAGENT_0;
    sock->entered = time(NULL);
    TCP_SERV_MNG_state_machine(sock, 0);
    if (sock->status == BIND_FAIL) {
        _close_duped_fd(sock->inagent->get_fd());
        delete sock->inagent;
        sock->inagent = NULL;
        close(sock->hfd);
        sock->hfd = -1;
        sock->app_status = CREATED;
        sock->status = FRESH;
        // making this socket fresh will be done if the app calls bind again
        errno = sock->error;
        return -1;
    }
    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->status == FULL_BOUND) {
            sock->app_status = BOUND;
            sock->bound =true;
            LOG(D_FULL, "\tOfficial: %s\n", _sin2str(sock->myaddr->proxy));
            _DPF_fd[sock->hfd].type = HEARTBEAT_FD;
            return 0;
        } else if (sock->status == BIND_FAIL) {
            _close_duped_fd(sock->inagent->get_fd());
            delete sock->inagent;
            sock->inagent = NULL;
            close(sock->hfd);
            sock->hfd = -1;
            sock->app_status = CREATED;
            sock->status = FRESH;
            // making this socket fresh will be done if the app calls bind again
            errno = sock->error;
            return -1;
        }
    }
}

static int
udp_bind(int fd, struct sockaddr_in *my_addr, socklen_t addrlen)
{
    DPF_UDP_sock *sock = (DPF_UDP_sock *)_DPF_fd[fd].info;

    // status check
    if (sock->app_status != CREATED) {
        LOG(D_BASIC, "GCB_bind: fd[%d] is in invalid status: %s\n",
                fd, _app_status_str[sock->app_status]);
        errno = EINVAL;
        return -1;
    }

    // unbind if the socket has bound for other purpose and the application wants to
    // bind it to a specific ip:port
    if (sock->myaddr &&
        (my_addr->sin_addr.s_addr != htonl(INADDR_ANY) || my_addr->sin_port != 0)) {
        if ((sock->make_fresh()) < 0) {
            LOG(D_BASIC, "socket is bound for other reason. Can't unbind it\n");
            return -1;
        }
        // purge peerCache
        if (sock->peerCache) {
            sock->rm_peerCache();
        }
    }

    // bind fd locally if necessary
    if (sock->myaddr == NULL) {
        struct sockaddr_in local;
        local.sin_family = AF_INET;
        if (my_addr->sin_addr.s_addr == htonl(INADDR_ANY)) {
            local.sin_addr.s_addr = _DPF_myIP;
        } else {
            local.sin_addr.s_addr = my_addr->sin_addr.s_addr;
        }
        local.sin_port = my_addr->sin_port;
        if (_local_bind(fd, &local, sizeof(local)) < 0) {
            LOG(D_BASIC, "_local_bind failed: %s\n", strerror(errno));
            sock->error = errno;
            return -1;
        }
        socklen_t len = sizeof(local);
        if (getsockname(fd, (struct sockaddr *)&local, &len) < 0) {
            LOG(D_BASIC, "DPF_bind: getsockname failed: %s\n", strerror(errno));
            sock->error = errno;
            return -1;
        }
        sock->myaddr = new ExAddr(&local, NULL, NULL);
        if (!(sock->myaddr)) { ABORT("new failed"); }
    }
    LOG(D_FULL, "\tLocal: %s\n", _sin2str(sock->myaddr->local));

    // If this machine does not have inagent, i.e., inbound connections to this
    // machine are allowed, or private ip is passed from the application, then we
    // have done
    if (!_DPF_inagent || _priv_net(ntohl(my_addr->sin_addr.s_addr))) {
        LOG(D_BASIC, "No inagent is defined or private ip passed with bind call\n");
        LOG(D_BASIC, "This socket will not get registered to inagent\n");
        LOG(D_FULL, "\tOfficial: %s\n", _sin2str(sock->myaddr->local));
        sock->status = LOCAL_BOUND;
        sock->app_status = BOUND;
        sock->bound = true;
        return 0;
    }

    /* Now we need to register the socket to the inagent */
    // calculate (partial) proxy address
    struct sockaddr_in proxy;
    calc_proxy(&proxy, my_addr, sock->myaddr->local);
    sock->myaddr->set_proxy(&proxy);
    // create a channel that will be connected to the inagent
    sock->inagent = new DPF_channel();
    // register fd of the channel just created
    _register_sock(sock->inagent->get_fd(), UDP_IN_MNGER, sock);
    // create sock->hfd
    sock->hfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock->hfd < 0) {
        LOG(D_BASIC, "can't create a UDP socket: %s\n",strerror(errno));
        return -1;
    }
    // connect fd to the inagent, send BIND, recv BIND_OK
    sock->status = CONNECT_INAGENT_0;
    sock->entered = time(NULL);
    UDP_state_machine(sock, 0);
    if (sock->status == BIND_FAIL) {
        _close_duped_fd(sock->inagent->get_fd());
        delete sock->inagent;
        sock->inagent = NULL;
        close(sock->hfd);
        sock->hfd = -1;
        sock->app_status = CREATED;
        sock->status = FRESH;
        // making this socket fresh will be done if the app calls bind again
        errno = sock->error;
        return -1;
    }
    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->status == FULL_BOUND) {
            _close_duped_fd(sock->inagent->get_fd());
            delete sock->inagent;
            sock->inagent = NULL;
            sock->app_status = BOUND;
            sock->bound =true;
            LOG(D_FULL, "\tOfficial: %s\n", _sin2str(sock->myaddr->proxy));
            _DPF_fd[sock->hfd].type = HEARTBEAT_FD;
            return 0;
        } else if (sock->status == BIND_FAIL) {
            _close_duped_fd(sock->inagent->get_fd());
            delete sock->inagent;
            sock->inagent = NULL;
            close(sock->hfd);
            sock->hfd = -1;
            sock->app_status = CREATED;
            sock->status = FRESH;
            // making this socket fresh will be done if the app calls bind again
            errno = sock->error;
            return -1;
        }
    }
}

extern "C" {
int
DPF_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    struct sockaddr_in *sinAddr = (struct sockaddr_in *)my_addr;
    LOG(D_FULL, "[DPF_bind(%d, %s, %d)]\n", fd, _sin2str(sinAddr), addrlen);

    // socket should be registered as TCP or UDP socket
    switch(_DPF_fd[fd].type) {
        case TCP_UNCLASSIFIED:
            return tcp_bind(fd, sinAddr, addrlen);
        case UDP:
            return udp_bind(fd, sinAddr, addrlen);
        case NOT_DEFINED:
            return bind(fd, my_addr, addrlen);
        default:
            LOG(D_BASIC, "DPF_bind is called for invalid fd\n");
            errno = EINVAL;
            return -1;
    }
}
} // extern "C"
