#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

static int
tcp_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
    char buf[50];
    unsigned short size;

    LOG(D_FULL, "[DPF_connect(%d, %s, %d)]\n",
            fd, _sin2str((struct sockaddr_in *)serv_addr), addrlen);

    DPF_TCP_sock *sock = (DPF_TCP_sock *)_DPF_fd[fd].info;

    DPF_client_status saved_status;
    switch (sock->status) {
        case FRESH:
            saved_status = FRESH;
            break;
        case FULL_BOUND:
            if (sock->inagent) {
                _close_duped_fd(sock->inagent->get_fd());
                delete sock->inagent;
                sock->inagent = NULL;
            }
            size = _stripe_UDP_msg(buf, TCP_OPEN_DEL, sock->myaddr->local, sock->myaddr->proxy);
            if (send(sock->hfd, buf, size, 0) < 0) {
                LOG(D_BASIC, "Can't send TCP_DEL: %s\n", strerror(errno));
            }
            sock->myaddr->set_proxy(NULL);
            _DPF_fd[sock->hfd].type = NOT_DEFINED;
            close(sock->hfd);
            sock->hfd = -1;
        case LOCAL_BOUND:
            saved_status = LOCAL_BOUND;
            break;
        default:
            LOG(D_BASIC, "DPF_connect called for invalid fd: %d\n", fd);
            errno = EINVAL;
            return -1;
    }

    // tag as client socket
    _DPF_fd[fd].type = TCP_CLIENT;

    if (!(sock->peer)) {
        sock->peer = new ExAddr(NULL, (struct sockaddr_in *)serv_addr, NULL);
        if (!(sock->peer)) { ABORT("new DPF_peer failed"); }
    }

    sock->status = INIT_CONNECTION;
    TCP_CLIENT_state_machine(sock, 0);

    if (sock->status == CONNECTION_FAILED) {
        goto error;
    } else if (sock->status == CONNECTION_SUCC) {
        return 0;
    }

    if (sock->nonblock) {
        errno = EINPROGRESS;
        return -1;
    }

    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->status == CONNECTION_SUCC) {
            sock->app_status = CONNECTED;
            return 0;
        } else if (sock->status == CONNECTION_FAILED) {
            goto error;
        }
    }

error:
    sock->status = saved_status;
    delete sock->peer;
    sock->peer = NULL;
    errno = sock->error;
    return -1;
}

static int
udp_unconnect(DPF_UDP_sock *sock, const struct sockaddr *serv_addr, socklen_t addrlen)
{
    int rst = 0;

    LOG(D_FULL, "[DPF_connect(%d, AF_UNSPEC)]\n", sock->fd);
    if (sock->status != CONNECTION_SUCC) {
        LOG(D_BASIC, "UDP unconnect is called for fd not connected\n");
        errno = EINVAL;
        return -1;
    }
    if (connect(sock->fd, serv_addr, addrlen) < 0) {
        LOG(D_BASIC, "udp_unconnect(%d, AF_UNSPEC) failed - %s\n",
                sock->fd, strerror(errno));
        rst = -1;
        sock->error = errno;
    }

    if (sock->peer) {
        sock->move_peer_2_cache();
    }
    if (sock->bound) {
        if (sock->myaddr->proxy) {
            if (sock->inagent) {
                sock->status = FULL_BOUND;
            } else {
                sock->status = HALF_BOUND;
            }
        } else {
            sock->status = LOCAL_BOUND;
        }
        sock->app_status = BOUND;
    } else {
        sock->make_fresh();
        sock->status = FRESH;
        sock->app_status = CREATED;
    }

    return rst;
}

static int
udp_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
    DPF_UDP_sock *sock = (DPF_UDP_sock *)_DPF_fd[fd].info;

    if (serv_addr->sa_family == AF_UNSPEC) {
        return udp_unconnect(sock, serv_addr, addrlen);
    } else if (serv_addr->sa_family != AF_INET || addrlen != sizeof(struct sockaddr_in)) {
        LOG(D_FULL, "udp_connect was called for non IPv4 address\n");
        if (connect(fd, serv_addr, addrlen) < 0) {
            sock->error = errno;
            sock->status = CONNECTION_FAILED;
            return -1;
        }
        sock->app_status = CONNECTED;
        sock->status = CONNECTION_SUCC;
        return 0;
    }

    LOG(D_FULL, "[DPF_connect(%d, %s, %d)]\n",
            fd, _sin2str((struct sockaddr_in *)serv_addr), addrlen);

    // if the socket is already connected, unconnect it. Unconnecting is simply done
    // by moving sock->peer to sock->peerCache
    if (sock->peer) {
        sock->move_peer_2_cache();
    }

    // search peer in sock->peerCache
    UDP_peer *peer = sock->find_peer((struct sockaddr_in *)serv_addr);
    if (peer) {
        sock->peer = new UDP_peer(peer->proxy, peer->local, peer->agent);
        if (connect(fd, (struct sockaddr *)peer->local, sizeof(struct sockaddr_in)) < 0) {
            LOG(D_BASIC, "\tthe peer is found in peerCache\n");
            LOG(D_BASIC, "\tthe direct connection failed: %s\n", strerror(errno));
            sock->error = errno;
            sock->status = CONNECTION_FAILED;
            return -1;
        }
        peer->lastContact = time(NULL);
        sock->app_status = CONNECTED;
        sock->status = CONNECTION_SUCC;
        return 0;
    }

    // mark that this socket is connecting to 'serv_addr'
    sock->peer = new UDP_peer(NULL, (struct sockaddr_in *)serv_addr, NULL);
    if (!(sock->peer)) { ABORT("new ExAddr failed"); }

    DPF_client_status saved_stat = sock->status;
    sock->status = INIT_CONNECTION;
    UDP_state_machine(sock, 0);
    if (sock->status == CONNECTION_SUCC) {
        goto finish;
    } else if (sock->status == CONNECTION_FAILED) {
        goto error;
    }
    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->status == CONNECTION_SUCC) {
            goto finish;
        } else if (sock->status == CONNECTION_FAILED) {
            goto error;
        }
    }

finish:
    if(connect(fd, (struct sockaddr *)sock->peer->local,
                sizeof(struct sockaddr_in)) < 0) {
        LOG(D_BASIC, "\tconnecting %d to %s...\n", fd, sock->peer->local);
        LOG(D_BASIC, "\tconnect failed\n");
        sock->error = errno;
        goto error;
    }
    sock->app_status = CONNECTED;
    sock->status = CONNECTION_SUCC;
    return 0;
error:
    if (sock->peer) {
        delete sock->peer;
        sock->peer = NULL;
    }
    sock->status = saved_stat;
    errno = sock->error;
    return -1;
}

extern "C" {
int
DPF_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    if (_DPF_fd[fd].type == NOT_DEFINED) {
        LOG(D_FULL, "connect(%d[NOT_DEFINED], ...)\n", fd);
        int result = connect(fd, serv_addr, addrlen);
        LOG(D_FULL, "\treturning %d\n", result);
        return result;
    }

    if (_DPF_fd[fd].type == TCP_UNCLASSIFIED) {
        return tcp_connect(fd, serv_addr, addrlen);
    } else if (_DPF_fd[fd].type == UDP) {
        return udp_connect(fd, serv_addr, addrlen);
    } else {
        LOG(D_BASIC, "connect called invalid fd [%d], ...)\n", fd);
        errno = EINVAL;
        return -1;
    }
}
} // extern "C"
