#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "DPF_client.h"

char *_client_status_str[] = {
    "FRESH",
    "INIT_CONNECTION", "CONNECT_DEST_AGENT_00", "CONNECT_DEST_AGENT_01",
    "CONNECT_DEST_AGENT_02", "CONNECT_DEST_AGENT_03", "CONNECT_DEST_AGENT_1",
    "CONNECT_DEST_AGENT_2", "CONTACT_DEST_AGENT_1", "CONTACT_DEST_AGENT_2",
    "ACTIVE_CONNECTING_00", "ACTIVE_CONNECTING_01", "ACTIVE_CONNECTING_02",
    "ACTIVE_CONNECTING_03", "ACTIVE_CONNECTING_1", "ACTIVE_CONNECTING_2",
    "PASSIVE_CONNECTING_1", "PASSIVE_CONNECTING_2",
    "RELAY_CONNECTING_00", "RELAY_CONNECTING_01", "RELAY_CONNECTING_02",
    "RELAY_CONNECTING_03", "RELAY_CONNECTING_1", "RELAY_CONNECTING_2",
    "FINISHING", "CONNECTION_FAILED", "CONNECTION_SUCC",
    "CONNECT_INAGENT_0", "CONNECT_INAGENT_1", "REGISTERING_1", "REGISTERING_2",
    "LOCAL_BOUND", "BIND_FAIL", "HALF_BOUND", "FULL_BOUND", "LISTENING_1", "LISTEN_RECOVERING",
    "LISTENING_2", "LOCAL_LISTEN", "HALF_LISTEN", "LISTEN_FAIL", "FULL_LISTEN", "CONTACTING"
};
char *_app_status_str[] = {
    "CREATED", "BOUND", "LISTENING", "CONNECTED"
};

int _DPF_initialized = 0;
int _DPF_maxfd = 0;
DPF_fdInfo _DPF_fd[DPF_OPEN_MAX + 1];
uint32_t _DPF_myIP;
// local = sinful address that the client must contact
// agent = sinful address that heartbeat must be sent to
ExAddr *_DPF_inagent = NULL;
char *_DPF_route = NULL; // the name of the file that tells connection method
time_t _DPF_last_heartbeat = 0;

void
_register_sock(int fd, DPF_fd_type type, ClientSock *info)
{
    _DPF_fd[fd].type = type;
    _DPF_fd[fd].info = info;
    if (fd >= _DPF_maxfd) {
        _DPF_maxfd = fd + 1;
    }
}

void
_queue_connection(DPF_TCP_sock *listener, DPF_TCP_sock *hidden)
{
    if (listener->chead) {
        listener->ctail->chead = hidden; // ctail as prev pointer
        hidden->ctail = listener->ctail; // chead as next pointer
        listener->ctail = hidden;
    } else {
        listener->chead = listener->ctail = hidden;
        hidden->ctail = listener;
    }
}

/* Replace target fd by another fd */
// @arg:    target: fd to be replaced. Socket options of target remain the same
//          withfd: fd by which "target" is replaced
// @return: 0, if succeed
//          -1, otherwise
// Note: Both fds are closed if failure occurs at some point
int
_replace_fd(int target, int withfd)
{
    int val;

    // copy socket options
    if ((val = fcntl(target, F_GETFL, 0)) < 0) {
        LOG(D_BASIC, "_replace_fd: fcntl: %s\n", strerror(errno));
        return -1;
    }
    if (fcntl(withfd, F_SETFL, val) < 0) {
        LOG(D_BASIC, "_replace_fd: fcntl: %s\n", strerror(errno));
        return -1;
    }
    if (dup2(withfd, target) < 0) {
        close(withfd);
        return -1;
    }
    close(withfd);

    return 0;
}

int
_close_fd(int fd)
{
    if (fd == _DPF_maxfd - 1) {
        _setmaxfd(fd);
    }

    DPF_TCP_sock *tcp_sock;
    DPF_UDP_sock *udp_sock;
    switch (_DPF_fd[fd].type) {
        case TCP_UNCLASSIFIED:
        case TCP_CLIENT:
        case TCP_SERVER:
        case HIDDEN:
        case ACCEPTED:
            tcp_sock = (DPF_TCP_sock *) _DPF_fd[fd].info;
            delete tcp_sock;
            break;
        case UDP:
            udp_sock = (DPF_UDP_sock *) _DPF_fd[fd].info;
            delete udp_sock;
            break;
        default:
            ABORT("_close_fd(%d) called for invalid fd", fd);
    }

    _DPF_fd[fd].type = NOT_DEFINED;
    _DPF_fd[fd].info = NULL;

    return close(fd);
}

int
_close_duped_fd(int fd)
{
    // adjust _DPF_maxfd
    if (fd == _DPF_maxfd - 1) {
        _setmaxfd(fd);
    }
    // clear _DPF_fd slot
    _DPF_fd[fd].type = NOT_DEFINED;
    _DPF_fd[fd].info = NULL;
    // close fd
    return close(fd);
}

void
_setmaxfd(int upper)
{
    if (_DPF_maxfd == 0) return;

    for (int i=upper-1; i>=0; i--) {
        if (_DPF_fd[i].type != NOT_DEFINED) {
            _DPF_maxfd = i + 1;
            return;
        }
    }
    LOG(D_FULL, "\t<setmaxfd> no DPF or mng socket remained\n");
    _DPF_maxfd = 0;
}

int
_copy_sock(int fd, int newfd)
{
    ClientSock *info = (ClientSock *)_DPF_fd[fd].info;

    LOG(D_FULL, "\tDPF_duped %d to %d\n", fd, newfd);

    // increase reference count of _CB_socket
    info->refs++;
    // SockInfo->fd has always the smallest fd
    if (info->fd > newfd) {
        info->fd = newfd;
    }
    LOG(D_FULL, "\tinfo->refs = %d\n", info->refs);
    // register newfd
    _DPF_fd[newfd].type = _DPF_fd[fd].type;
    _DPF_fd[newfd].info = info;
    // adjust _CB_maxfd if necessary
    if (_DPF_maxfd <= newfd) {
        _DPF_maxfd = newfd + 1;
    }

    return 0;
}
