#include <string.h>
#include <errno.h>
#include <time.h>

#include "DPF_client.h"

extern int _DPF_maxfd;
extern DPF_fdInfo _DPF_fd[];
extern time_t _DPF_last_heartbeat;

static void
handleUDPmsg(DPF_UDP_sock *sock)
{
    char buf[DPF_MAX_UDP_PACKET];
    struct sockaddr_in from;
    socklen_t addrlen = sizeof(struct sockaddr_in);
    // recv a UDP msg
    int rcved;
    if (sock->status == CONNECTION_SUCC) {
        rcved = recv(sock->fd, buf, DPF_MAX_UDP_PACKET, 0);
    } else {
        rcved = recvfrom(sock->fd, buf, DPF_MAX_UDP_PACKET, 0,
            (struct sockaddr *)&from, &addrlen);
        LOG(D_FULL, "handleUDPmsg: [fd=%d]packet[%d bytes] rcved  from %s\n",
                sock->fd, rcved, _sin2str(&from));
    }
    if (rcved < 0) {
        if (errno == EAGAIN || errno == EINTR) { return; }
        LOG(D_BASIC, "handleUDPmsg(fd=%d) recvfrom failed: %s\n",
                sock->fd, strerror(errno));
        return;
    }
    // if the msg is a data packet, then queue it
    if ((unsigned char)buf[0] != DPF_TAG ||
        (unsigned char)buf[1] != DPF_VERSION ||
        (DPF_msg_type)buf[4] != DPF_CMD ||
        (DPF_CMD_opcode)buf[5] != DUMMY)
    {
        sock->queue(buf, rcved, &from);
        LOG(D_FULL, "\tqueued\n");
        return;
    }
    // if the msg is DUMMY, it must be a delayed or come-too-early msg,
    // so throw it away.
    LOG(D_FULL, "<< Delayed or too-early DUMMY to fd=%d\n", sock->fd);
    return;
}

void
_DPF_add_fds(int *maxfd, fd_set *rdfds, fd_set *wrfds)
{
    DPF_TCP_sock *tcp_sock;
    DPF_UDP_sock *udp_sock;

    //str[0] = '\0';
    for (int i = 0; i < _DPF_maxfd; i++) {
        int mfd;
        switch(_DPF_fd[i].type) {
            case TCP_UNCLASSIFIED:
                tcp_sock = (DPF_TCP_sock *) _DPF_fd[i].info;
                if (tcp_sock->error) break;
                switch(tcp_sock->status) {
                    case CONNECT_INAGENT_1: // sock->inagent is connecting to InAgent
                        tcp_sock->inagent->setfd(maxfd, rdfds, wrfds);
                        break;
                    case REGISTERING_1: // sock->inagent is sending TCP_BIND
                        mfd = tcp_sock->inagent->get_fd();
                        FD_SET(mfd, wrfds);
                        if (*maxfd <= mfd) { *maxfd = mfd + 1; }
                        break;
                    case REGISTERING_2: // sock->inagent is receiving BIND_OK
                        mfd = tcp_sock->inagent->get_fd();
                        FD_SET(mfd, rdfds);
                        if (*maxfd <= mfd) { *maxfd = mfd + 1; }
                        break;
                }
                break;
            case TCP_SERVER:
                tcp_sock = (DPF_TCP_sock *) _DPF_fd[i].info;
                if (tcp_sock->error) break;
                switch(tcp_sock->status) {
                    case LISTENING_1: // sock->inagent is sending LISTEN
                        mfd = tcp_sock->inagent->get_fd();
                        FD_SET(mfd, wrfds);
                        if (*maxfd <= mfd) { *maxfd = mfd + 1; }
                        // sock->fd must be prepared to accept connections
                        FD_SET(i, rdfds);
                        if (*maxfd <= i) { *maxfd = i + 1; }
                        break;
                    case LISTENING_2: // sock->inagent is receiving LISTEN_OK
                        mfd = tcp_sock->inagent->get_fd();
                        FD_SET(mfd, rdfds);
                        if (*maxfd <= mfd) { *maxfd = mfd + 1; }
                        // sock->fd must be prepared to accept connections
                        FD_SET(i, rdfds);
                        if (*maxfd <= i) { *maxfd = i + 1; }
                        break;
                    case LOCAL_LISTEN:
                    case FULL_LISTEN:
                        FD_SET(i, rdfds);
                        if (*maxfd <= i) { *maxfd = i + 1; }
                        break;
                }
                break;
            case TCP_CLIENT:
                tcp_sock = (DPF_TCP_sock *) _DPF_fd[i].info;
                if (tcp_sock->error) break;
                switch(tcp_sock->status) {
                    case CONNECT_DEST_AGENT_2: // sock->peer_agent is connecting
                        tcp_sock->peer_agent->setfd(maxfd, rdfds, wrfds);
                        FD_CLR(i, wrfds);
                        break;
                    case CONTACT_DEST_AGENT_1: // sock->peer_agent is sending TCP_CONNECT
                        mfd = tcp_sock->peer_agent->get_fd();
                        FD_SET(mfd, wrfds);
                        if (*maxfd <= mfd) {
                            *maxfd = mfd + 1;
                        }
                        FD_CLR(i, wrfds);
                        break;
                    case CONTACT_DEST_AGENT_2: // sock->peer_agent is receiving
                        mfd = tcp_sock->peer_agent->get_fd();
                        FD_SET(mfd, rdfds);
                        if (*maxfd <= mfd) {
                            *maxfd = mfd + 1;
                        }
                        FD_CLR(i, wrfds);
                        break;
                    case ACTIVE_CONNECTING_2: // sock->fd is connecting
                        FD_SET(i, wrfds);
                        if (*maxfd <= i) {
                            *maxfd = i + 1;
                        }
                        break;
                }
                break;
            case UDP:
                udp_sock = (DPF_UDP_sock *) _DPF_fd[i].info;
                if (udp_sock->error) break;
                switch(udp_sock->status) {
                    case CONNECT_INAGENT_1: // sock->inagent is connecting
                        udp_sock->inagent->setfd(maxfd, rdfds, wrfds);
                        break;
                    case REGISTERING_1: // sock->inagent is sending REGISTERING_1
                        mfd = udp_sock->inagent->get_fd();
                        FD_SET(mfd, wrfds);
                        if (*maxfd <= mfd) {
                            *maxfd = mfd + 1;
                        };
                        break;
                    case REGISTERING_2: // sock->inagent is receiving UDP_BIND_OK
                        mfd = udp_sock->inagent->get_fd();
                        FD_SET(mfd, rdfds);
                        if (*maxfd <= mfd) {
                            *maxfd = mfd + 1;
                        };
                        break;
                    case CONNECT_DEST_AGENT_2: // sock->peer_agent is connecting
                        udp_sock->peer_agent->setfd(maxfd, rdfds, wrfds);
                        break;
                    case CONTACT_DEST_AGENT_1: // sock->peer_agent is sending UDP_CONNECT
                        mfd = udp_sock->peer_agent->get_fd();
                        FD_SET(mfd, wrfds);
                        if (*maxfd <= mfd) {
                            *maxfd = mfd + 1;
                        };
                        break;
                    case CONTACT_DEST_AGENT_2: // sock->peer_agent is receiving
                        mfd = udp_sock->peer_agent->get_fd();
                        FD_SET(mfd, rdfds);
                        if (*maxfd <= mfd) {
                            *maxfd = mfd + 1;
                        };
                        break;
                    default:
                        break;
                }
                // add owner for read ready check
                FD_SET(i, rdfds);
                if (*maxfd <= i) {
                    *maxfd = i + 1;
                };
                break;
            default:
                break;
        } // switch(_DPF_fd[i].type)
    } // for (int i=0; _DPF_maxfd; i++)
}

void
_DPF_handle_event(int maxfd, fd_set *rdfds, fd_set *wrfds)
{
    DPF_TCP_sock *tcp_sock;
    DPF_UDP_sock *udp_sock;

    for (int i=0; i < maxfd; i++) {
        if (!FD_ISSET(i, rdfds) && !FD_ISSET(i, wrfds)) { continue; }
        switch(_DPF_fd[i].type) {
            case TCP_IN_MNGER:
                tcp_sock = (DPF_TCP_sock *)_DPF_fd[i].info;
                TCP_SERV_MNG_state_machine(tcp_sock, 0);
                break;
            case TCP_SERVER:
                TCP_SERVER_state_machine(i, 0);
                break;
            case TCP_CLIENT:
            case TCP_PEER_IN_MNGER:
                tcp_sock = (DPF_TCP_sock *)_DPF_fd[i].info;
                if (tcp_sock->status != CONNECTION_SUCC) {
                    TCP_CLIENT_state_machine(tcp_sock, 0);
                }
                break;
            case UDP:
                udp_sock = (DPF_UDP_sock *)_DPF_fd[i].info;
                handleUDPmsg(udp_sock);
                break;
            case UDP_IN_MNGER:
            case UDP_PEER_IN_MNGER:
                udp_sock = (DPF_UDP_sock *)_DPF_fd[i].info;
                UDP_state_machine(udp_sock, 0);
                break;
            default:
                break;
        } // switch(_DPF_fd[i].type)
    } // for (int i=0; maxfd; i++)
}

void
_DPF_heartbeat(time_t current)
{
    DPF_TCP_sock *tcp_sock;
    DPF_UDP_sock *udp_sock;
    int mfd;
    unsigned short size;
    char buf[100];

    _DPF_last_heartbeat = current;

    for (int i=0; i<_DPF_maxfd; i++) {
        switch (_DPF_fd[i].type) {
            case NOT_DEFINED:
                break;
            case TCP_UNCLASSIFIED:
                tcp_sock = (DPF_TCP_sock *) _DPF_fd[i].info;
                switch(tcp_sock->status) {
                    case CONNECT_INAGENT_1: // sock->inagent is connecting
                    case REGISTERING_1: // sock->inagent is sending TCP_BIND
                    case REGISTERING_2: // waiting for BIND_OK
                        TCP_SERV_MNG_state_machine(tcp_sock, current);
                        break;
                    default:
                        break;
                }
                break;
            case TCP_SERVER:
                tcp_sock = (DPF_TCP_sock *) _DPF_fd[i].info;
                switch(tcp_sock->status) {
                    case LISTENING_1: // sock->inagent is sending LISTEN
                    case LISTENING_2: // sock->inagent is receiving
                        TCP_SERV_MNG_state_machine(tcp_sock, current);
                        break;
                    case FULL_LISTEN:
                        // send HEARTBEAT
                        size = _stripe_UDP_msg(buf, TCP_HEARTBEAT,
                                tcp_sock->myaddr->local, tcp_sock->myaddr->proxy);
                        (void) send(tcp_sock->hfd, buf, size, 0);
                        break;
                    default:
                        break;
                }
                break;
            case TCP_CLIENT:
                tcp_sock = (DPF_TCP_sock *) _DPF_fd[i].info;
                switch(tcp_sock->status) {
                    case CONNECT_DEST_AGENT_2: // sock->peer_agent is connecting
                    case CONTACT_DEST_AGENT_1: // sock->peer_agent is sending CONNECT
                    case CONTACT_DEST_AGENT_2: // sock->peer_agent is receiving the reply
                    case ACTIVE_CONNECTING_2: // connecting to the dest
                        TCP_CLIENT_state_machine(tcp_sock, current);
                        break;
                    default:
                        break;
                }
                break;
            case UDP:
                udp_sock = (DPF_UDP_sock *)_DPF_fd[i].info;
                switch(udp_sock->status) {
                    case CONNECT_INAGENT_1: // connecting to or handshaking with inagent
                    case REGISTERING_1: // sending UDP_BIND
                    case REGISTERING_2: // waiting for UDP_BIND_OK
                    case CONNECT_DEST_AGENT_2: // sock->peer_agent is connecting
                    case CONTACT_DEST_AGENT_1: // sock->peer_agent is sending CONNECT
                    case CONTACT_DEST_AGENT_2: // sock->peer_agent is receiving the reply
                        UDP_state_machine(udp_sock, current);
                        break;
                }
                udp_sock->flush_peers(current);
                if (udp_sock->app_status == BOUND && udp_sock->hfd >= 0) {
                    size = _stripe_UDP_msg(buf, UDP_HEARTBEAT,
                            udp_sock->myaddr->local, udp_sock->myaddr->proxy);
                    (void) send(udp_sock->hfd, buf, size, 0);
                }
                break;
            default:
                break;
        }
    }
}

void
_DPF_wait_and_handle_event()
{
    int maxfd;
    fd_set rdfds, wrfds;

    while(true) {
        time_t current = time(NULL);
        if (current - _DPF_last_heartbeat > DPF_HEARTBEAT_RATE) {
            _DPF_heartbeat(current);
        }
        maxfd = 0;
        FD_ZERO(&rdfds);
        FD_ZERO(&wrfds);
        _DPF_add_fds(&maxfd, &rdfds, &wrfds);
        if (maxfd == 0) { return; }
        struct timeval timeout;
        timeout.tv_sec = DPF_HEARTBEAT_RATE/4;
        timeout.tv_usec = 0;
        char str1[200], str2[200];
        strcpy(str1, _prtfds(maxfd, &rdfds));
        strcpy(str2, _prtfds(maxfd, &wrfds));
        LOG(D_FULL, "_DPF_wait_and_handle_event: selecting with rd=%s, wr=%s\n", str1, str2);
        int ready = select(maxfd, &rdfds, &wrfds, NULL, &timeout);
        if (ready < 0) {
            if (errno != EINTR) {
                LOG(D_BASIC, "_DPF_wait_and_handle_event: select failed -\
                        %s\n", strerror(errno));
                return;
            }
        } else if (ready == 0) {
            continue;
        }
        // here, ready > 0
        return _DPF_handle_event(maxfd, &rdfds, &wrfds);
    } // while(true)
}
