#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdarg.h>

#include "DPF.h"

int _priv_net(uint32_t ip);

int
useDPF()
{
    char *answer;

    answer = getenv("DPF_ENABLE");
    if (!answer || strcasecmp(answer, "yes")) {
        return 0;
    }
    return 1;
}

int
useGeneric(void)
{
    char *answer;

    answer = getenv("GENERIC_ENABLE");
    if (!answer || strcasecmp(answer, "yes")) {
        return 0;
    }
    return 1;
}

int
Generic_socket(int domain, int type, int protocol)
{
	if (useDPF()) {
		return DPF_socket(domain, type, protocol);
	}
	return socket(domain, type, protocol);
}

static void
calc_proxy(struct sockaddr_in *proxy, struct sockaddr_in *appaddr)
{
    proxy->sin_family = AF_INET;
    // set ip
    if (_priv_net(ntohl(appaddr->sin_addr.s_addr))) {
        // If Condor calls bind with private ip address, we must let DPF choose ip
        // address appropriately
        proxy->sin_addr.s_addr = htonl(INADDR_ANY);
        proxy->sin_port = appaddr->sin_port;
    } else {
        proxy->sin_addr.s_addr = appaddr->sin_addr.s_addr;
        proxy->sin_port = appaddr->sin_port;
    }
    return;
}
int
Generic_bind(int fd, struct sockaddr *my_addr, socklen_t addrlen)
{
    struct sockaddr_in proxy;
	if (useDPF()) {
        calc_proxy(&proxy, (struct sockaddr_in *)my_addr);
		return DPF_bind(fd, (struct sockaddr *)&proxy, sizeof(struct sockaddr_in));
	}
	return bind(fd, my_addr, addrlen);
}

int
Generic_listen(int fd, int backlog)
{
	if (useDPF()) {
		return DPF_listen(fd, backlog);
	}
	return listen(fd, backlog);
}

int
Generic_connect(int fd, const struct sockaddr *serv_addr, socklen_t addrlen)
{
	if (useDPF()) {
		return DPF_connect(fd, serv_addr, addrlen);
	}
	return connect(fd, serv_addr, addrlen);
}

int
Generic_accept(int fd, struct sockaddr * peer, socklen_t *addrlen)
{
	if (useDPF()) {
		return DPF_accept(fd, peer, addrlen);
	}
	return accept(fd, peer, addrlen);
}

int
Generic_select(
		int maxfd,
		fd_set *readfds,
		fd_set *writefds,
		fd_set *exceptfds,
		struct timeval *timeout)
{
	if (useDPF()) {
		return DPF_select(maxfd, readfds, writefds, exceptfds, timeout);
	}
	return select(maxfd, readfds, writefds, exceptfds, timeout);
}

int
Generic_recv(int fd, void *buf, size_t len, int flags)
{
	if (useDPF()) {
		return DPF_recv(fd, buf, len, flags);
	}
	return recv(fd, buf, len, flags);
}

int
Generic_read(int fd, void *buf, size_t count)
{
	if (useDPF()) {
		return DPF_read(fd, buf, count);
	}
	return read(fd, buf, count);
}

int
Generic_recvfrom(
		int fd,  void *buf, size_t len, int flags,
		struct sockaddr *from, socklen_t *fromlen)
{
	if (useDPF()) {
		return DPF_recvfrom(fd, buf, len, flags, from, fromlen);
	}
	return recvfrom(fd, buf, len, flags, from, fromlen);

}

int
Generic_send(int fd, const void *msg, size_t len, int flags)
{
	if (useDPF()) {
		return DPF_send(fd, msg, len, flags);
	}
	return send(fd, msg, len, flags);
}

int
Generic_sendto(
		int fd,
		const void *msg, size_t len, int flags,
		const struct sockaddr *to, socklen_t tolen)
{
	if (useDPF()) {
		return DPF_sendto(fd, msg, len, flags, to, tolen);
	}
	return sendto(fd, msg, len, flags, to, tolen);
}

int
Generic_write(int fd, const void *buf, size_t count)
{
	if (useDPF()) {
		return DPF_write(fd, buf, count);
	}
	return write(fd, buf, count);
}

int
Generic_close(int fd)
{
	if (useDPF()) {
		return DPF_close(fd);
	}
	return close(fd);
}

int
Generic_dup(int fd)
{
	if (useDPF()) {
		return DPF_dup(fd);
	}
	return dup(fd);
}

int
Generic_dup2(int oldfd, int newfd)
{
	if (useDPF()) {
		return DPF_dup2(oldfd, newfd);
	}
	return dup2(oldfd, newfd);
}

int
Generic_getsockname(int fd, struct sockaddr * name, socklen_t *namelen)
{
	if (useDPF()) {
		return DPF_getsockname(fd, name, namelen);
	}
	return getsockname(fd, name, namelen);
}

int
Generic_getsockopt(
		int fd,
		int level,
		int optname,
		void *opt_val,
		socklen_t *optlen)
{
	if (useDPF()) {
		return DPF_getsockopt(fd, level, optname, opt_val, optlen);
	}
	return getsockopt(fd, level, optname, opt_val, optlen);
}

int
Generic_fcntl(int fd, int cmd, void *arg)
{
    if (useDPF()) {
        return DPF_fcntl(fd, cmd, arg);
    }
    return fcntl(fd, cmd, arg);
}

pid_t
Generic_fork(void)
{
    if (useDPF()) {
        return DPF_fork();
    }
    return fork();
}

int
Generic_execv(const char *path, char *const argv[])
{
	if (useDPF()) {
		return DPF_execv(path, argv);
	}
	return execv(path, argv);
}

int
Generic_execl(const char *path, const char *arg, ...)
{
	va_list ap;
	char *next_arg;
	char *argv[20];
	int noargs = 0;

	argv[noargs++] = arg;

	va_start(ap, arg);
	while(1) {
		next_arg = va_arg(ap, char *);
		argv[noargs++] = next_arg;
		if (!next_arg) {
			break;
		}
	}
	va_end(ap);

	if (useDPF()) {
		return DPF_execv(path, argv);
	}
	return execv(path, argv);
}

int
Generic_execve(const char *path, char *const argv [], char *const envp[])
{
	if (useDPF()) {
		return DPF_execve(path, argv, envp);
	}
	return execve(path, argv, envp);
}

void
Generic_cleanup(void)
{
    return DPF_cleanup();
}
