#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_maxfd;
extern DPF_fdInfo _DPF_fd[];

int
DPF_inheritFrom()
{
    // inherit filename: .DPF_INHERIT + pid
    // Content: a list of DPF_TCP_sock or DPF_UDP_sock serializations.
    //          Each socket is either owner or duplicated socket. When a socket is
    //          duplicated, the one with the smallest fd becomes the owner and others
    //          duplicated
        // Serialization of owner socket:
            // type(1 byte)
            // the slot number where the socket is registered in _DPF_fd(2 bytes)
            // tag = 1(1 byte)
            // the serialization of DPF_TCP_sock or DPF_UDP_sock
        // Serialization of duplicate sockets:
            // type(1 byte)
            // the slot number where the socket is registered in _DPF_fd(2 bytes)
            // tag = 0(1 byte)
            // fd of the owner socket(2 bytes)
        // End of the list
            // type = -1

	LOG(D_FULL, "%d\t[DPF_inheritFrom]\n", getpid());

    // open inherit file
    char filename[100];
	strcpy(filename, "/tmp/.DPF_INHERIT");
	sprintf(&filename[strlen(filename)], "%d", getpid());
	LOG(D_FULL, "\tinherit filename = %s\n", filename);
    // open the file
    int inheritfd = open(filename, O_RDONLY);
    if (inheritfd < 0) {
        if (errno == ENOENT) {
            LOG(D_BASIC, "the inherit file does not exist. Done.\n");
            return 0;
        }
        LOG(D_BASIC, "can't open the inherit file: %s\n", strerror(errno));
        goto fail;
    }

    char inheritbuf[5000];
    uint16_t nshort;
    uint32_t nlong;
    DPF_TCP_sock *tcpsock, *hidden;
    DPF_UDP_sock *udpsock;
    char type;
    int fd;
    char tag;
    int owner;
    bool deep;
    while (true) {
        // read the type
        if (read(inheritfd, &type, 1) != 1) {
            LOG(D_BASIC, "can't read from the inherit file: %s\n", strerror(errno));
            goto fail;
        }
        if (type == -1) { break; }
        // read fd
        if (read(inheritfd, &nshort, 2) != 2) {
            LOG(D_BASIC, "can't read from the inherit file: %s\n", strerror(errno));
            goto fail;
        }
        fd = ntohs(nshort);
        // read tag
        if (read(inheritfd, &tag, 1) != 1) {
            LOG(D_BASIC, "can't read from the inherit file: %s\n", strerror(errno));
            goto fail;
        }
        if (tag == 1) { // owner socket
            // read the length field
            if (read(inheritfd, inheritbuf, 2) != 2) {
                LOG(D_BASIC, "can't read from the inherit file: %s\n", strerror(errno));
                goto fail;
            }
            memcpy(&nshort, inheritbuf, 2);
            unsigned short sock_len = ntohs(nshort);
            LOG(D_FULL, "Inheriting User socket(%d bytes): [type=%d][fd=%d][tag=1]\n",
                    sock_len+6, type, fd);
            // read the remaining part of DPF_sock serialization, which must be sock_len bytes
            if (read(inheritfd, &inheritbuf[2], sock_len) != sock_len) {
                LOG(D_BASIC, "can't read from the inherit file: %s\n", strerror(errno));
                goto fail;
            }
            switch((DPF_fd_type)type) {
                case TCP_UNCLASSIFIED:
                case TCP_SERVER:
                case TCP_CLIENT:
                case ACCEPTED:
                    deep = ((DPF_fd_type)type == TCP_SERVER) ? true : false;
                    tcpsock = new DPF_TCP_sock(-1); 
                    if (tcpsock->unstripe(inheritbuf, deep) != sock_len+2) {
                        LOG(D_BASIC, "tcpsock->unstripe failed\n");
                        goto fail;
                    }
                    // register owner
                    _register_sock(fd, (_DPF_fd_type)type, tcpsock);
                    // register hfd
                    if (tcpsock->hfd >= 0) {
                        _DPF_fd[tcpsock->hfd].type = HEARTBEAT_FD;
                    }
                    // register managers
                    if (tcpsock->inagent) {
                        _register_sock(tcpsock->inagent->get_fd(), TCP_IN_MNGER, tcpsock);
                    }
                    if (tcpsock->peer_agent) {
                        _register_sock(tcpsock->peer_agent->get_fd(), TCP_PEER_IN_MNGER, tcpsock);
                    }
                    // register sockets in the connection queue
                    for (hidden = tcpsock->chead; hidden; hidden = hidden->chead) {
                        _register_sock(hidden->fd, HIDDEN, hidden);
                    }
                    break;
                case UDP:
                    udpsock = new DPF_UDP_sock(-1); 
                    if (udpsock->unstripe(inheritbuf) != sock_len+2) {
                        LOG(D_BASIC, "udpsock->unstripe failed\n");
                        goto fail;
                    }
                    // register owner
                    _register_sock(fd, (_DPF_fd_type)type, udpsock);
                    // register hfd
                    if (udpsock->hfd >= 0) {
                        _DPF_fd[tcpsock->hfd].type = HEARTBEAT_FD;
                    }
                    // register managers
                    if (udpsock->inagent) {
                        _register_sock(udpsock->inagent->get_fd(), UDP_IN_MNGER, udpsock);
                    }
                    if (udpsock->peer_agent) {
                        _register_sock(udpsock->peer_agent->get_fd(), UDP_PEER_IN_MNGER, udpsock);
                    }
                    break;
                default:
                    LOG(D_BASIC, "inherit buf has strange type value\n");
                    goto fail;
            }
        } else { // duplicated socket
            // read owner fd
            if (read(inheritfd, &nshort, 2) != 2) {
                LOG(D_BASIC, "can't read from the inherit file: %s\n", strerror(errno));
                goto fail;
            }
            owner = ntohs(nshort);
            _register_sock(fd, _DPF_fd[owner].type, _DPF_fd[owner].info);
        }
    } // while(true)

    // delete the inherit file
    close(inheritfd);
    unlink(filename);
    return 0;

fail:
    close(inheritfd);
    unlink(filename);
    return -1;
}

int
DPF_inheritTo()
{
    // inherit filename: .DPF_INHERIT + pid
    // Content: a list of DPF_TCP_sock or DPF_UDP_sock serializations.
    //          Each socket is either owner or duplicated socket. When a socket is
    //          duplicated, the one with the smallest fd becomes the owner and others
    //          duplicated
        // Serialization of owner socket:
            // type(1 byte)
            // the slot number where the socket is registered in _DPF_fd(2 bytes)
            // tag = 1(1 byte)
            // the serialization of DPF_TCP_sock or DPF_UDP_sock
        // Serialization of duplicate sockets:
            // type(1 byte)
            // the slot number where the socket is registered in _DPF_fd(2 bytes)
            // tag = 0(1 byte)
            // fd of the owner socket(2 bytes)
        // End of the list
            // type = -1

	LOG(D_FULL, "%d\t[DPF_inheritTo]\n", getpid());

    // Scan _DPF_fd[] and close all GCB fds with close-on-exec bit set
    LOG(D_FULL, "Closing fd with close-on-exec set...\n");
    int rst;
    for (int i = 0; i <= DPF_OPEN_MAX; i++) {
        if (_DPF_fd[i].info) {
            rst = fcntl(i, F_GETFD);
            if (rst == -1) {
                LOG(D_FULL, "\tfcntl for fd=%d: %s\n", i, strerror(errno));
                DPF_close(i);
            } else if (rst) {
                LOG(D_FULL, "\tdeleting fd=%d because of it's close-on-exec tag set\n", i);
                DPF_close(i);
            }
        }
    }

    // Make up the name of the file that will contain the serialization
    char filename[30];
	strcpy(filename, ".DPF_INHERIT");
	sprintf(&filename[strlen(filename)], "%d.tmp", getpid());
	LOG(D_FULL, "\ttemp filename = %s\n", filename);
    // open the file
    mode_t omask = umask(0000);
    int inheritfd = creat(filename, S_IRWXU|S_IRWXG|S_IRWXO);
    if (inheritfd < 0) {
        LOG(D_BASIC, "can't create the file for inheriting sockets: %s\n", filename);
        return -1;
    }
    (void) umask(omask);

    uint16_t nshort;
    uint32_t nlong;
    int length;
	char inheritbuf[9000];
    DPF_TCP_sock *tcpsock;
    DPF_UDP_sock *udpsock;
    bool deep;
	for (int i = 0; i < _DPF_maxfd; i++) {
        switch(_DPF_fd[i].type) {
            case TCP_UNCLASSIFIED:
            case ACCEPTED:
            case TCP_SERVER:
                tcpsock = (DPF_TCP_sock *)_DPF_fd[i].info;
                inheritbuf[0] = _DPF_fd[i].type;
                nshort = htons(i);
                memcpy(&inheritbuf[1], &nshort, 2);
                if (tcpsock->fd == i) {
                    inheritbuf[3] = 1;
                    deep = (_DPF_fd[i].type == TCP_SERVER) ? true : false;
                    length = 4 + tcpsock->stripe(&inheritbuf[4], deep);
                    LOG(D_FULL, "Inheriting User socket(%d bytes): [type=%d][fd=%d][tag=1]\n",
                            length, _DPF_fd[i].type, i);
                } else {
                    inheritbuf[3] = 0;
                    nshort = htons(tcpsock->fd);
                    memcpy(&inheritbuf[4], &nshort, 2);
                    length = 6;
                    LOG(D_FULL, "Inheriting User socket(%d bytes): [type=%d][fd=%d][tag=0][owner=%d]\n",
                            length, _DPF_fd[i].type, i, tcpsock->fd);
                }
                if (write(inheritfd, inheritbuf, length) != length) { goto fail; }
                break;
            case TCP_CLIENT:
                tcpsock = (DPF_TCP_sock *)_DPF_fd[i].info;
                inheritbuf[0] = _DPF_fd[i].type;
                nshort = htons(i);
                memcpy(&inheritbuf[1], &nshort, 2);
                if (tcpsock->fd == i) {
                    inheritbuf[3] = 1;
                    length = 4 + tcpsock->stripe(&inheritbuf[4], false);
                } else {
                    inheritbuf[3] = 0;
                    nshort = htons(tcpsock->fd);
                    memcpy(&inheritbuf[4], &nshort, 2);
                    length = 6;
                }
                if (write(inheritfd, inheritbuf, length) != length) { goto fail; }
                break;
            case UDP:
                udpsock = (DPF_UDP_sock *)_DPF_fd[i].info;
                inheritbuf[0] = UDP;
                nshort = htons(i);
                memcpy(&inheritbuf[1], &nshort, 2);
                if (udpsock->fd == i) {
                    inheritbuf[3] = 1;
                    length = 4 + udpsock->stripe(&inheritbuf[4]);
                } else {
                    inheritbuf[3] = 0;
                    nshort = htons(udpsock->fd);
                    memcpy(&inheritbuf[4], &nshort, 2);
                    length = 6;
                }
                if (write(inheritfd, inheritbuf, length) != length) { goto fail; }
                break;
            default:
                break;
        }
    }
    
    // mark EOF
    inheritbuf[0] = -1;
    if (write(inheritfd, inheritbuf, 1) != 1) { goto fail; }

    // close the temporary file
    close(inheritfd);

    // rename the temporary to real file name
    char inheritfilename[100];
    sprintf(inheritfilename, "/tmp/%s", filename);
    inheritfilename[strlen(inheritfilename) - 4] = 0;
    rename(filename, inheritfilename);

    return 0;

fail:
    close(inheritfd);
    unlink(filename);
    return -1;
}
