#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

static int
tcp_recv(int fd, void *buf, int maxsize, int flags)
{
    DPF_TCP_sock *sock = (DPF_TCP_sock *)_DPF_fd[fd].info;
    
    if (sock->status != CONNECTION_SUCC) {
        LOG(D_BASIC, "DPF_recv called for fd not in valid status\n");
        errno = EINVAL;
        return -1;
    }

    // if pending error, return error immediately
    if (sock->error) {
        LOG(D_BASIC, "DPF_recv called for fd with a pending error\n");
        errno = sock->error;
        return -1;
    }

    // If nonblocking socket or nonblocking recv, then call recv and return the
    // result immediately. Note that every socket in DPF has been made as
    // nonblocking.
    int rst = recv(fd, buf, maxsize, flags);
    if (rst >= 0 || sock->nonblock || flags & MSG_DONTWAIT) {
        return rst;
    } else if (errno != EAGAIN) {
        LOG(D_BASIC, "\trecv failed: %s\n", strerror(errno));
        return -1;
    }

    // Now we have to wait until fd becomes read ready
    fd_set rdfds;
    FD_ZERO(&rdfds);
    FD_SET(fd, &rdfds);
    int ready = 0;
    while(!ready) {
        ready = DPF_select(fd+1, &rdfds, NULL, NULL, NULL);
        if (ready < 0) {
            if (errno == EINTR) { continue; }
            LOG(D_BASIC, "select failed: %s\n", strerror(errno));
            return -1;
        }
    }
    return recv(fd, buf, maxsize, flags);
}

static int
udp_recv(int fd, void *buf, int maxsize, int flags)
{
    DPF_UDP_sock *sock = (DPF_UDP_sock *)_DPF_fd[fd].info;

    // check status
    if (sock->status != CONNECTION_SUCC) {
        LOG(D_BASIC, "DPF_recv called for fd not connected\n");
        errno = EINVAL;
        return -1;
    }

    // if recv buffer is not empty, return the content
    if (sock->rbuf.no_msg != 0) {
        // dequeue a msg
        int size = sock->dequeue(NULL, buf);
        return size;
    } else if (sock->error) {
        LOG(D_BASIC, "DPF_recv called for fd with a pending error\n");
        errno = sock->error;
        return -1;
    } else if (sock->nonblock || flags & MSG_DONTWAIT) {
        errno = EAGAIN;
        return -1;
    }

    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->rbuf.no_msg == 0) { // recv buffer is empty
            if (sock->error) { // pending error
                errno = sock->error;
                return -1;
            }
        } else {
            // dequeue a msg
            int size = sock->dequeue(NULL, buf);
            return size;
        }
    }
}

extern "C" {
int
DPF_recv(int fd, void *msg, size_t len, int flags)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // check fd
    if (_DPF_fd[fd].type == TCP_CLIENT || _DPF_fd[fd].type == ACCEPTED) {
        //LOG(D_FULL, "DPF_recv(TCP[%d], msg, size=%d)\n", fd, len);
        return tcp_recv(fd, msg, len, flags);
    } else if (_DPF_fd[fd].type == UDP) {
        //LOG(D_FULL, "DPF_recv(UDP[%d], msg, size=%d)\n", fd, len);
        return udp_recv(fd, msg, len, flags);
    } else if (_DPF_fd[fd].type == NOT_DEFINED) {
        //LOG(D_FULL, "DPF_recv(NOT_DEFINED[%d], msg, size=%d)\n", fd, len);
        return recv(fd, msg, len, flags);
    } else {
        LOG(D_BASIC, "DPF_recv called for invalid fd=%d\n", fd);
        errno = EBADF;
        return -1;
    }
}

int
DPF_read(int fd, void *msg, size_t len)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // check fd
    if (_DPF_fd[fd].type == TCP_CLIENT || _DPF_fd[fd].type == ACCEPTED) {
        //LOG(D_FULL, "DPF_read(TCP[%d], msg, size=%d)\n", fd, len);
        return tcp_recv(fd, msg, len, 0);
    } else if (_DPF_fd[fd].type == UDP) {
        //LOG(D_FULL, "DPF_read(UDP[%d], msg, size=%d)\n", fd, len);
        return udp_recv(fd, msg, len, 0);
    } else if (_DPF_fd[fd].type == NOT_DEFINED) {
        //LOG(D_FULL, "DPF_read(NOT_DEFINED[%d], msg, size=%d)\n", fd, len);
        return read(fd, msg, len);
    } else {
        LOG(D_BASIC, "DPF_read called for invalid fd=%d\n", fd);
        errno = EBADF;
        return -1;
    }
}
} // extern "C"
