#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

// Global data structure
extern int _DPF_initialized;
extern int _DPF_maxfd;
extern DPF_fdInfo _DPF_fd[];
extern time_t _DPF_last_heartbeat;

static int adjustFDs(int maxfd, fd_set *orgrfds, fd_set *orgwfds,
        fd_set *rfds, fd_set *wfds, fd_set *frdfds, fd_set *fwrfds);

extern "C" {
int
DPF_select(int mxfd,
        fd_set *rdfds,
        fd_set *wrfds,
        fd_set *exfds,
        struct timeval *timeout)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    char str1[50], str2[50], str3[50];
    strcpy(str1, _prtfds(mxfd, rdfds));
    strcpy(str2, _prtfds(mxfd, wrfds));
    strcpy(str3, _prtfds(mxfd, exfds));
    if (timeout) {
        LOG(D_FULL, "[DPF_select(mxfd=%d, rd=%s, wr=%s, ex=%s <%d, %d>)]\n",
                mxfd, str1, str2, str3, timeout->tv_sec, timeout->tv_usec);
    } else {
        LOG(D_FULL, "[DPF_select(mxfd=%d, rd=%s, wr=%s, ex=%s <infinite>)]\n",
                mxfd, str1, str2, str3);
    }

    long long remained;
    if (timeout) {
        remained = timeout->tv_sec * 1000000 + timeout->tv_usec;
    } else {
        // need to be equal to or greater than DPF_HEARTBEAT_RATE
        remained = DPF_HEARTBEAT_RATE * 1000000;
    }

    // check fds
    for (int i=0; i<mxfd; i++) {
        if ((!rdfds || !FD_ISSET(i, rdfds)) &&
            (!wrfds || !FD_ISSET(i, wrfds)) &&
            (!exfds || !FD_ISSET(i, exfds)))
        { continue; }
        switch(_DPF_fd[i].type) {
            case TCP_IN_MNGER:
            case UDP_IN_MNGER:
            case TCP_OUT_MNGER:
            case UDP_OUT_MNGER:
            case TCP_PEER_IN_MNGER:
            case UDP_PEER_IN_MNGER:
            case HIDDEN:
                LOG(D_BASIC, "DPF_select called for invalid fd (=%d) type\n", i);
                errno = EBADF;
                return -1;
            default:
                break;
        }
    }

    // Some applications are smart enough to know that some kernels do not check
    // fds greater than mxfd and pass fd_set that is just as big as mx * 8. Hence
    // we must not access fds bigger than mxfd so as not to segfault
    int bytes = (mxfd % 8 == 0) ? mxfd/8: mxfd/8 + 1;
    char first = 1;
    fd_set frdfds, fwrfds;
    fd_set readfds, writefds, expfds;
    int maxfd;
    int fready;
    do {
        // trigger heartbeat and flush, if necessary
        time_t current = time(NULL);
        if (current - _DPF_last_heartbeat > DPF_HEARTBEAT_RATE) {
            _DPF_heartbeat(current);
        }
    
        // copy fds and maxfd
        FD_ZERO(&readfds);
        FD_ZERO(&writefds);
        FD_ZERO(&expfds);
        if (rdfds) { memcpy(&readfds, rdfds, bytes); }
        if (wrfds) { memcpy(&writefds, wrfds, bytes); }
        if (exfds) { memcpy(&expfds, exfds, bytes); }
        maxfd = mxfd;

        // Add fds
        _DPF_add_fds(&maxfd, &readfds, &writefds);

        // For the case that some fds are already ready, we do nonblocking select
        // for the first time.  We also chop timeout value into multiple of
        // GCB_HEARTBEAT_RATE so that we can send HEARTBEAT
        struct timeval tv;
        if (first) {
            tv.tv_sec = 0;
            tv.tv_usec = 0;
        } else if (remained > DPF_HEARTBEAT_RATE * 1000000) {
            tv.tv_sec = DPF_HEARTBEAT_RATE;
            tv.tv_usec = 0;
        } else {
            tv.tv_sec = remained / 1000000;
            tv.tv_usec = remained % 1000000;
        }
        struct timeval before;
        if (gettimeofday(&before, NULL) < 0) {
            LOG(D_BASIC, "DPF_select: gettimeofday - %s\n", strerror(errno));
            return -1;
        }
        strcpy(str1, _prtfds(maxfd, &readfds));
        strcpy(str2, _prtfds(maxfd, &writefds));
        strcpy(str3, _prtfds(maxfd, &expfds));
        LOG(D_FULL, "\tselecting with %d, rd=%s, wr=%s, ex=%s\n", maxfd, str1, str2, str3);
        int retval = select(maxfd, &readfds, &writefds, &expfds, &tv);
        first = 0;
        struct timeval after;
        if (gettimeofday(&after, NULL) < 0) {
            LOG(D_BASIC, "DPF_select: gettimeofday - %s\n", strerror(errno));
            return -1;
        }
        if (timeout) {
            remained = remained -
                (after.tv_sec - before.tv_sec) * 1000000 - (after.tv_usec - before.tv_usec);
        }
        if (retval < 0 ) {
            if (errno != EINTR) {
                LOG(D_BASIC, "DPF_select: rudp_select failed - %s\n", strerror(errno));
            }
            return -1;
        } else if (retval > 0) {
            strcpy(str1, _prtfds(maxfd, &readfds));
            strcpy(str2, _prtfds(maxfd, &writefds));
            strcpy(str3, _prtfds(maxfd, &expfds));
            LOG(D_FULL, "\tselect returned: ready=%d, rd=%s, wr=%s, ex=%s\n",
                    retval, str1, str2, str3);

            // handle async events that happen during selecting
            _DPF_handle_event(maxfd, &readfds, &writefds);
        }
        // Adjust read and write ready set
        FD_ZERO(&frdfds);
        FD_ZERO(&fwrfds);
        fready = adjustFDs(maxfd, rdfds, wrfds, &readfds, &writefds, &frdfds, &fwrfds);
        if (fready < 0) {
            LOG(D_BASIC, "DPF_select: adjustFDs failed\n");
            return -1;
        } else if (fready > 0) {
            break;
        }
    } while ( remained > 0);

    if (rdfds) {
        memcpy(rdfds, &frdfds, bytes);
    }
    if (wrfds) {
        memcpy(wrfds, &fwrfds, bytes);
    }
    if (exfds) {
        memcpy(exfds, &expfds, bytes);
    }
    strcpy(str1, _prtfds(maxfd, &frdfds));
    strcpy(str2, _prtfds(maxfd, &fwrfds));
    strcpy(str3, _prtfds(maxfd, &expfds));
    LOG(D_FULL, "\treturning ready = %d: rd=%s, wr=%s, ex=%s\n", fready, str1, str2, str3);
    return fready;
}
} // extern "C"

static int
adjustFDs(int maxfd, fd_set *orgrfds, fd_set *orgwfds, fd_set *rfds,
        fd_set *wfds, fd_set *frdfds, fd_set *fwrfds)
{
    int ready = 0;
    DPF_TCP_sock *tcp_sock;
    DPF_UDP_sock *udp_sock;

    for (int i = 0; i < maxfd; i++) {
        switch(_DPF_fd[i].type) {
            case TCP_SERVER:
                tcp_sock = (DPF_TCP_sock *)_DPF_fd[i].info;
                // read ready if connection queue is not empty
                if (orgrfds && FD_ISSET(i, orgrfds)) {
                    if (tcp_sock->chead) {
                        FD_SET(i, frdfds);
                        ready++;
                    }
                }
                break;
            case TCP_CLIENT:
                tcp_sock = (DPF_TCP_sock *)_DPF_fd[i].info;
                if (orgwfds && FD_ISSET(i, orgwfds)) {
                    if (tcp_sock->app_status != CONNECTED &&
                        (tcp_sock->status == CONNECTION_SUCC ||
                         tcp_sock->status == CONNECTION_FAILED))
                    {
                        // Application wants to know if a connection succeeds or fails
                        FD_SET(i, fwrfds);
                        ready++;
                    }
                    else if (tcp_sock->app_status == CONNECTED && FD_ISSET(i, wfds))
                    {
                        // Application wants to know if data can be sent
                        FD_SET(i, fwrfds);
                        ready++;
                    }
                }
                if (orgrfds && FD_ISSET(i, orgrfds)) {
                    if (tcp_sock->status == CONNECTION_SUCC &&
                        tcp_sock->app_status == CONNECTED &&
                        FD_ISSET(i, rfds))
                    {
                        FD_SET(i, frdfds);
                        ready++;
                    }
                }
                break;
            case UDP:
                udp_sock = (DPF_UDP_sock *)_DPF_fd[i].info;
                // read ready if rbuf is not empty
                if (orgrfds && FD_ISSET(i, orgrfds)) {
                    if (udp_sock->rbuf.no_msg > 0) { // rbuf is not empty
                        FD_SET(i, frdfds);
                        ready++;
                    }
                }
                if (orgwfds && FD_ISSET(i, orgwfds)) {
                    // UDP is always write ready
                    FD_SET(i, fwrfds);
                    ready++;
                }
                break;
            default: // return as the real select says
                // This is for:
                //    - NOT_DEFINED
                //    - TCP_UNCLASSIFIED
                //    - TCP_CLIENT
                //    - ACCEPTED
                if (orgrfds && FD_ISSET(i, orgrfds) && FD_ISSET(i, rfds)) {
                    FD_SET(i, frdfds);
                    ready++;
                }
                if (orgwfds && FD_ISSET(i, orgwfds) && FD_ISSET(i, wfds)) {
                    FD_SET(i, fwrfds);
                    ready++;
                }
                break;
        }
    }
    return ready;
}
