#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

static int
tcp_send(int fd, const void *data, int maxsize, int flags)
{
    DPF_TCP_sock *sock = (DPF_TCP_sock *)_DPF_fd[fd].info;
    
    if (sock->status != CONNECTION_SUCC) {
        LOG(D_BASIC, "DPF_send called for fd not in valid status\n");
        errno = EBADF;
        return -1;
    }

    // if pending error, return error immediately
    if (sock->error) {
        errno = sock->error;
        LOG(D_BASIC, "sock has a pending error. Returning error\n");
        return -1;
    }

    // If nonblocking socket or nonblocking send, then call send and return the
    // result immediately. Note that every socket in DPF has been made as
    // nonblocking.
    int rst = send(fd, data, maxsize, flags);
    if (rst > 0 || sock->nonblock || flags & MSG_DONTWAIT) {
        return rst;
    } else if (errno != EAGAIN) {
        LOG(D_BASIC, "\tsend failed: %s\n", strerror(errno));
        return -1;
    }

    // Now we have to wait until fd becomes write ready
    fd_set wrfds;
    FD_ZERO(&wrfds);
    FD_SET(fd, &wrfds);
    int ready = 0;
    while(!ready) {
        ready = DPF_select(fd+1, NULL, &wrfds, NULL, NULL);
        if (ready < 0) {
            if (errno == EINTR) { continue; }
            LOG(D_BASIC, "select failed: %s\n", strerror(errno));
            return -1;
        }
    }
    return send(fd, data, maxsize, flags);
}

static int
udp_send(int fd, const void *data, int datasize, int flags)
{
    if (datasize > DPF_MAX_UDP_PACKET) {
        errno = EINVAL;
        LOG(D_BASIC, "ERROR: msg is too big\n");
        return -1;
    }

    DPF_UDP_sock *sock = (DPF_UDP_sock *)_DPF_fd[fd].info;

    // check status
    if (sock->status != CONNECTION_SUCC) {
        LOG(D_BASIC, "DPF_send called for fd not connected\n");
        errno = EBADF;
        return -1;
    }

    // if pending error, return error immediately
    if (sock->error) {
        LOG(D_BASIC, "sock has a pending error. Returning error\n");
        errno = sock->error;
        return -1;
    }

    // send the msg
    return send(fd, data, datasize, flags);
}

extern "C" {
int
DPF_send(int fd, const void *msg, size_t len, int flags)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // check fd
    if (_DPF_fd[fd].type == TCP_CLIENT || _DPF_fd[fd].type == ACCEPTED) {
        LOG(D_FULL, "DPF_send(TCP[%d], msg, size=%d)\n", fd, len);
        return tcp_send(fd, msg, len, flags);
    } else if (_DPF_fd[fd].type == UDP) {
        LOG(D_FULL, "DPF_send(UDP[%d], msg, size=%d)\n", fd, len);
        return udp_send(fd, msg, len, flags);
    } else if (_DPF_fd[fd].type == NOT_DEFINED) {
        LOG(D_FULL, "DPF_send(NOT_DEFINED[%d], msg, size=%d)\n", fd, len);
        return send(fd, msg, len, flags);
    } else {
        LOG(D_BASIC, "DPF_send called for invalid fd=%d\n", fd);
        errno = EBADF;
        return -1;
    }
}

int
DPF_write(int fd, const void *msg, size_t len)
{
    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // check fd
    if (_DPF_fd[fd].type == TCP_CLIENT || _DPF_fd[fd].type == ACCEPTED) {
        //LOG(D_FULL, "DPF_write(TCP[%d], msg, size=%d)\n", fd, len);
        return tcp_send(fd, msg, len, 0);
    } else if (_DPF_fd[fd].type == UDP) {
        //LOG(D_FULL, "DPF_write(UDP[%d], msg, size=%d)\n", fd, len);
        return udp_send(fd, msg, len, 0);
    } else if (_DPF_fd[fd].type == NOT_DEFINED) {
        //LOG(D_FULL, "DPF_send(NOT_DEFINED[%d], msg, size=%d)\n", fd, len);
        return write(fd, msg, len);
    } else {
        LOG(D_BASIC, "DPF_send called for invalid fd=%d\n", fd);
        errno = EBADF;
        return -1;
    }
}
} // extern "C"
