#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <errno.h>

#include "DPF_client.h"
#include "DPF.h"

extern int _DPF_initialized;
extern DPF_fdInfo _DPF_fd[];

extern "C" {

int
DPF_sendto(int fd, const void *msg, size_t len, int flags,
       const struct sockaddr *to, socklen_t addrlen)
{
    int rst;

    if (!_DPF_initialized) {
        DPF_init(NULL, NULL, 0);
    }

    // check fd type
    if (_DPF_fd[fd].type == NOT_DEFINED) {
        return sendto(fd, msg, len, flags, to, addrlen);
    } else if (_DPF_fd[fd].type != UDP) {
        LOG(D_BASIC, "sendto(%d, ...) is called for not UDP\n", fd);
        errno = EINVAL;
        return -1;
    }

    // check peer address
    if (to->sa_family != AF_INET || addrlen != sizeof(struct sockaddr_in)) {
        LOG(D_FULL, "sendto was called for non IPv4 address\n");
        return sendto(fd, msg, len, flags, to, addrlen);
    }

    // check size
    if (len > DPF_MAX_UDP_PACKET) {
        LOG(D_BASIC, "sendto was called with too big a msg\n");
        errno = EINVAL;
        return -1;
    }

    LOG(D_FULL, "DPF_sendto(%d, msg(%d bytes), %s, %d)\n",
            fd, len, _sin2str((struct sockaddr_in *)to), addrlen);
    DPF_UDP_sock *sock = (DPF_UDP_sock *)_DPF_fd[fd].info;

    // check status
    if (sock->status == CONNECTION_SUCC) {
        LOG(D_BASIC, "sendto was called for a connected socket\n");
        errno = EINVAL;
        return -1;
    }

    // search peer in sock->peerCache
    UDP_peer *peer = sock->find_peer((struct sockaddr_in *)to);
    if (peer) {
        peer->lastContact = time(NULL);
        LOG(D_FULL, "\tthe peer is found in peerCache\n");
        return sendto(fd, msg, len, flags, (struct sockaddr *)peer->local,
                sizeof(struct sockaddr_in));
    }

    // Since peerCache does not have the entry for 'to', we need to open up a channel
    // to the peer
    //
    // save the status to resume later
    DPF_client_status saved_stat = sock->status;

    // mark that this socket is connecting to 'to'
    sock->peer = new UDP_peer(NULL, (struct sockaddr_in *)to, NULL);
    if (!(sock->peer)) { ABORT("new ExAddr failed"); }

    sock->status = INIT_CONNECTION;
    UDP_state_machine(sock, 0);
    if (sock->status == CONNECTION_SUCC) {
        goto finish;
    } else if (sock->status == CONNECTION_FAILED) {
        goto error;
    }
    while(true) {
        _DPF_wait_and_handle_event();
        if (sock->status == CONNECTION_SUCC) {
            goto finish;
        } else if (sock->status == CONNECTION_FAILED) {
            goto error;
        }
    }

finish:
    rst = sendto(fd, msg, len, flags, (struct sockaddr *)sock->peer->local,
            sizeof(struct sockaddr_in));
    sock->move_peer_2_cache();
    sock->status = saved_stat;
    return rst;
error:
    errno = sock->error;
    sock->status = saved_stat;
    if (sock->peer) {
        delete sock->peer;
        sock->peer = NULL;
    }
    return -1;
}

} // extern "C"
