#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include "DPF_common.h"

#define DPF_HSK_MAX_STAY 600

char *_DPF_hsk_stat_str[] = {
    "InitSec",
    "Connecting_1", "Connecting_2",
    "ClientHello", "ServerHello",
    "ClientCertificate", "ServerCertificate",
    "ComputeMasterSecret", "KeyACK", "ComputeKeyBlock",
    "ClientFinished", "ServerFinished",
    "Connected"
};
extern char *_CMD_opcode_str[];

// Client channel
DPF_channel::DPF_channel()
{
    fd = _tcp_nonblock_socket();
    if (fd < 0) { ABORT("socket creation failed"); }
    server = false;
    status = Connecting_1;
    cmd_buf = new DPF_RecBuf(fd, DPF_MAX_CMD_SIZE, DPF_TAG, DPF_VERSION);
}

// Server channel. 'newfd' is the fd that 'accept' has returned
DPF_channel::DPF_channel(int newfd)
{
    fd = newfd;
    server = true;
    status = Connected;
    cmd_buf = new DPF_RecBuf(fd, DPF_MAX_CMD_SIZE, DPF_TAG, DPF_VERSION);
}

DPF_channel::~DPF_channel()
{
    // fd will get closed by _close_fd or _close_duped_fd
    if (cmd_buf) { delete cmd_buf; }
}

// return the addr that fd is bound to
bool
DPF_channel::getaddr(struct sockaddr_in *addr)
{
    socklen_t addrlen = sizeof(struct sockaddr_in);
    if (getsockname(fd, (struct sockaddr *) addr, &addrlen) < 0) {
        LOG(D_BASIC, "getsockname failed: %s\n", strerror(errno));
        return false;
    }
    return true;
}

// bind fd to an ephemeral addr and return the addr that fd is bound to
bool
DPF_channel::bind(struct sockaddr_in *addr)
{
    struct sockaddr_in sin;

    memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = htonl(INADDR_ANY);
    sin.sin_port = htons(0);
    if (_local_bind(fd, &sin, sizeof(sin)) < 0) {
        LOG(D_BASIC, "_local_bind failed: %s\n", strerror(errno));
        return false;
    }
    return getaddr(addr);
}

DPF_channel_result
DPF_channel::TCP_handshake_client(
        struct sockaddr *addr,
        socklen_t addrlen,
        time_t current)
{
    int error;
    socklen_t len = sizeof(error);

    time_t cur = time(NULL);
    while(true) {
        LOG(D_FULL, "[TCP_handshake_client, fd=%d] %s (current=%lu)\n",
            fd, _DPF_hsk_stat_str[status], current);
        switch (status) {
            case Connecting_1:
                if (::connect(fd, addr, addrlen) < 0) {
                    if (errno == EINPROGRESS) {
                        status = Connecting_2;
                        entered = cur;
                        return BLOCKED;
                    }
                    LOG(D_BASIC, "connect failed: %s\n", strerror(errno));
                    return FAILED;
                }
                status = ClientHello;
                entered = cur;
                break;
            case Connecting_2:
                // Condition to enter here:
                //    - heartbeat timer went off
                //    - fd write ready
                if (current != 0) { // called by DPF_heartbeat
                    if (current - entered > DPF_HSK_MAX_STAY) {
                        errno = ETIMEDOUT;
                        return FAILED;
                    }
                    return BLOCKED;
                }
                error = 0;
                if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
                    LOG(D_BASIC, "getsockopt failed:%s\n", strerror(errno));
                    return FAILED;
                }
                if (error) {
                    if (error == EINPROGRESS) { return BLOCKED; }
                    errno = error;
                    return FAILED;
                }
                status = Connected;
                return SUCC;
            default:
                ABORT("Unknown state reached");
        }
    }
}

void
DPF_channel::add_fd_TCP_CLIENT(int *maxfd, fd_set *rdfds, fd_set *wrfds)
{
    switch (status) {
        case Connecting_2: // waiting for the connection completed
            FD_SET(fd, wrfds);
            if (*maxfd <= fd) { *maxfd = fd + 1; }
        default:
            break;
    }
    return;
}

// Stripe DPF command
// @args
//  - buf: the striped msg is returned via this arg. The striped msg is of the
//         format: DPF_CMD[0] + enc{opcode + addr1 + addr2 + byte} 
// @ret: the size of the striped msg in 'buf'
// Note: DPF_TAG, version, length must be prefixed by the caller
unsigned short
DPF_channel::stripe_cmd(char *buf, DPF_CMD_opcode opcode,
        struct sockaddr_in *addr1, struct sockaddr_in *addr2, char byte)
{
    char body[300];
    unsigned short body_len;
    char str[300];

    // stripe the body
    body[0] = (char)opcode;
    sprintf(str, "%s ", _CMD_opcode_str[opcode]);
    switch (opcode) {
        case UDP_BIND:
        case TCP_BIND:
        case UDP_OPEN:
        case TCP_OPEN:
        case UDP_OPEN_DEL:
        case TCP_OPEN_DEL:
        case DUMMY:
            memcpy(&body[1], &(addr1->sin_addr.s_addr), 4);
            memcpy(&body[5], &(addr1->sin_port), 2);
            sprintf(&str[strlen(str)], "local=%s ", _sin2str(addr1));
            memcpy(&body[7], &(addr2->sin_addr.s_addr), 4);
            memcpy(&body[11], &(addr2->sin_port), 2);
            sprintf(&str[strlen(str)], "proxy=%s", _sin2str(addr2));
            body_len = 13;
            break;
        case LISTEN:
            memcpy(&body[1], &(addr1->sin_addr.s_addr), 4);
            memcpy(&body[5], &(addr1->sin_port), 2);
            sprintf(&str[strlen(str)], "local=%s ", _sin2str(addr1));
            if (addr2) {
                memcpy(&body[7], &(addr2->sin_addr.s_addr), 4);
                memcpy(&body[11], &(addr2->sin_port), 2);
                sprintf(&str[strlen(str)], "proxy=%s", _sin2str(addr2));
                body_len = 13;
            } else {
                body_len = 7;
            }
            break;
        case UDP_BIND_OK:
        case TCP_BIND_OK:
        case UDP_OPEN_OK:
        case TCP_OPEN_OK:
        case CONTACT:
        case RELAY_CONTACT:
        case TRIGGER:
        case RELAY_TRIGGER:
            memcpy(&body[1], &(addr1->sin_addr.s_addr), 4);
            memcpy(&body[5], &(addr1->sin_port), 2);
            sprintf(&str[strlen(str)], "addr=%s ", _sin2str(addr1));
            body_len = 7;
            break;
        case LISTEN_OK:
            body_len = 1;
            break;
        case UDP_CONNECT:
        case TCP_CONNECT:
            memcpy(&body[1], &(addr1->sin_addr.s_addr), 4);
            memcpy(&body[5], &(addr1->sin_port), 2);
            sprintf(&str[strlen(str)], "from=%s ", _sin2str(addr1));
            memcpy(&body[7], &(addr2->sin_addr.s_addr), 4);
            memcpy(&body[11], &(addr2->sin_port), 2);
            sprintf(&str[strlen(str)], "to=%s ", _sin2str(addr2));
            // byte is used to denote if inbound is allowed to the connector
            body[13] = byte;
            if (byte == ALLOWED) {
                sprintf(&str[strlen(str)], "allowed");
            } else if (byte == INSIDE) {
                sprintf(&str[strlen(str)], "inside");
            } else {
                sprintf(&str[strlen(str)], "not blocked");
            }
            body_len = 14;
            break;
        case ACTIVE_INIT:
        case RELAY_INIT:
        case ACTIVE_CONNECT:
        case RELAY_CONNECT:
        case PASSIVE_INIT:
        case PASSIVE_CONNECT:
            memcpy(&body[1], &(addr1->sin_addr.s_addr), 4);
            memcpy(&body[5], &(addr1->sin_port), 2);
            sprintf(&str[strlen(str)], "%s ", _sin2str(addr1));
            body_len = 7;
            break;
        case PASSIVE_ACK:
        case DUMMY_ACK:
            body_len = 1;
            break;
        case NAK:
            body[1] = byte;
            sprintf(&str[strlen(str)], "err=%s ", strerror(byte));
            body_len = 2;
            break;
        default:
            ABORT("_stripe_cmd: unknown opcode passed");
    }
    LOG(D_FULL, ">> DPF_CMD: %s\n", str);

    buf[0] = DPF_CMD;
    memcpy(&buf[1], body, body_len);
    return (body_len + 1);
}

bool
DPF_channel::send_cmd(
        DPF_CMD_opcode opcode,
        struct sockaddr_in *addr1,
        struct sockaddr_in *addr2,
        char byte)
{
    char buf[DPF_MAX_CMD_SIZE];
    unsigned short size;

    size = stripe_cmd(&buf[4], opcode, addr1, addr2, byte);
    return cmd_buf->send(buf, size);
}

// Unstripe DPF command
// @args:
//  - buf: the buffer that contains a striped msg:
//         format: DPF_CMD[0] + enc{opcode + addr1 + addr2 + byte} 
//  - buf_len: the size of the striped msg in 'buf'
bool
DPF_channel::unstripe_cmd(char *buf, unsigned short buf_len, DPF_CMD_opcode *opcode,
        struct sockaddr_in *addr1, struct sockaddr_in *addr2, char *byte)
{
    if (buf[0] != DPF_CMD) {
        LOG(D_BASIC, "not a DPF command\n");
        return false;
    }
    char body[300];
    unsigned short body_len;
    memcpy(body, &buf[1], buf_len-1);
    body_len = buf_len-1;

    // unstripe the command 
    char str[300];
    *opcode = (DPF_CMD_opcode) body[0];
    sprintf(str, "%s ", _CMD_opcode_str[*opcode]);
    switch(*opcode) {
        case UDP_BIND:
        case TCP_BIND:
        case UDP_OPEN:
        case TCP_OPEN:
        case UDP_OPEN_DEL:
        case TCP_OPEN_DEL:
        case DUMMY:
            memcpy(&(addr1->sin_addr.s_addr), &body[1], 4);
            memcpy(&(addr1->sin_port), &body[5], 2);
            sprintf(&str[strlen(str)], "local = %s ", _sin2str(addr1));
            memcpy(&(addr2->sin_addr.s_addr), &body[7], 4);
            memcpy(&(addr2->sin_port), &body[11], 2);
            sprintf(&str[strlen(str)], "proxy = %s ", _sin2str(addr2));
            break;
        case LISTEN:
            memcpy(&(addr1->sin_addr.s_addr), &body[1], 4);
            memcpy(&(addr1->sin_port), &body[5], 2);
            sprintf(&str[strlen(str)], "local = %s ", _sin2str(addr1));
            if (body_len == 13) {
                memcpy(&(addr2->sin_addr.s_addr), &body[7], 4);
                memcpy(&(addr2->sin_port), &body[11], 2);
                sprintf(&str[strlen(str)], "proxy = %s ", _sin2str(addr2));
            } else {
                bzero(addr2, sizeof(struct sockaddr_in));
                sprintf(&str[strlen(str)], "proxy = no ");
            }
            break;
        case UDP_BIND_OK:
        case TCP_BIND_OK:
        case UDP_OPEN_OK:
        case TCP_OPEN_OK:
        case CONTACT:
        case RELAY_CONTACT:
        case TRIGGER:
        case RELAY_TRIGGER:
            memcpy(&(addr1->sin_addr.s_addr), &body[1], 4);
            memcpy(&(addr1->sin_port), &body[5], 2);
            sprintf(&str[strlen(str)], "addr = %s ", _sin2str(addr1));
            break;
        case LISTEN_OK:
            break;
        case UDP_CONNECT:
        case TCP_CONNECT:
            memcpy(&(addr1->sin_addr.s_addr), &body[1], 4);
            memcpy(&(addr1->sin_port), &body[5], 2);
            sprintf(&str[strlen(str)], "from = %s ", _sin2str(addr1));
            memcpy(&(addr2->sin_addr.s_addr), &body[7], 4);
            memcpy(&(addr2->sin_port), &body[11], 2);
            sprintf(&str[strlen(str)], "to = %s ", _sin2str(addr2));
            *byte = body[13];
            if (*byte == ALLOWED) {
                sprintf(&str[strlen(str)], "allowed");
            } else if (*byte == INSIDE) {
                sprintf(&str[strlen(str)], "inside");
            } else {
                sprintf(&str[strlen(str)], "not blocked");
            }
            break;
        case ACTIVE_INIT:
        case RELAY_INIT:
        case ACTIVE_CONNECT:
        case RELAY_CONNECT:
        case PASSIVE_INIT:
        case PASSIVE_CONNECT:
            memcpy(&(addr1->sin_addr.s_addr), &body[1], 4);
            memcpy(&(addr1->sin_port), &body[5], 2);
            sprintf(&str[strlen(str)], "%s ", _sin2str(addr1));
            break;
        case PASSIVE_ACK:
        case DUMMY_ACK:
            break;
        case NAK:
            *byte = body[1];
            sprintf(&str[strlen(str)], "err=%s ", strerror(*byte));
            break;
        default:
            ABORT("_unstripe_cmd: unknown opcode passed");
    }
    LOG(D_FULL, "<< DPF_CMD: %s\n", str);

    return true;
}

bool
DPF_channel::recv_cmd(
        DPF_CMD_opcode *opcode,
        struct sockaddr_in *addr1,
        struct sockaddr_in *addr2,
        char *byte)
{
    char buf[DPF_MAX_CMD_SIZE];
    unsigned short size;
    int result, rcved;

    rcved = DPF_MAX_CMD_SIZE;
    result = cmd_buf->recv(buf, &rcved);
    if (result < 0) {
        return false;
    } else if (result == 0) {
        errno = EINVAL;
        return false;
    }
    if (!unstripe_cmd(buf, rcved, opcode, addr1, addr2, byte)) {
        LOG(D_BASIC, "failed to unstripe cmd received\n");
        return false;
    }
    return true;
}

void
DPF_channel::tostr(const char *prefix, char *str)
{
    if (server) {
        sprintf(str, "%sfd=%d, server cmd_buf=", prefix, fd);
    } else {
        sprintf(str, "%sfd=%d, client cmd_buf=", prefix, fd);
    }
    if (cmd_buf) {
        cmd_buf->tostr("", &str[strlen(str)]);
    } else {
        strcat(str,"NULL");
    }
    sprintf(&str[strlen(str)], " status=%s", _DPF_hsk_stat_str[status]);
}

int
DPF_channel::stripe(char *buf)
{
    // length of the serialization excluding this field(2 bytes)
    // fd in network byte order(4 bytes)
    // server or client(1 byte): 1 = server; 0 = client;
    // DPF_RecBuf *cmd_buf
        // cmd_buf serialization length. 0 means that cmd_buf = NULL(2 bytes)
        // cmd_buf serialization(length bytes)
    // DPF_channel_status status(1 byte)
    // time_t entered in network byte order(4 bytes)

    char str[500];
    uint32_t nlong;
    uint16_t nshort;
    int index = 2;
    // fd in network byte order(4 bytes)
    nlong = htonl(fd);
    memcpy(&buf[index], &nlong, 4);
    sprintf(str, "[%d]fd=%d", index, fd);
    index += 4;
    // server or client(1 byte): 1 = server; 0 = client;
    if (server) {
        sprintf(str, "[%d]server", index, fd);
        buf[index++] = 1;
    } else {
        sprintf(str, "[%d]client", index, fd);
        buf[index++] = 0;
    }
    // DPF_RecBuf *cmd_buf
        // cmd_buf serialization length. 0 means that cmd_buf = NULL(2 bytes)
        // cmd_buf serialization(length bytes)
    sprintf(&str[strlen(str)], "[%d]cmd_buf_serialization", index);
    if (cmd_buf) {
        unsigned short cmd_buf_len = cmd_buf->stripe(&buf[index]);
        index += cmd_buf_len;
    } else {
        nshort = htons(0);
        memcpy(&buf[index], &nshort, 2);
        index += 2;
    }
    // DPF_channel_status status(1 byte)
    sprintf(&str[strlen(str)], "[%d]status=%s", index, _DPF_hsk_stat_str[status]);
    buf[index++] = status;
    // time_t entered in network byte order(4 bytes)
    nlong = htonl(entered);
    memcpy(&buf[index], &nlong, 4);
    sprintf(&str[strlen(str)], "[%d]entered", index);
    index += 4;
    // length of the serialization excluding this field(2 bytes)
    nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of DPF_channel: [0]len=%d%s\n", index-2, str);

    return index;
}

int
DPF_channel::unstripe(char *buf)
{
    // length of the serialization excluding this field(2 bytes)
    // fd in network byte order(4 bytes)
    // server or client(1 byte): 1 = server; 0 = client;
    // DPF_RecBuf *cmd_buf
        // cmd_buf serialization length. 0 means that cmd_buf = NULL(2 bytes)
        // cmd_buf serialization(length bytes)
    // DPF_channel_status status(1 byte)
    // time_t entered in network byte order(4 bytes)

    char str[500];
    uint32_t nlong;
    uint16_t nshort;
    int index = 0;
    // length of the serialization excluding this field(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // fd in network byte order(4 bytes)
    memcpy(&nlong, &buf[index], 4);
    fd = ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]fd=%d", index, fd);
    index += 4;
    // server or client(1 byte): 1 = server; 0 = client;
    if (buf[index]) {
        sprintf(str, "[%d]server", index, fd);
        server = true;
    } else {
        sprintf(str, "[%d]client", index, fd);
        server = false;
    }
    index++;
    // DPF_RecBuf *cmd_buf
        // cmd_buf serialization length. 0 means that cmd_buf = NULL(2 bytes)
        // cmd_buf serialization(length bytes)
    memcpy(&nshort, &buf[index], 2);
    sprintf(&str[strlen(str)], "[%d]cmd_buf_serialization", index);
    unsigned short cmd_buf_len = ntohs(nshort);
    if (cmd_buf_len > 0) {
        if (!cmd_buf) {
            cmd_buf = new DPF_RecBuf(fd, DPF_MAX_CMD_SIZE, DPF_TAG, DPF_VERSION);
        }
        index += cmd_buf->unstripe(&buf[index]);
    } else {
        index += 2;
    }
    // DPF_channel_status status(1 byte)
    status = (DPF_channel_status)buf[index];
    sprintf(&str[strlen(str)], "[%d]status=%s", index, _DPF_hsk_stat_str[status]);
    index++;
    // time_t entered in network byte order(4 bytes)
    memcpy(&nlong, &buf[index], 4);
    entered = ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]entered", index);
    index += 4;

    LOG(D_FULL, "Unserialization of DPF_channel: %s\n", str);

    if (length != index-2) {
        ABORT("length(%d) does not match with the actual size(index=%d) of the serialization", length, index);
    }

    return index;
}
