#ifndef _DPF_COMMON
#define _DPF_COMMON

#define DPF_TAG 0xAA
#define DPF_VERSION 0x10

#define DPF_OPEN_MAX 1024

#define DPF_MAX_CMD_SIZE 500 // Maximum size of DPF command
#define DPF_MAX_UDP_PACKET 65530 // Maximum size of a UDP packet
#define DPF_HEARTBEAT_RATE 60
#define DPF_FLUSH_RATE 120 // How often UDP_socket->peerCache flushed
#define INAGENT_HEARTBEAT_RATE 60
#define DPF_MAX_STAY 600 // How long can a socket stay in a stage
                         // where it can't stay infinitely
#define DPF_HASH_SIZE 173

/* fd type */
typedef enum _DPF_fd_type {
    NOT_DEFINED,
    TCP_UNCLASSIFIED, TCP_SERVER, TCP_CLIENT,
    UDP,
    TCP_IN_MNGER, TCP_OUT_MNGER, TCP_PEER_IN_MNGER,
    UDP_IN_MNGER, UDP_OUT_MNGER, UDP_PEER_IN_MNGER,
    HIDDEN, ACCEPTED,
    LISTENER, CONNECTOR, TCP_RELAY, UDP_RELAY,
    HEARTBEAT_FD,
    RESERVED
} DPF_fd_type;

typedef enum _DPF_msg_type {
    DPF_CMD,
    CLIENT_HELLO, SERVER_HELLO,
    CLIENT_CERTIFICATE, SERVER_CERTIFICATE,
    KEY_EXCHANGE, KEY_ACK,
    CLIENT_FINISHED, SERVER_FINISHED
} DPF_msg_type;

typedef enum _DPF_CMD_opcode {
    TCP_BIND, UDP_BIND,
    TCP_OPEN, UDP_OPEN,
    TCP_OPEN_DEL, UDP_OPEN_DEL,
    LISTEN,
    TCP_BIND_OK, UDP_BIND_OK,
    TCP_OPEN_OK, UDP_OPEN_OK,
    CONTACT, TRIGGER, RELAY_CONTACT, RELAY_TRIGGER,
    LISTEN_OK,
    TCP_CONNECT, UDP_CONNECT,
    ACTIVE_CONNECT, ACTIVE_INIT, RELAY_CONNECT, RELAY_INIT,
    PASSIVE_CONNECT, PASSIVE_INIT,
    PASSIVE_ACK,
    NAK,
    DUMMY, DUMMY_ACK,
    TCP_HEARTBEAT, UDP_HEARTBEAT
} DPF_CMD_opcode;
// Definition for 'byte' of TCP_CONNECT and UDP_CONNECT commands
#define ALLOWED 'A'
#define NOT_ALLOWED 'N'
#define INSIDE 'I'

/****************************************************************************/
/*                               ExAddr Class                               */
/****************************************************************************/
#include <netinet/in.h>
class ExAddr {
    private:
    public:
        struct sockaddr_in *local;
        struct sockaddr_in *proxy;
        struct sockaddr_in *agent;
        ExAddr(struct sockaddr_in *local,
               struct sockaddr_in *proxy,
               struct sockaddr_in *agent);
        ~ExAddr();
        void set_local(struct sockaddr_in *addr);
        void set_proxy(struct sockaddr_in *addr);
        void set_agent(struct sockaddr_in *addr);
        bool same_proxy(ExAddr *addr);
        bool same_proxy(struct sockaddr_in *sin);
        // connect socket fd to 'local'
        int connect(int fd);
        // connect socket fd to 'agent'
        int connect_agent(int fd);
        void tostr(const char *prefix, char *str);
        int stripe(char *outbuf);
        int unstripe(char *inbuf);
};

/****************************************************************************/
/*                            DPF_RecBuf Class                             */
/****************************************************************************/
// This class provides a mechanism for DPF record-based communications.
// DPF command and secure data exchange require this record based communication.
//
// This class uses a single buffer both for sending and receiving. Therefore, this
// class must be used only for stop-and-wait fashion communications. In other words,
// application should not try to send or receive the next record until the previous
// record is fully flushed into the kernel buffer or fully consumed by the
// application. This single record buffering makes the implementation of state transition
// machine easy because whether the buffer has a pending record can be used to designate
// a state.
//
// A record is of the format: DPF_TAG(1) + Version(1) + the length of the record(2) +
// data. send method prefixes record header (TAG, Version, Length of the record) before
// sending the record. At the receiving side, the header is removed and only the data
// part of the record is returned to the application
class DPF_RecBuf {
    private:
        static const short header_size = 4;
        static const char EMPTY = 'e';
        static const char SENDING = 's';
        static const char RECEIVING = 'r';

        unsigned char TAG;
        unsigned char VER;
        int fd; // TCP connection through which records are exchanged
        char status; // EMPTY when 'record' has nothing. SENDING when it contains
                     // data for sending. RECEIVING if partially received data is
                     // buffered
        unsigned char *record;
        int maxsize; // size of record[]
        // The following two fields represent either:
        //   -  For sending purpose
        //      . size: the size of the leftover of the record being sent
        //      . bytes: bytes sent so far
        //   -  For receiving purpose
        //      . size: the size of the record being received
        //      . bytes: bytes received so far
        int size;
        int bytes;
        bool header_checked; // if the header field has been checked

    public:
        DPF_RecBuf(int sockfd, int max, unsigned char tag, unsigned char ver);
        ~DPF_RecBuf();
        // Check if the next msg can be sent
        inline bool can_send() { return (status == EMPTY); };
        // Push the content of record[byte ~ size] to the network. Caller must
        // make sure that the buffer is not empty by calling can_send().
        // @ret:
        //    - true if the buffer is totally flushed
        //    - false otherwise w/ errno set to:
        //      + partially flushed: EAGAIN
        //      + send fails: errno returned by send
        bool flush();
        // Pack 'data' of 'len' bytes into a record and sends the record.
        // This method pushs the record to the network as much as possible by a single
        // send, and then put leftover into the buffer--'record'
        // This routine overwrites previous msg, so, caller must make sure that
        // the previous msg was fully flushed into the TCP buffer by calling
        // can_send()
        // @args:
        //   - data: the char array holding the payload to send. To avoid memory
        //           copy, this routine puts header (TAG, version, length) into
        //           data[0 - 3]. So application must put the payload from data[4]
        //   - len: the size of the payload. Should not be greater than the
        //          size of 'record[]'
        // @ret:
        //    - true if the record is totally sent
        //    - false w/ errno set to:
        //      + partially sent and the leftover is buffered: EAGAIN
        //      + send fails: errno returned by send
        //      + len > maxsize: EINVAL
        bool send(char *data, int len);
        // Receive the next record from the network. Unlike send, this method returns
        // the payload by copying it to the buffer passed from the application.
        // @args:
        //   - data: the payload will be returned via this arg. Should be big enough
        //           to hold that record
        //   - len: the size of the payload copied to 'data'
        // @ret:
        //   - 1 if the whole record is read and returned to 'data'
        //   - 0 if the connection closed
        //   - -1 otherwise w/ the errno set to:
        //     + EAGAIN: if recv returns EAGAIN or only partial record has been
        //               received
        //     + other errno: recv failed
        int recv(char *data, int *len);
        void tostr(const char *prefix, char *str);
        int stripe(char *outbuf);
        int unstripe(char *inbuf);
};

/****************************************************************************/
/*                           DPF_channel Class                             */
/****************************************************************************/
typedef enum _DPF_channel_status {
    InitSec,
    Connecting_1, Connecting_2,
    ClientHello, ServerHello,
    ClientCertificate, ServerCertificate,
    ComputeMasterSecret, KeyACK, ComputeKeyBlock,
    ClientFinished, ServerFinished,
    Connected
} DPF_channel_status;
typedef enum _channel_result {
    SUCC, BLOCKED, FAILED
} DPF_channel_result;
class DPF_channel {
    private:
        int fd; // socket fd
        bool server; // server or client
        DPF_RecBuf *cmd_buf;
        DPF_channel_status status;
        time_t entered; // time when the socket entered this status
        // methods
        DPF_channel_result TCP_handshake_client(
                struct sockaddr *addr,
                socklen_t addrlen,
                time_t current);
        void add_fd_TCP_CLIENT(int *maxfd, fd_set *rdfds, fd_set *wrfds);
        unsigned short stripe_cmd(char *buf, DPF_CMD_opcode opcode,
                struct sockaddr_in *addr1, struct sockaddr_in *addr2,
                char byte);
        bool unstripe_cmd(char *buf, unsigned short buf_len, DPF_CMD_opcode *opcode,
                struct sockaddr_in *addr1, struct sockaddr_in *addr2,
                char *byte);
    public:
        DPF_channel(); // client channel creation
        DPF_channel(int newfd); // server channel creation
        ~DPF_channel();
        bool getaddr(struct sockaddr_in *addr);
        bool bind(struct sockaddr_in *addr);
        // send DPF command
        bool send_cmd(DPF_CMD_opcode opcode, struct sockaddr_in *addr1,
                struct sockaddr_in *addr2, char byte);
        // recv DPF command
        bool recv_cmd(DPF_CMD_opcode *opcode, struct sockaddr_in *addr1,
                struct sockaddr_in *addr2, char *byte);
        void tostr(const char *prefix, char *str);
        int stripe(char *outbuf);
        int unstripe(char *inbuf);
        inline int get_fd() { return fd; };
        inline DPF_channel_result connect(
                struct sockaddr *addr,
                socklen_t addrlen) {
            return TCP_handshake_client(addr, addrlen, 0);
        };
        inline DPF_channel_result finish_connect(time_t current) {
            return TCP_handshake_client(NULL, 0, current);
        };
        inline void setfd(int *maxfd, fd_set *rdfds, fd_set *wrfds) {
            return add_fd_TCP_CLIENT(maxfd, rdfds, wrfds);
        };
        inline bool can_send() { return cmd_buf->can_send(); };
        inline bool flush() { return cmd_buf->flush(); };
};

/****************************************************************************/
/*                              utility routines                            */
/****************************************************************************/
#include "log.h"
#include "sock.h"
#include "internet.h"
#include "nb_wbuf.h"

/****************************************************************************/
/*                              common routines                             */
/****************************************************************************/
unsigned short _stripe_UDP_msg(char *buf,
        DPF_CMD_opcode opcode,
        struct sockaddr_in *local,
        struct sockaddr_in *proxy);
bool _unstripe_UDP_msg(char *buf, unsigned short buf_len,
        DPF_CMD_opcode *opcode,
        struct sockaddr_in *addr1,
        struct sockaddr_in *addr2,
        char *byte);

#endif
