#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#include "DPF_common.h"

DPF_RecBuf::DPF_RecBuf(int sockfd, int max,
        unsigned char tag, unsigned char ver)
{
    fd = sockfd;
    maxsize = max;
    record = (unsigned char *)malloc(maxsize);
    TAG = tag;
    VER = ver;
    bytes = size = 0;
    header_checked = false;
    status = EMPTY;
}

DPF_RecBuf::~DPF_RecBuf()
{
    if (record) {
        free(record);
    }
}

bool
DPF_RecBuf::flush()
{
    int sent;
    if (status != SENDING) {
        ABORT("flush called when status is not SENDING");
    }

    sent = ::send(fd, &record[bytes], size - bytes, 0);
    if (sent < 0) { // error
        if (errno != EAGAIN) {
            size = bytes = 0;
        }
        return false;
    } else if (sent < size - bytes) { // not totally flushed
        bytes += sent;
        errno = EAGAIN;
        LOG(D_FULL, "sent: %d, bytes: %d, size: %d\n", sent, bytes, size);
        return false;
    } else { // flushed
        bytes = size = 0;
        LOG(D_FULL, "sent: %d, bytes: %d, size: %d\n", sent, bytes, size);
        status = EMPTY;
        return true;
    }
}

bool
DPF_RecBuf::send(char *data, int len)
{
    if (status == RECEIVING) {
        ABORT("send called when status == RECEIVING");
    }

    // check if the record is too big
    if (len > maxsize - 4) {
        LOG(D_BASIC, "maxsize: %d, len: %d\n", maxsize, len);
        LOG(D_BASIC, "DPF_RecBuf::send - too big a record\n");
        errno = EINVAL;
        return false;
    }

    // make a record
    data[0] = TAG;
    data[1] = VER;
    size = len + 4;
    uint16_t nsize = htons(size);
    memcpy(&data[2], &nsize, 2);

    // send the record and copy leftover to the internal buffer
    bytes = ::send(fd, data, size, 0);
    if (bytes < 0) {
        if (errno != EAGAIN) { size = 0; }
        memcpy(record, data, size);
        bytes = 0;
        status = SENDING;
        LOG(D_FULL, "sent: %d, buffered: %d\n", bytes, size);
        return false;
    } else if (bytes < size) {
        errno = EAGAIN;
        LOG(D_FULL, "sent: %d, buffered: %d\n", bytes, size-bytes);
        memcpy(record, &data[bytes], size-bytes);
        bytes = 0;
        status = SENDING;
        return false;
    } else { // fully sent
        LOG(D_FULL, "sent: %d, buffered: 0\n", bytes);
        size = bytes = 0;
        status = EMPTY;
        return true;
    }
}

// receive the next record
// @args:
//   - data: should be big enough to hold the payload of the record
//   - len: the size of the payload being returned
// @ret:
//   - 1 if the whole record is read and returned to 'record'
//   - 0 if the connection closed
//   - -1 otherwise w/ the errno set to:
//     + EAGAIN: if recv returns EAGAIN or whole record is not ready yet
//     + other errno: recv failed
int
DPF_RecBuf::recv(char *data, int *len)
{
    if (status == SENDING) {
        ABORT("recv called when status == SENDING");
    }

    // if the buffer does not hold a full record, recv from the TCP connection upto
    // the size of the buffer
    if (!header_checked || bytes < size) {
        int rcved = ::recv(fd, &record[bytes], maxsize - bytes, 0);
        if (rcved < 0) {
            if (errno == EINTR) { errno = EAGAIN; }
            return -1;
        } else if (rcved == 0) {
            return 0;
        }
        bytes += rcved;
        LOG(D_FULL, "\t\tRBuf::recv: fd = %d: bytes = %d\n", fd, bytes);
    }

    // parse the header
    if (!header_checked && bytes >= 4) {
        if ((unsigned char)record[0] != TAG) {
            ABORT("Non DPF message received");
            LOG(D_FULL, "Non DPF message received. Through that away\n");
            bytes = 0;
            errno = EAGAIN;
            return -1;
        }
        if ((unsigned char)record[1] != VER) {
            ABORT("Invalid version number");
            LOG(D_FULL, "Invalid version number. Through that away\n");
            bytes = 0;
            errno = EAGAIN;
            return -1;
        }
        uint16_t nsize;
        memcpy(&nsize, &record[2], 2);
        size = ntohs(nsize);
        header_checked = true;
        if (size > maxsize) {
            ABORT("The record is too big\n");
            LOG(D_BASIC, "The record is too big\n");
            errno = EINVAL;
            return -1;
        }
        if (*len < size - 4) {
            LOG(D_BASIC, "size of the msg: %d, size of buffer passed by app: %d\n", size, *len);
            ABORT("not enough space passed for the record\n");
            LOG(D_BASIC, "not enough space passed for the record\n");
            errno = EINVAL;
            return -1;
        }
        LOG(D_FULL, "\t\tRBuf::recv - the size of the current record (%d)\n", size);
    }

    if (!header_checked || bytes < size) { // partial recv
        errno = EAGAIN;
        status = RECEIVING;
        return -1;
    }

    // full record has been received
    *len = size - 4;
    memcpy(data, &record[4], *len);
    // Note that this class must be used only for stop-and-wait fashion
    // communications, i.e., send a record, recv a record, send a record, and so on.
    // So we throw away any extra data beyond the current record boundary
    bytes = 0;
    size = 0;
    header_checked = false;
    status = EMPTY;
    return 1;
}

void
DPF_RecBuf::tostr(const char *prefix, char *str)
{
    if (header_checked) {
        sprintf(str, "%sfd = %d [size:%d bytes:%d] header checked",
                prefix, fd, size, bytes); 
    } else {
        sprintf(str, "%sfd = %d [size:%d bytes:%d] header not checked",
                prefix, fd, size, bytes); 
    }
}

int
DPF_RecBuf::stripe(char *buf)
{
    // the size of the serialization excluding this field(2 bytes)
    // fd: network byte order(4 bytes)
    // maxsize(2 bytes)
    // status(1 byte)
    // switch(status)
        // EMPTY: done
        // SENDING
            // size of data stored in the next field(2 bytes)
            // record[bytes ~ size-1](size bytes)
        // RECEIVING
            // bytes(2 bytes)
            // record[0 ~ bytes-1](bytes bytes)
            // header_checked: 0 or 1(1 byte)
            // size: present only if header_checked == true(2 bytes)

    char str[300];

    int index = 2;
    uint32_t nlong;
    uint16_t nshort;
    // fd: network byte order(4 bytes)
    nlong = htonl(fd);
    memcpy(&buf[index], &nlong, 4);
    sprintf(str, "[%d]fd=%d", index, fd);
    index += 4;
    // maxsize(2 bytes)
    nshort = htons(maxsize);
    memcpy(&buf[index], &nshort, 2);
    sprintf(&str[strlen(str)], "[%d]maxsize=%d", index, maxsize);
    index += 2;
    // status(1 byte)
    buf[index++] = status;
    unsigned short datalen;
    switch(status) {
        case EMPTY:
            sprintf(&str[strlen(str)], "[%d]status=EMPTY", index-1);
            break;
        case SENDING:
            // size of data stored in the next field(2 bytes)
            sprintf(&str[strlen(str)], "[%d]status=SENDING", index-1);
            datalen = size - bytes;
            nshort = htons(datalen);
            memcpy(&buf[index], &nshort, 2);
            sprintf(&str[strlen(str)], "[%d]data.len", index);
            index += 2;
            if (datalen > 0) {
                // record[bytes ~ size-1](datalen bytes)
                memcpy(&buf[index], &record[bytes], datalen);
                sprintf(&str[strlen(str)], "[%d]data", index);
                index += datalen;
            }
            break;
        case RECEIVING:
            // bytes(2 bytes)
            sprintf(&str[strlen(str)], "[%d]status=RECEIVING", index-1);
            nshort = htons(bytes);
            memcpy(&buf[index], &nshort, 2);
            sprintf(&str[strlen(str)], "[%d]bytes=%d", index, bytes);
            index += 2;
            if (bytes > 0) {
                // record[0 ~ bytes-1](bytes bytes)
                memcpy(&buf[index], record, bytes);
                sprintf(&str[strlen(str)], "[%d]data", index);
                index += bytes;
            }
            // header_checked: 0 or 1(1 byte)
            sprintf(&str[strlen(str)], "[%d]header_checked=%d", index, header_checked);
            buf[index++] = (header_checked) ? 1 : 0;
            if (header_checked) {
                // size(2 bytes)
                nshort = htons(size);
                memcpy(&buf[index], &nshort, 2);
                sprintf(&str[strlen(str)], "[%d]size=", index, size);
                index += 2;
            }
            break;
        default:
            ABORT("DPF_RecBuf is in invalid status");
    }

    // the size of the serialization excluding this field(2 bytes)
    nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of DPF_RecBuf: [0]len=%d%s\n", index-2, str);

    return index;
}

int
DPF_RecBuf::unstripe(char *buf)
{
    // the size of the serialization excluding this field(2 bytes)
    // fd: network byte order(4 bytes)
    // maxsize(2 bytes)
    // status(1 byte)
    // switch(status)
        // EMPTY: done
        // SENDING
            // size of data stored in the next field(2 bytes)
            // record[bytes ~ size-1](size bytes)
        // RECEIVING
            // bytes(2 bytes)
            // record[0 ~ bytes-1](bytes bytes)
            // header_checked: 0 or 1(1 byte)
            // size: present only if header_checked == true(2 bytes)

    char str[300];

    int index = 0;
    uint32_t nlong;
    uint16_t nshort;
    // the size of the serialization excluding this field(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // fd: network byte order(4 bytes)
    memcpy(&nlong, &buf[index], 4);
    fd = ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]fd=%d", index, fd);
    index += 4;
    // maxsize(2 bytes)
    unsigned short newsize;
    memcpy(&nshort, &buf[index], 2);
    newsize = ntohs(nshort);
    if (newsize != maxsize) {
        free(record);
        record = (unsigned char *)malloc(newsize);
        if (!record) { ABORT("malloc failed"); }
        maxsize = newsize;
    }
    sprintf(&str[strlen(str)], "[%d]maxsize=%d", index, maxsize);
    index += 2;
    // status(1 byte)
    status = buf[index++];
    unsigned short datalen;
    switch(status) {
        case EMPTY:
            sprintf(&str[strlen(str)], "[%d]status=EMPTY", index-1);
            break;
        case SENDING:
            // size of data stored in the next field(2 bytes)
            sprintf(&str[strlen(str)], "[%d]status=SENDING", index-1);
            memcpy(&nshort, &buf[index], 2);
            datalen = ntohs(nshort);
            sprintf(&str[strlen(str)], "[%d]data.len", index);
            index += 2;
            if (datalen > 0) {
                // record[0 ~ size-1](datalen bytes)
                memcpy(record, &buf[index], datalen);
                sprintf(&str[strlen(str)], "[%d]data", index);
                size = datalen;
                index += datalen;
            }
            break;
        case RECEIVING:
            // bytes(2 bytes)
            sprintf(&str[strlen(str)], "[%d]status=RECEIVING", index-1);
            memcpy(&nshort, &buf[index], 2);
            datalen = ntohs(nshort);
            sprintf(&str[strlen(str)], "[%d]bytes=%d", index, datalen);
            index += 2;
            if (datalen > 0) {
                // record[0 ~ bytes-1](bytes bytes)
                memcpy(record, &buf[index], datalen);
                sprintf(&str[strlen(str)], "[%d]data", index);
                bytes = datalen;
                index += datalen;
            }
            // header_checked: 0 or 1(1 byte)
            header_checked = (buf[index++] == 1) ? true : false;
            sprintf(&str[strlen(str)], "[%d]header_checked=%d", index-1, header_checked);
            if (header_checked) {
                // size(2 bytes)
                memcpy(&nshort, &buf[index], 2);
                size = ntohs(nshort);
                sprintf(&str[strlen(str)], "[%d]size=", index, size);
                index += 2;
            }
            break;
        default:
            ABORT("DPF_RecBuf is in invalid status");
    }

    LOG(D_FULL, "Unserialization of DPF_RecBuf: %s\n", str);

    if (length != index-2) {
        ABORT("length does not match with the actual size of the serialization");
    }

    return index;
}
