#include <time.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "inagent.h"

extern char *_server_status_str[];

void
InAgent::CONNECTOR_state_machine(Connector *connector, time_t current)
{
    DPF_CMD_opcode opcode;
    struct sockaddr_in addr1, addr2;
    char byte;
    int result;
    Listener *listener;

    int mfd = connector->conn->get_fd();
    time_t cur = time(NULL);

    while(true) {
        LOG(D_FULL, "[Connector fd=%d] %s (current=%lu)\n", mfd,
                _server_status_str[connector->status], current);
        switch(connector->status) {
            case CONTACT_DEST_AGENT_1:
                // Condition to enter here:
                //    - connector->conn->fd is read ready
                //    - timeout
                // Action:
                //    - recv TCP_CONNECT or UDP_CONNECT
                //    - find the listener
                //    - tie the connector and the listener
                if (current != 0) { // called by heartbeat
                    if (current - connector->entered > DPF_MAX_STAY) {
                        LOG(D_BASIC, "[fd=%d]timedout\n", mfd);
                        delete_connector(connector);
                    }
                    return;
                }
                result = connector->conn->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv failed: %s\n", strerror(errno));
                    delete_connector(connector);
                    return;
                }
                if (opcode == TCP_CONNECT) {
                    listener = find_listener(&addr2, TCP);
                } else if (opcode == UDP_CONNECT) {
                    listener = find_listener(&addr2, UDP);
                } else {
                    LOG(D_BASIC, "invalid msg received\n");
                    delete_connector(connector);
                    return;
                }
                if (listener == NULL) {
                    connector->error = ECONNREFUSED;
                    connector->status = CONNECTION_FAILED;
                    connector->entered = cur;
                    break;
                }
                connector->listener = listener;
                connector->status = CONTACT_DEST_AGENT_2;
                connector->entered = cur;
                break;
            case CONNECTION_FAILED:
                // Pre: CONNECT or UDP_CONNECT request could not be served.
                //      connector->error contains the error code of the reason
                // Condition to enter here:
                //    - from CONTACT_DEST_AGENT_1
                //    - connector->conn->fd is write ready
                //    - timeout
                // Action: send NAK(errcode)
                // Next: FINISHED
                if (current != 0) { // called by DPF_heartbeat
                    if (current - connector->entered > DPF_MAX_STAY) {
                        LOG(D_BASIC, "[fd=%d]timedout\n", mfd);
                        delete_connector(connector);
                    }
                    return;
                }
                if (!(connector->conn->can_send())) {
                    if (!(connector->conn->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush connector->conn: %s\n", strerror(errno));
                        delete_connector(connector);
                        return;
                    }
                } else {
                    result = connector->conn->send_cmd(NAK, NULL, NULL, connector->error);
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "send(NAK) failed: %s\n", strerror(errno));
                        delete_connector(connector);
                        return;
                    }
                }
                connector->status = FINISHED;
                break;
            case CONTACT_DEST_AGENT_2:
                // Pre: TCP_CONNECT(UDP_CONNECT) received
                // Condition to enter here:
                //    - from CONTACT_DEST_AGENT_1
                //    - connector->conn->fd is write ready
                //    - timeout
                // Action: send ACTIVE_CONNECT (ACTIVE_INIT)
                // Next: FINISHED
                if (current != 0) { // called by DPF_heartbeat
                    if (current - connector->entered > DPF_MAX_STAY) {
                        LOG(D_BASIC, "[fd=%d] timedout\n", mfd);
                        delete_connector(connector);
                    }
                    return;
                }
                if (!(connector->conn->can_send())) {
                    if (!(connector->conn->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        delete_connector(connector);
                        return;
                    }
                } else {
                    listener = connector->listener;
                    if (listener->proto == TCP) {
                        opcode = ACTIVE_CONNECT;
                    } else {
                        opcode = ACTIVE_INIT;
                    }
                    result = connector->conn->send_cmd(opcode, listener->myaddr->local, NULL, 0);
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "send(ACTIVE_CONN) failed: %s\n", strerror(errno));
                        delete_connector(connector);
                        return;
                    }
                }
                connector->status = FINISHED;
                break;
            case FINISHED:
                LOG(D_FULL, "the connector done\n");
                delete_connector(connector);
                return;
        }
    }
}
