#include <time.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include "inagent.h"

extern char *_server_status_str[];

void
InAgent::LISTENER_state_machine(Listener *listener, time_t current)
{
    DPF_CMD_opcode opcode;
    struct sockaddr_in addr1, addr2;
    char byte;
    int result;

    time_t cur = time(NULL);

    while(true) {
        LOG(D_FULL, "[Listener, fd=%d] %s (current=%lu)\n", listener->conn->get_fd(),
                _server_status_str[listener->status], current);
        switch(listener->status) {
            case REGISTERING_1:
                // Condition to enter here:
                //    - listener->conn->fd is read ready
                //    - timeout
                if (current != 0) {
                    if (current - listener->entered > DPF_MAX_STAY) {
                        delete_listener(listener);
                    }
                    return;
                }
                result = listener->conn->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv failed: %s\n", strerror(errno));
                    delete_listener(listener);
                    return;
                }
                if (opcode == TCP_BIND) {
                    listener->proto = TCP;
                    if (_firewall) {
                        if (!(firewall_open(listener, &addr1))) {
                            LOG(D_BASIC, "Can't open firewall for the listener\n");
                            listener->status = BIND_FAIL;
                            listener->entered = cur;
                            break;
                        }
                        listener->opened = true;
                        listener->myaddr = new ExAddr(&addr1, &addr1, NULL);
                    } else {
                        if (!(nat_bind(listener, &addr1, &addr2))) {
                            LOG(D_BASIC, "Can't give the address to the listener again\n");
                            listener->status = BIND_FAIL;
                            listener->entered = cur;
                            break;
                        }
                        listener->opened = true;
                        listener->myaddr = new ExAddr(&addr1, &addr2, NULL);
                    }
                    listener->status = REGISTERING_2;
                } else if (opcode == UDP_BIND) {
                    listener->proto = UDP;
                    if (_firewall) {
                        if (!(firewall_open(listener, &addr1))) {
                            LOG(D_BASIC, "Can't open firewall for the listener\n");
                            listener->status = BIND_FAIL;
                            listener->entered = cur;
                            break;
                        }
                        listener->opened = true;
                        listener->myaddr = new ExAddr(&addr1, &addr1, NULL);
                    } else {
                        if (!(nat_bind(listener, &addr1, &addr2))) {
                            LOG(D_BASIC, "Can't give the address to the listener again\n");
                            listener->status = BIND_FAIL;
                            listener->entered = cur;
                            break;
                        }
                        listener->opened = true;
                        listener->myaddr = new ExAddr(&addr1, &addr2, NULL);
                    }
                    listener->status = REGISTERING_2;
                } else {
                    LOG(D_BASIC, "unexpected msg received\n");
                    delete_listener(listener);
                    return;
                }
                listener->entered = cur;
                listener_2_hash(listener);
                break;
            case REGISTERING_2:
                // Condition to enter here:
                //    - from REGISTERING_1
                //    - listener->conn->fd is write ready
                //    - timeout
                if (current != 0) {
                    if (current - listener->entered > DPF_MAX_STAY) {
                        delete_listener(listener);
                    }
                    return;
                }
                if (!(listener->conn->can_send())) {
                    if (!(listener->conn->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        delete_listener(listener);
                        return;
                    }
                } else {
                    opcode = (listener->proto == TCP) ? TCP_BIND_OK : UDP_BIND_OK;
                    result = listener->conn->send_cmd(opcode, listener->myaddr->proxy, NULL, 0);
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        delete_listener(listener);
                        return;
                    }
                }
                listener->status = FULL_BOUND;
                if (listener->proto == TCP) {
                    listener->entered = cur;
                } else {
                    delete_connection(listener);
                }
                return;
            case BIND_FAIL:
            case LISTEN_FAIL:
                // Condition to enter here:
                //    - something went wrong
                //    - listener->conn->fd is write ready
                //    - timeout
                if (current != 0) {
                    if (current - listener->entered > DPF_MAX_STAY) {
                        delete_listener(listener);
                    }
                    return;
                }
                if (!(listener->conn->can_send())) {
                    if (!(listener->conn->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                    }
                } else {
                    result = listener->conn->send_cmd(NAK, NULL, NULL, listener->error);
                    if (!result && errno == EAGAIN) { return; }
                }
                delete_listener(listener);
                return;
            case FULL_BOUND:
                // Condition to enter here and action:
                //   + TCP:
                //     - listener->conn->fd read ready
                //     - timeout
                if (current != 0) {
                    if (current - listener->entered > DPF_MAX_STAY) {
                        delete_listener(listener);
                    }
                    return;
                }
                result = listener->conn->recv_cmd(&opcode, &addr1, &addr2, &byte);
                if (!result) {
                    if (errno == EAGAIN) { return; }
                    LOG(D_BASIC, "recv failed: %s\n", strerror(errno));
                    delete_listener(listener);
                    return;
                }
                if (opcode != LISTEN) {
                    LOG(D_BASIC, "strange msg received\n");
                    delete_listener(listener);
                    return;
                }
                listener->status = LISTENING_2;
                listener->entered = cur;
                break;
            case LISTENING_2:
                // Condition to enter here:
                //    - from FULL_BOUND
                //    - listener->conn->fd is write ready
                //    - timeout
                if (current != 0) {
                    if (current - listener->entered > DPF_MAX_STAY) {
                        delete_listener(listener);
                    }
                    return;
                }
                if (!(listener->conn->can_send())) {
                    if (!(listener->conn->flush())) {
                        if (errno == EAGAIN) { return; }
                        LOG(D_BASIC, "can't flush sock: %s\n", strerror(errno));
                        delete_listener(listener);
                        return;
                    }
                } else {
                    result = listener->conn->send_cmd(LISTEN_OK, NULL, NULL, 0);
                    if (!result) {
                        if (errno == EAGAIN) { return; }
                        delete_listener(listener);
                        return;
                    }
                }
                listener->listening = true;
                listener->status = FULL_LISTEN;
                delete_connection(listener);
                return;
            case FULL_LISTEN:
                return;
            default:
                ABORT("Listener falls in invalid state");
        }
    }
}
