#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "portfw.h"
extern "C" {
#include "fwkernel.h"
}
#include "fw_mnger.h"
#include "log.h"
#include "sock.h"
#include "internet.h"

#ifndef IPPROTO_NONE
#define IPPROTO_NONE    59
#endif
#define IP_PORTFW_DEF_PREF 10

/*
 *  *  Nice glibc2 kLuDge
 *   */
#if defined(__GLIBC__) || (__GLIBC__ > 2)
#define _LINUX_IN_H
#define _LINUX_IP_H
#define _LINUX_ICMP_H
#define _LINUX_UDP_H
#define _LINUX_TCP_H
#define _LINUX_BYTEORDER_GENERIC_H
#endif

FW_Mnger::FW_Mnger(char proto, char *logfile)
{
	for (int i=0; i<HASH_SIZE; i++) {
		_rules[i] = NULL;
	}

    if (proto == TCP) {
		_protocol = SOCK_STREAM;
	} else if (proto == UDP) {
		_protocol = SOCK_DGRAM;
	} else {
		ABORT ("FW_Mnger::FW_Mnger - invalid protocol");
	}

    strcpy(_logFileName, logfile);
}

// Add a rule to the internal data structure. A rule is internally represented by
// (private address, local address, remote address), where pviate address is ip:port
// of a socket behind firewall, local address is ip:port that a NAT binding will
// be created for the private address, and remote address is ip:port of a socket
// outside the private/firewalled network. For firewall management, we use private
// address as the source address and remote address as the destination address.
// @return: true, if success
//          false if error
// @error:
//    - EINVAL: the rule with the same value already exists
bool
FW_Mnger::addInternal(uint32_t src_ip, uint16_t src_port,
        uint32_t dst_ip, uint16_t dst_port)
{
    char str1[30], str2[30];
    sprintf(str1, "src=%s", _ipport2str(src_ip, src_port));
    sprintf(str2, "dest=%s", _ipport2str(dst_ip, dst_port));
    LOG(D_FULL, "[addInternal] %s %s\n", str1, str2);
    // Hash with private ip
    struct fwRule *rule = _rules[src_ip % HASH_SIZE];
	while (rule) {
		if (rule->priv_ip == src_ip && rule->priv_port == src_port &&
            rule->rmt_ip == dst_ip && rule->rmt_port == dst_port)
        { // found
            LOG(D_FULL, "addInternal: rule with the same value exist\n");
            errno = EINVAL;
            return false;
		} else {
			rule = rule->next;
		}
	}

    rule = (struct fwRule *)malloc(sizeof(struct fwRule));
    if (!rule) {
        ABORT("addInternal: malloc failed");
    }
	rule->priv_ip = src_ip;
	rule->priv_port = src_port;
	rule->rmt_ip = dst_ip;
	rule->rmt_port = dst_port;
    rule->pre = NULL;
	struct fwRule *rt = _rules[src_ip % HASH_SIZE];
	if (rt) {
		rt->pre = rule;
	}
    rule->next = rt;
	_rules[src_ip % HASH_SIZE] = rule;

	return true;
}

// Delete a rule from the internal representation. See the comments on addInternal
// for how rules are internally represented
// @return: true, if success
//          false if error
// @error:
//    - EINVAL: the rule with the given value does not exist
bool
FW_Mnger::deleteInternal(uint32_t src_ip, uint16_t src_port,
        uint32_t dst_ip, uint16_t dst_port)
{
    char str1[30], str2[30];
    sprintf(str1, "src=%s", _ipport2str(src_ip, src_port));
    sprintf(str2, "dst=%s", _ipport2str(dst_ip, dst_port));
    LOG(D_FULL, "[deleteInternal] %s %s\n", str1, str2);
	struct fwRule *rptr;
	rptr = _rules[src_ip % HASH_SIZE];
	while (rptr) {
		if (rptr->priv_ip == src_ip &&
            rptr->priv_port == src_port &&
            rptr->rmt_ip == dst_ip &&
            rptr->rmt_port == dst_port)
        {
            if (rptr->pre) {
                rptr->pre->next = rptr->next;
            } else {
                _rules[src_ip % HASH_SIZE] = rptr->next;
            }
            if (rptr->next) {
                rptr->next->pre = rptr->pre;
            }
            free(rptr);
            return true;
		}
		rptr = rptr->next;
	}

    LOG(D_FULL, "deleteInternal: rule not found\n");
    errno = EINVAL;
	return false;
}

// Delete rules remained in the log file and start with a fresh log file.
// This routine must be called at the startup time
void
FW_Mnger::cleanup()
{
    /* Recover the status where the FW_Mnger crashed */
    // open the log file
    int fd = ::open(_logFileName, O_RDWR | O_SYNC, S_IRWXU);
    if (fd < 0) {
        if (errno == ENOENT) {
            LOG(D_FULL, "FW_Mnger::cleanup - logfile does not exist\n");
            fd = ::open(_logFileName, O_RDWR | O_CREAT | O_SYNC, S_IRWXU);
            if (fd < 0) { ABORT("failed to create a fresh log file"); }
            _logFD = fdopen(fd, "w+");
            if (!_logFD) { ABORT("fdopen failed"); }
            return;
        } else {
            ABORT("failed to open the log file");
        }
    }
    _logFD = fdopen(fd, "r+");
    if (!_logFD) { ABORT("fdopen failed"); }
    // Since the rules that the FW_Mnger had at the time of crash are still remained
    // in the kernel, we just need to recreate internal representation
    char buf[100];
    while (fgets(buf, 100, _logFD) != NULL) {
        // read an entry from log file: ip and port are in network order
        char cmd[10], addr1[50], addr2[50];
        if (sscanf(buf, "%s %s %s", cmd, addr1, addr2) != 3) {
            ABORT("FW_Mnger::cleanup - sscanf() failed");
        }
        char target;
        struct sockaddr_in src, dst;
        // a filter rule is recorded as ADD/DEL src dst 
        if (_str2sin(addr1, &src) < 0) {
            ABORT("FW_Mnger::cleanup - src address is in a bad format");
        }
        if (_str2sin(addr2, &dst) < 0) {
            ABORT("FW_Mnger::cleanup - dst address is in a bad format");
        }
        if (!strncasecmp(cmd, "ADD", 3)) {
            (void) addInternal(src.sin_addr.s_addr, src.sin_port,
                               dst.sin_addr.s_addr, dst.sin_port);
        } else if (!strncasecmp(cmd, "DEL", 3)) {
            (void) deleteInternal(src.sin_addr.s_addr, src.sin_port,
                                  dst.sin_addr.s_addr, dst.sin_port);
        } else {
            ABORT("FW_Mnger::cleanup - strange command found in the log file");
        }
    }
    if ( !feof(_logFD) ) {
        ABORT ("FW_Mnger::sync - fgets(c_fp) failed: ");
    }
    fclose(_logFD);

    /* Delete rules in the internal representation from the kernel table */
    struct fwRule *rules = getRules();
    struct fwRule *rule = rules;
    while (rule) {
        struct fwRule *temp = rule;
        rule = rule->next;
        // delete the rule from the kernel
        (void) filterRuleSet(_protocol, CMD_DEL,
                             temp->priv_ip, temp->priv_port,
                             temp->rmt_ip, temp->rmt_port);
        free(temp);
    }

    /* flush the internal representation */
    flush_internal();

    /* Delete the log file */
    fd = ::open(_logFileName, O_RDWR | O_TRUNC | O_SYNC, S_IRWXU);
    if (fd < 0) { ABORT("failed to create a new log file"); }
    _logFD = fdopen(fd, "w+");
}

// Create a pinhole for the pair of ("src", "dst"). Either but not both
// can be NULL, meaning anybody.
bool
FW_Mnger::open(struct sockaddr_in *src, struct sockaddr_in *dst)
{
    if (!src && !dst) {
        LOG(D_BASIC, "At least one of src and dst must not be NULL\n");
        return false;
    }

    char str1[50];
    uint32_t src_ip;
    uint16_t src_port;
    if (src) {
        src_ip = src->sin_addr.s_addr;
        src_port = src->sin_port;
        sprintf(str1, "%s", _sin2str(src));
    } else {
        src_ip = htonl(INADDR_ANY);
        src_port = htons(0);
        sprintf(str1, "%s", _ipport2str(src_ip, src_port));
    }
    char str2[50];
    uint32_t dst_ip;
    uint16_t dst_port;
    if (dst) {
        dst_ip = dst->sin_addr.s_addr;
        dst_port = dst->sin_port;
        sprintf(str2, "%s", _sin2str(dst));
    } else {
        dst_ip = htonl(INADDR_ANY);
        dst_port = htons(0);
        sprintf(str2, "%s", _ipport2str(dst_ip, dst_port));
    }
    char buf[100];
    sprintf(buf, "ADD %s %s\n", str1, str2);
    LOG(D_FULL, "\t%s", buf);

    // update internal data structure
    if (!addInternal(src_ip, src_port, dst_ip, dst_port)) {
        LOG(D_BASIC, "addInternal failed\n");
        return false;
    }
    LOG(D_FULL, "\t\tInternal: done\n");

    // write the rule to the log file: we write the forwarding rule to the
    // persist file before seting up actual forwarding rule, which will
    // write the forwarding rule to kernel. At some bad situation, we will have some
    // garbage rules remained in our condor persistent file, but this causes no harm
    // and the garbage will be cleaned at the next startup time
    fputs(buf, _logFD);
    fflush(_logFD);
    LOG(D_FULL, "\t\tLog: done\n");

    // now add the rule to the kernel
    if (filterRuleSet(_protocol, CMD_ADD, src_ip, src_port, dst_ip, dst_port) < 0) {
        LOG(D_BASIC, "filterRuleSet failed\n");
        ABORT("\t%s\n", strerror(errno));
        return false;
    }
    LOG(D_FULL, "\t\tKernel: done\n");
    return true;
}

// Delete the pinhole for the pair of ("src", "dst"). Either but not both
// can be NULL, meaning anybody.
bool
FW_Mnger::close(struct sockaddr_in *src, struct sockaddr_in *dst)
{
    if (!src && !dst) {
        LOG(D_BASIC, "At least one of src and dst must not be NULL\n");
        return false;
    }

    char str1[50];
    uint32_t src_ip;
    uint16_t src_port;
    if (src) {
        src_ip = src->sin_addr.s_addr;
        src_port = src->sin_port;
        sprintf(str1, "%s", _sin2str(src));
    } else {
        src_ip = htonl(INADDR_ANY);
        src_port = htons(0);
        sprintf(str1, "%s", _ipport2str(src_ip, src_port));
    }
    char str2[50];
    uint32_t dst_ip;
    uint16_t dst_port;
    if (dst) {
        dst_ip = dst->sin_addr.s_addr;
        dst_port = dst->sin_port;
        sprintf(str2, "%s", _sin2str(dst));
    } else {
        dst_ip = htonl(INADDR_ANY);
        dst_port = htons(0);
        sprintf(str2, "%s", _ipport2str(dst_ip, dst_port));
    }
    char buf[100];
    sprintf(buf, "DEL %s %s\n", str1, str2);
    LOG(D_FULL, "\t%s", buf);

    // delete internal data structure first
    if (!deleteInternal(src_ip, src_port, dst_ip, dst_port)) {
        LOG(D_BASIC, "FW_Mnger::deleteRule - deleteInternal failed\n");
        return false;
    }
    LOG(D_FULL, "\tInternal done: %s", buf);

    // unset forwarding rule here
    if (filterRuleSet(_protocol, CMD_DEL, src_ip, src_port, dst_ip, dst_port) < 0) {
        LOG(D_BASIC, "FW_Mnger::deleteRule - failed to delete the rule to the kernel\n");
        ABORT("\t%s\n", strError());
    }
    LOG(D_FULL, "\tKernel: done\n");

    // update log
    fputs(buf, _logFD);
    fflush(_logFD);
    LOG(D_FULL, "\tLog: done\n");

    return true;
}

void
FW_Mnger::compress_log()
{
    // close the current log file
    fclose(_logFD);

    // open a temporary log file
    char tempname[30];
    sprintf(tempname, ".%s.tmp", _logFileName);
	FILE *temp_log = fopen(tempname, "w+");
	if (!temp_log) {
		ABORT("FW_Mnger::compress_log - faile to fopen(temp file)");
	}

    // dump the snapshot of internal representation to a temporary file
	struct fwRule *rules = getRules();
    struct fwRule *rule = rules;
    while (rule) {
        struct fwRule *temp = rule;
        rule = rule->next;
        char buf[100], str1[50], str2[50];
        sprintf(str1, "%s", _ipport2str(temp->priv_ip, temp->priv_port));
        sprintf(str2, "%s", _ipport2str(temp->rmt_ip, temp->rmt_port));
        sprintf(buf, "ADD %s %s\n", str1, str2);
        fputs(buf, temp_log); 
        free(temp);
    }
    fclose(temp_log);

    // replace the log file with the temporary log
    if (rename(tempname, _logFileName)) {
        ABORT("FW_Mnger::compress_log - rename failed: %s", strerror(errno));
    }

    // open the log file
    _logFD = fopen(_logFileName, "a+");
    if (!_logFD) {
        ABORT("FW_Mnger::compress_log - faile to fopen(compressed log file)");
    }
}

struct fwRule *
FW_Mnger::getRules()
{
	struct fwRule *rules = NULL; 
	for (int i=0; i<HASH_SIZE; i++) {
	    struct fwRule *rptr = _rules[i];
		while (rptr) {
			struct fwRule * temp = (struct fwRule *)malloc(sizeof(struct fwRule));
            //temp->target = rptr->target;
            temp->priv_ip = rptr->priv_ip;
            temp->priv_port = rptr->priv_port;
            //temp->local_ip = rptr->local_ip;
            //temp->local_port = rptr->local_port;
            temp->rmt_ip = rptr->rmt_ip;
            temp->rmt_port = rptr->rmt_port;
			rptr = rptr->next;
			temp->next = rules;
			rules = temp;
		}
	}
	return rules;
}

void
FW_Mnger::flush_internal()
{
    for (int i=0; i<HASH_SIZE; i++) {
        struct fwRule *rptr = _rules[i];
        while (rptr) {
            struct fwRule *temp = rptr;
            rptr = rptr->next;
            free(temp);
        }
        _rules[i] = NULL;
    }
}
