#ifndef __INAGENT
#define __INAGENT

#include "free_port_mnger.h"
#include "fw_mnger.h"
#include "nat_mnger.h"
#include "DPF_server.h"

#define INAGENT_FLUSH_FREQ 3 // unit = INAGENT_HEARTBEAT_RATE
#define INAGENT_FLUSH_RATE INAGENT_HEARTBEAT_RATE * INAGENT_FLUSH_FREQ
#define INAGENT_COMPRESS_FREQ (60 * 24) // unit = INAGENT_HEARTBEAT_RATE
#define INAGENT_COMPRESS_RATE INAGENT_HEARTBEAT_RATE * INAGENT_COMPRESS_FREQ

// connection method
#define ACTIVE 'a'
#define PASSIVE 'p'
#define RELAY 'r'
#define INVALID 'i'

class InAgent;
class Listener;
class Connector;

class Listener {
    friend class InAgent;
    friend class Connector;
    private:
        char proto; // TCP, UDP, or UNKNOWN
        DPF_channel *conn;
        DPF_server_status status;
        time_t entered;
        ExAddr *myaddr; // listener's local and proxy address
        bool opened; // Has NAT binding been created for this listener?
        bool listening; // Is the socket in listening status? Since TCP listener
                        // becomes listening through two DPF commands, we need
                        // this tag. We do not need this for UDP listeners
        char error;
        unsigned short missed; // how many time HEARTBEAT from this listener missed
        Listener *pre; // the listener is hashed by myaddr->proxy. These two links
        Listener *next;// are used to chain the listener to the hash bucket
    public:
        Listener(int fd);
        ~Listener();
        void tostr(const char *prefix, char *str);
};

class Connector {
    friend class InAgent;
    private:
        DPF_channel *conn;
        DPF_server_status status;
        time_t entered;
        Listener *listener; // the listener that this connector is trying to connect
                          // or has been connected to
        char error;

    public:
        Connector(int fd);
        ~Connector();
        void tostr(const char *prefix, char *str);
};

class InAgent {
	public:
        // Constructor for NAT manager
        InAgent(struct sockaddr_in *proxyaddr,
                struct sockaddr_in *servaddr,
                struct sockaddr_in *clientaddr,
                int tcp_serverfd,
                int udp_serverfd,
                int clientfd,
                int time_rfd);
        // Constructor for firewall manager
        InAgent(struct sockaddr_in *servaddr,
                int tcp_serverfd,
                int udp_serverfd,
                int time_rfd);
        ~InAgent();
		void run();

	private:
        bool _firewall; // true = inagent of firewall; false = inagent of NAT
        // if _firewall == true
        FW_Mnger *_TCP_FW_mnger;
        FW_Mnger *_UDP_FW_mnger;
        // else if _firewall == false
        FreePortMnger *_TCP_free_port;
        FreePortMnger *_UDP_free_port;
        NAT_Mnger *_TCP_NAT_mnger;
        NAT_Mnger *_UDP_NAT_mnger;
        struct sockaddr_in _proxyaddr;
        struct sockaddr_in _clientaddr;
        int _clientfd;
        // endif
        int _maxfd;
        int _tcp_servfd;
        int _udp_servfd;
        int _timeout;
        struct sockaddr_in _servaddr;
        fd_desc _client_fd[DPF_OPEN_MAX+1];
        Listener *_proxy_hash[DPF_HASH_SIZE];
        /* local methods */
        // Accept a new connection to _local and register a new Listener at
        // _client_fd[fd] as LISTENER. _maxfd is updated if necessary
        void add_listener();
        void add_listener(char proto, struct sockaddr_in *local, struct sockaddr_in *proxy);
        // Open a pinhole for the listener so that any client can talk to the
        // listener at 'dest'
        bool firewall_open(Listener *listener, struct sockaddr_in *dest);
        void firewall_close(Listener *listener);
        // Create a NAT binding for the listener. The NAT binding will be
        // {anybody->proxy} --> {anybody->local}. "proxy" is value-result argument
        bool nat_bind(Listener *listener, struct sockaddr_in *local, struct sockaddr_in *proxy);
        void nat_close(Listener *listener);
        // Add a listener to the head of _proxy_hash[]. Bucket is determined by
        // listener's proxy address. Note that this function does not check if the
        // listener is already chained to _proxy_hash[] or not.
        void listener_2_hash(Listener *listener);
        // Find the listener, from _proxy_hash[], with the proxy address--ip and
        // port--and proto same as 'addr' and 'proto', respectively. NULL is returned
        // if not found
        Listener *find_listener(struct sockaddr_in *addr, char proto);
        void handle_UDP_msg();
        void delete_connection(Listener *listener);
        // Delete a listener.
        //   - unchain the listener from _proxy_hash[], if necessary
        //   - if an address has been rented to the listener, then make it available
        //     for renting to another listener
        //   - unregister from _client_fd[], with _maxfd adjusted accordingly
        //   - close fd of the listener
        void delete_listener(Listener *listener);
        // Accept a new connection to _public and register a new Connector at
        // _client_fd[fd] as CONNECTOR. _maxfd is updated if necessary
        void add_connector();
        // Delete a connector.
        //   - if the firewall is opened for the connector, then close the pinhole
        //   - unregister from _client_fd[], with _maxfd adjusted accordingly
        //   - close fd of the connector
        void delete_connector(Connector *connector);
        // Set _maxfd appropriately
        void setmaxfd();
        // state transition machines
        void LISTENER_state_machine(Listener *listener, time_t current);
        void CONNECTOR_state_machine(Connector *connector, time_t current);
        // asynchronous event handling routines
        void set_fds(fd_set *rdfds, fd_set *wrfds);
        void handle_event(fd_set *rdfds, fd_set *wrfds);
        void heartbeat();
};

#endif
