#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>

#include "inagent.h"
#include "DPF_server_state_str.h"

int _time_wfd;

void
usage (char *msg)
{
    printf("%s\n\n", msg);

    printf("usage: inagent -S serv_addr [-P proxy_addr -C client_addr]\n");
    printf("\t\t-f debug_file [-g debug_level] [-x max_of_debug_file]\n");
    printf("\n");
    printf("serv_addr: the address to which server sockets must send DPF commands.\n");
    printf("           Must be ip:port. E.g., \"192.168.100.1:65432\"\n");
    printf("           One TCP and UDP sockets will be created at this address.\n");
    printf("proxy_addr: the address that NAT binding will be created at.\n");
    printf("            If the network behind is not private, then this\n");
    printf("            must not be specified. Must be one of the public\n");
    printf("            IP address this machine has. E.g., \"128.105.175.121\"\n");
    printf("client_addr: the address to which client sockets must send DPF commands.\n");
    printf("             If the network behind is not private, then this must not be\n");
    printf("             specified. Must be ip:port. E.g., \"192.168.100.1:65431\"\n");
    printf("             One TCP sockets will be created at this address.\n");
    printf("debug_file: the full path name of the file that log and debug messages go\n");
    printf("debug_level: 'fulldebug', 'network', or 'basic'. 'fulldebug' means extensive\n");
    printf("             logging and 'basic' means minimum logging. Default is 'basic'\n");
    printf("max_of_debug_file: maximum size of the debug file. When the file reaches\n");
    printf("                   the size, the log is moved to *.old file and new log file\n");
    printf("                   will be created. Default is 64000\n");
    printf("\n");
	return;
}

void
heartbeat(int signum)
{
    write(_time_wfd, "c", 1);
    alarm(INAGENT_HEARTBEAT_RATE);
}

int
main(int argc, char *argv[])
{
    bool proxy_given = false, serv_given = false, client_given = false;
    bool dfile_given = false;
    char dlevel[30], dfile[100];
    int debug_level = D_BASIC;
    long maxsize = 64000;
    struct in_addr inaddr;
    struct sockaddr_in proxyaddr, servaddr, clientaddr;
    /* get command line arguments */
    char c;
    while ( (c = getopt(argc, argv, "P:S:C:f:g:x:")) != -1 ) {
        switch (c) {
            case 'P': // proxy_addr
                if (proxy_given) {
                    usage("proxy_addr specified multiple times");
                    return -1;
                }
                if (inet_aton(optarg, &inaddr) < 0) {
                    usage("invalid proxy_addr given");
                    return -1;
                }
                proxyaddr.sin_family = AF_INET;
                proxyaddr.sin_addr.s_addr = inaddr.s_addr;
                proxyaddr.sin_port = htons(0);
                proxy_given = true;
                break;
            case 'S': // serv_addr
                if (serv_given) {
                    usage("serv_addr specified multiple times");
                    return -1;
                }
                if (_str2sin(optarg, &servaddr) < 0) {
                    usage("invalid serv_addr given");
                    return -1;
                }
                serv_given = true;
                break;
            case 'C': // client_addr
                if (client_given) {
                    usage("client_addr specified multiple times");
                    return -1;
                }
                if (_str2sin(optarg, &clientaddr) < 0) {
                    usage("invalid client_addr given");
                    return -1;
                }
                client_given = true;
                break;
            case 'f': // debug file
                if (dfile_given) {
                    usage("debug_file specified multiple times");
                    return -1;
                }
				strcpy(dfile, optarg);
                dfile_given = true;
                break;
            case 'g': // debug level
				strcpy(dlevel, optarg);
				if (!strcasecmp(dlevel, "fulldebug")) {
					debug_level = D_FULL;
				}
				break;
            case 'x': // maxsize of debug file
                maxsize = atol(optarg);
                break;
            default:
                usage("unknown option");
                return -1;
        }
    }

    if (!serv_given || !dfile_given) {
        usage("one or more mandatory fields are missing");
        return -1;
    }
    if (proxy_given && !client_given || !proxy_given && client_given) {
        usage("proxy_addr and client_addr must be used together");
        return -1;
    }

    Generic_logInit(dfile, debug_level, maxsize);

    // Create server TCP and UDP sockets
    int _tcp_serverfd = _tcp_nonblock_socket();
    int on = 1;
    if (setsockopt(_tcp_serverfd, SOL_SOCKET, SO_REUSEADDR , (int *)&on, sizeof(on))) {
        printf("setsockopt failed: %s", strerror(errno));
        return -1;
    }
    if (bind(_tcp_serverfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
        printf("failed to bind _tcp_serverfd: %s\n", strerror(errno));
        return -1;
    }
    if (listen(_tcp_serverfd, 5) < 0) {
        printf("failed to listen(_serverfd)\n");
        return -1;
    }
    int _udp_serverfd = _udp_nonblock_socket();
    on = 1;
    if (setsockopt(_udp_serverfd, SOL_SOCKET, SO_REUSEADDR , (int *)&on, sizeof(on))) {
        printf("setsockopt failed: %s", strerror(errno));
        return -1;
    }
    if (bind(_udp_serverfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
        printf("failed to bind _udp_serverfd: %s\n", strerror(errno));
        return -1;
    }
    // Create client socket
    int _clientfd;
    if (client_given) {
        _clientfd = _tcp_nonblock_socket();
        if (setsockopt(_clientfd, SOL_SOCKET, SO_REUSEADDR , (int *)&on, sizeof(on))) {
            printf("setsockopt failed: %s", strerror(errno));
            return -1;
        }
        if (bind(_clientfd, (struct sockaddr *)&clientaddr, sizeof(clientaddr)) < 0) {
            printf("failed to bind _clientfd: %s\n", strerror(errno));
            return -1;
        }
        if (listen(_clientfd, 5) < 0) {
            printf("failed to listen(_clientfd)\n");
            return -1;
        }
    }

    // create timeout pipe.
    int val;
    int pipe_fds[2];
    if (pipe(pipe_fds) < 0) {
        printf("pipe failed: %s\n", strerror(errno));
        return -1;
    }
    int _time_rfd = pipe_fds[0];
    if (_make_nonblock(_time_rfd) < 0) {
        LOG(D_BASIC, "can't make _time_rfd nonblock\n");
        close(_time_rfd);
    }
    _time_wfd = pipe_fds[1];
    if (_make_nonblock(_time_wfd) < 0) {
        LOG(D_BASIC, "can't make _time_wfd nonblock\n");
        close(_time_wfd);
    }

    // install sigpipe handler
    struct sigaction act;
    act.sa_handler = SIG_IGN;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if (sigaction(SIGPIPE, &act, NULL) < 0) {
        ABORT("Failed sigaction - %s\n", strerror(errno));
    }

    // install sigalarm handler
    act.sa_handler = heartbeat;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    if (sigaction(SIGALRM, &act, NULL) < 0) {
        ABORT("Failed sigaction - %s\n", strerror(errno));
    }

    // create an InAgent
    InAgent *inagent;
    if (proxy_given) {
        inagent = new InAgent(&proxyaddr, &servaddr, &clientaddr,
                _tcp_serverfd, _udp_serverfd, _clientfd, _time_rfd);
    } else {
        inagent = new InAgent(&servaddr, _tcp_serverfd, _udp_serverfd, _time_rfd);
    }

    alarm(INAGENT_HEARTBEAT_RATE);

    inagent->run();
    return -1;
}
