#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "internet.h"
#include "portfw.h"
#include "fwkernel.h"

int
snat_list(struct fwRule *rules)
{
    char str1[30], str2[30], str3[30];
    struct fwRule *rule;
    rule = rules;
    printf("\nSNAT rules:\n");
    while(rule) {
        sprintf(str1, "%s", _ipport2str(rule->priv_ip, rule->priv_port));
        sprintf(str2, "%s", _ipport2str(rule->local_ip, rule->local_port));
        sprintf(str3, "%s", _ipport2str(rule->rmt_ip, rule->rmt_port));
        printf("\t{%s->%s}-->{%s->%s}\n", str1, str3, str2, str3);
        rule = rule->next;
    }
    return 0;
}

int
dnat_list(struct fwRule *rules)
{
    char str1[30], str2[30], str3[30];
    struct fwRule *rule;
    rule = rules;
    printf("\nDNAT rules:\n");
    while(rule) {
        sprintf(str1, "%s", _ipport2str(rule->rmt_ip, rule->rmt_port));
        sprintf(str2, "%s", _ipport2str(rule->local_ip, rule->local_port));
        sprintf(str3, "%s", _ipport2str(rule->priv_ip, rule->priv_port));
        printf("\t{%s->%s}-->{%s->%s}\n", str1, str2, str1, str3);
        rule = rule->next;
    }
    return 0;
}

int
filter_list(struct fwRule *rules)
{
    char str1[30], str2[30], str3[30];
    struct fwRule *rule;
    rule = rules;
    printf("\nFilter rules:\n");
    while(rule) {
        sprintf(str1, "%s", _ipport2str(rule->priv_ip, rule->priv_port));
        sprintf(str2, "%s", _ipport2str(rule->rmt_ip, rule->rmt_port));
        printf("\t{%s}-->{%s}\n", str1, str2);
        rule = rule->next;
    }
    return 0;
}

static void
usage()
{
    printf("\nusage: one of the following:\n");
    printf("  - kRule.test [TCP|UDP] [add|del] dnat sip sport dip dport lip lport\n");
    printf("  - kRule.test [TCP|UDP] list dnat\n");
    printf("  - kRule.test [TCP|UDP] [add|del] snat sip sport dip dport lip lport\n");
    printf("  - kRule.test [TCP|UDP] list snat\n");
    printf("  - kRule.test [TCP|UDP] [add|del] filter sip sport dip dport\n");
    printf("  - kRule.test [TCP|UDP] list filter\n");
}

// 1. dnat rule is denoted as "{sip:sport->lip:lport}-->{sip:sport -> dip:dport}",
// meaning that packets from sip:sport to lip:lport are rewritten as if from
// sip:sport to dip:dport
// 2. snat rule is denoted as "{sip:sport->dip:dport}-->{lip:lport->dip:dport}",
// meaning that packets fromsip:sport to dip:dport are rewritten as if from
// lip:lport to dip:dport
//
// usage: one of the following
//  - kRule.test TCP/UDP add/del dnat sip sport lip lport dip dport
//  - kRule.test TCP/UDP list dnat
//  - kRule.test TCP/UDP add/del snat sip sport lip lport dip dport
//  - kRule.test TCP/UDP list snat
int
main(int argc, char *argv[])
{
    int proto;
    int cmd;
    struct in_addr inaddr;
    uint32_t sip, lip, dip;
    uint16_t sport, lport, dport;
    struct fwRule *rules;

    if (argc != 4 && argc != 8 && argc != 10) {
        printf("invalid number of arguments\n");
        usage();
        return -1;
    }

    // protocol
    if (!strcmp(argv[1], "TCP")) {
        proto = SOCK_STREAM;
    } else {
        proto = SOCK_DGRAM;
    }
    // command
    if (!strcmp(argv[2], "add")) {
        cmd = CMD_ADD;
    } else if (!strcmp(argv[2], "del")) {
        cmd = CMD_DEL;
    } else if (!strcmp(argv[2], "list")) {
        if (!strcmp(argv[3], "snat")) {
            if (snatRuleList(proto, &rules ) < 0) {
                printf("snatRuleList failed\n");
                return -1;
            }
            return snat_list(rules);
        } else if (!strcmp(argv[3], "dnat")) {
            if (dnatRuleList(proto, &rules ) < 0) {
                printf("dnatRuleList failed\n");
                return -1;
            }
            return dnat_list(rules);
        } else if (!strcmp(argv[3], "filter")) {
            if (filterRuleList(proto, &rules ) < 0) {
                printf("dnatRuleList failed\n");
                return -1;
            }
            return filter_list(rules);
        } else {
            printf("invalid table\n");
            return -1;
        }
    } else {
        printf("invalid command\n");
        return -1;
    }

    // sip sport dip dport lip lport
    if (inet_aton(argv[4], &inaddr) < 0) {
        printf("invalid sip passed\n");
        return -1;
    }
    sip = inaddr.s_addr;
    sport = htons(atoi(argv[5]));
    if (inet_aton(argv[6], &inaddr) < 0) {
        printf("invalid lip passed\n");
        return -1;
    }
    dip = inaddr.s_addr;
    dport = htons(atoi(argv[7]));
    if (argc == 10) {
        if (inet_aton(argv[8], &inaddr) < 0) {
            printf("invalid dip passed\n");
            return -1;
        }
        lip = inaddr.s_addr;
        lport = htons(atoi(argv[9]));
    }

    if (!strcmp(argv[3], "snat")) {
        if (snatRuleSet(proto, cmd, sip, sport, lip, lport, dip, dport) < 0) {
            printf("snatRuleSet failed\n");
            return -1;
        }
    } else if (!strcmp(argv[3], "dnat")) { // dnat
        // src_ip src_port pub_ip pub_port priv_ip priv_port
        if (dnatRuleSet(proto, cmd, sip, sport, lip, lport, dip, dport) < 0) {
            printf("snatRuleSet failed\n");
            return -1;
        }
    } else {
        if (filterRuleSet(proto, cmd, sip, sport, dip, dport) < 0) {
            printf("filterRuleSet failed\n");
            return -1;
        }
    }
    return 0;
}
