#include <getopt.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <ctype.h>
#include <stdarg.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>
#include "iptables.h"
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>

#include "portfw.h"
#include "fwkernel.h"

char errMsg[100];

struct ipt_natinfo
{
	struct ipt_entry_target t;
	struct ip_nat_multi_range mr;
};
struct iptables_target snat = {
    NULL,
    "SNAT",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ip_nat_multi_range)),
    IPT_ALIGN(sizeof(struct ip_nat_multi_range)),
};
struct iptables_target dnat = {
    NULL,
    "DNAT",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ip_nat_multi_range)),
    IPT_ALIGN(sizeof(struct ip_nat_multi_range)),
};
struct iptables_target standard = {
    NULL,
    "standard",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(int)),
    IPT_ALIGN(sizeof(int)),
};
struct iptables_match tcp = {
    NULL,
    "tcp",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_tcp)),
    IPT_ALIGN(sizeof(struct ipt_tcp)),
};
struct iptables_match udp = {
    NULL,
    "udp",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_udp)),
    IPT_ALIGN(sizeof(struct ipt_udp)),
};

#define NAT_TABLE "nat"
#define FILTER_TABLE "filter"
#define PREROUTING "PREROUTING"
#define POSTROUTING "POSTROUTING"
#define FORWARD "FORWARD"
#define TCP_SRC_PORTS 0x01
#define UDP_SRC_PORTS 0x01
#define TCP_DST_PORTS 0x02
#define UDP_DST_PORTS 0x02


/* The following function is defined in "/usr/include/linux/netfilter_ipv4/ip_tables.h"
   in my system. So uncomment the function if your system does not have the header file

struct ipt_entry_target *
ipt_get_target(struct ipt_entry *e)
{
	return (void *)e + e->target_offset;
}
*/

static struct ipt_entry *
generate_entry(const struct ipt_entry *fw,
		struct iptables_match *match,
		struct ipt_entry_target *target)
{
	unsigned int size;
	struct ipt_entry *e;

	size = sizeof(struct ipt_entry);
	size += match->m->u.match_size;

	if ( (e = malloc(size + target->u.target_size)) == NULL ) {
		sprintf (errMsg, "malloc failed: %s", strerror(errno));
		return NULL;
	}
	*e = *fw;
	e->target_offset = size;
	e->next_offset = size + target->u.target_size;

	memcpy(e->elems, match->m, match->m->u.match_size);
	size = match->m->u.match_size;
	memcpy(e->elems + size, target, target->u.target_size);

	return e;
}

static unsigned char *
make_delete_mask(struct ipt_entry *fw, struct iptables_match *m, struct iptables_target *t)
{
	/* Establish mask for comparison */
	unsigned int size;
	unsigned char *mask, *mptr;

	size = sizeof(struct ipt_entry);
	size += IPT_ALIGN(sizeof(struct ipt_entry_match)) + m->size;

	mask = calloc(1, size
			+ IPT_ALIGN(sizeof(struct ipt_entry_target))
			+ t->size);
	if (mask == NULL) {
		sprintf(errMsg, "calloc failed: %s", strerror(errno));
		return NULL;
	}

	memset(mask, 0xFF, sizeof(struct ipt_entry));
	mptr = mask + sizeof(struct ipt_entry);

	memset(mptr, 0xFF,
			IPT_ALIGN(sizeof(struct ipt_entry_match))
			+ m->userspacesize);
		mptr += IPT_ALIGN(sizeof(struct ipt_entry_match)) + m->size;

	memset(mptr, 0xFF,
			IPT_ALIGN(sizeof(struct ipt_entry_target))
			+ t->userspacesize);

	return mask;
}

// Append or delete a DNAT rule. A rule is denoted by
// "{sip:sport->lip:lport}-->{sip:sport -> dip:dport}", meaning that packets from
// sip:sport to lip:lport are rewritten as if from sip:sport to dip:dport. lip:lport
// must be a valid public address in this machine and dip:dport be a valid address
// in a machine behind this machine
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - command: CMD_ADD or CMD_DEL
//  - sip, sport, lip, lport, dip, dport: in network byte order
int
dnatRuleSet(int protocol, int command,
        uint32_t sip, uint16_t sport,
        uint32_t lip, uint16_t lport,
        uint32_t dip, uint16_t dport)
{
	iptc_handle_t handle;
	struct iptables_target *target;
	struct iptables_match *match;
	struct ipt_entry fw, *e = NULL;
	struct ipt_natinfo info;
	struct ip_nat_range range;
	size_t size;
	struct ipt_tcp *tcpinfo;
	struct ipt_udp *udpinfo;
	unsigned char *mask;

	// read the snapshot of the table
	handle = iptc_init(NAT_TABLE);
	if (!handle) {
		sprintf(errMsg, "can't read the snapshot of iptables");
		return -1;
	}

    // set which parts of a packet should be examined
	memset(&fw, 0, sizeof(fw));
	fw.nfcache |= NFC_UNKNOWN;
	fw.nfcache |= NFC_IP_PROTO;
    // with DNAT, destination ip(=lip) must be given
	fw.nfcache |= NFC_IP_DST;
	fw.ip.dmsk.s_addr = 0xFFFFFFFF;
    if (dport != 0) {
	    fw.nfcache |= NFC_IP_DST_PT;
    }
    if (sip != htonl(INADDR_ANY)) {
        fw.nfcache |= NFC_IP_SRC;
	    fw.ip.smsk.s_addr = 0xFFFFFFFF;
    }
    if (sport != 0) {
        fw.nfcache |= NFC_IP_SRC_PT;
    }

	if(protocol == SOCK_STREAM) {
		fw.ip.proto = IPPROTO_TCP;
	} else if(protocol == SOCK_DGRAM) {
		fw.ip.proto = IPPROTO_UDP;
	} else {
		sprintf(errMsg, "invalid protocol passed");
		return -1;
	}

	/* setup target */
	target = &dnat;
	size = IPT_ALIGN(sizeof(struct ipt_entry_target)) + target->size;
	memset(&info, 0, sizeof(info));
	target->t = (void *)&info;
	target->t->u.target_size = size;
	strcpy(target->t->u.user.name, "DNAT");
	target->tflags = 1;
	// --to dip:dport
	memset(&range, 0, sizeof(range));
	range.flags |= IP_NAT_RANGE_MAP_IPS;
	range.max_ip = range.min_ip = dip;
	range.flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
	range.min.tcp.port = range.max.tcp.port = dport;
	size = IPT_ALIGN(sizeof(info) + info.mr.rangesize * sizeof(range));
	//info.t.u.target_size = size;
	info.mr.range[info.mr.rangesize] = range;
	info.mr.rangesize++;

	/* setup match */
	if(protocol == SOCK_STREAM) {
		match = &tcp;
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + match->size;
		if ( (match->m = calloc(1, size)) == NULL ) {
			sprintf (errMsg, "calloc failed: %s", strerror(errno));
			return -1;
		}
		match->m->u.match_size = size;
		strcpy(match->m->u.user.name, match->name);
		// You should be aware that enough space has been allocated above
		tcpinfo = (struct ipt_tcp *) match->m->data;
        if (sport == 0) {
            // src port = (0, 65535)
            tcpinfo->spts[0] = 0;
            tcpinfo->spts[1] = 0xFFFF;
        } else {
            tcpinfo->spts[0] = tcpinfo->spts[1] = ntohs(sport);
		    match->mflags |= TCP_SRC_PORTS;
        }
        if (lport == 0) {
            // dst port = (0, 65535)
            tcpinfo->dpts[0] = 0;
            tcpinfo->dpts[1] = 0xFFFF;
        } else {
		    // dst port = (lport, lport)
		    tcpinfo->dpts[0] = tcpinfo->dpts[1] = ntohs(lport);
		    match->mflags |= TCP_DST_PORTS;
        }
	} else if(protocol == SOCK_DGRAM) {
		match = &udp;
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + match->size;
		if ( (match->m = calloc(1, size)) == NULL ) {
			sprintf (errMsg, "calloc failed: %s", strerror(errno));
			return -1;
		}
		match->m->u.match_size = size;
		strcpy(match->m->u.user.name, match->name);
		// You should be aware that enough space has been allocated above
		udpinfo = (struct ipt_udp *) match->m->data;
        if (sport == 0) {
            // src port = (0, 65535)
            udpinfo->spts[0] = 0;
            udpinfo->spts[1] = 0xFFFF;
        } else {
		    // src port = (sport, sport)
		    udpinfo->spts[0] = udpinfo->spts[1] = ntohs(sport);
		    match->mflags |= UDP_SRC_PORTS;
        }
        if (lport == 0) {
            // dst port = (0, 65535)
            udpinfo->dpts[0] = 0;
            udpinfo->dpts[1] = 0xFFFF;
        } else {
		    // dst port = (lport, lport)
		    udpinfo->dpts[0] = udpinfo->dpts[1] = ntohs(lport);
		    match->mflags |= UDP_DST_PORTS;
        }
	} else {
        return -1;
    }

	e = generate_entry(&fw, match, target->t);
	if (e == NULL) {
        free(match->m);
		return -1;
	}

	switch(command) {
		case CMD_ADD:
			e->ip.src.s_addr = sip;
			e->ip.dst.s_addr = lip;
			if(!iptc_append_entry(PREROUTING, e, &handle)) {
				printf("iptc_append_entry failed\n");
                free(match->m);
                free(e);
				return -1;
			}
			break;
		case CMD_DEL:
			mask = make_delete_mask(e, match, target);
			if (mask == NULL) {
                free(match->m);
                free(e);
				return -1;
			}
			e->ip.src.s_addr = sip;
			e->ip.dst.s_addr = lip;
			if (!iptc_delete_entry(PREROUTING, e, mask, &handle)) {
				printf("iptc_delete_entry failed\n");
                free(match->m);
                free(e);
				return -1;
			}
			free(mask);
			break;
		default:
			sprintf(errMsg, "invalid command passed: %d\n", command);
            free(match->m);
            free(e);
			return -1;
	}

	if (!iptc_commit(&handle)) {
		printf("iptc_commit failed\n");
        free(match->m);
        free(e);
		return -1;
	}

	// free up memory
	free(match->m);
	free(e);

	return 0;
}

// List DNAT rules. A rule is denoted by
// "{sip:sport->lip:lport}-->{sip:sport -> dip:dport}", meaning that packets from
// sip:sport to lip:lport are rewritten as if from sip:sport to dip:dport. lip:lport
// must be a valid public address in this machine and dip:dport be a valid address
// in a machine behind this machine
int
dnatRuleList(int protocol, struct fwRule **rules)
{
	iptc_handle_t handle;
	char *cName;
	const struct ipt_entry *entry;
	const struct ipt_entry_target *t;
	struct ipt_natinfo *info;
	struct ipt_entry_match *m;
	struct ipt_tcp *tcp;
	struct ipt_udp *udp;
	int proto;
	struct ip_nat_range range;
    uint32_t sip, lip, dip;
	uint16_t sport, lport, dport;
	struct fwRule *tRule;

	if (protocol == SOCK_STREAM) {
		proto = IPPROTO_TCP;
	} else if (protocol == SOCK_DGRAM) {
		proto = IPPROTO_UDP;
	} else {
        return -1;
    }

	// read the snapshot of the table
	handle = iptc_init(NAT_TABLE);
	if (!handle) {
		sprintf(errMsg, "can't read the snapshot of iptables\n");
		return -1;
	}

	*rules = NULL;
	for (cName = iptc_first_chain(&handle); cName; cName = iptc_next_chain(&handle)) {

		if (strcmp(cName, PREROUTING) != 0)
			continue;

		for (entry = iptc_first_rule(cName, &handle); entry; entry = iptc_next_rule(entry, &handle)) {
			if (entry->ip.proto != proto) {
				continue;
			}

			// get the match of the rule: simple because we know that our rules have only one match
			m = (void *)entry + sizeof(struct ipt_entry);
			if (proto == IPPROTO_TCP) {
				tcp = (void *) m->data;
                if (tcp->spts[0] == 0 && tcp->spts[1] == 0xFFFF) {
                    sport = 0;
                } else if (tcp->spts[0] == tcp->spts[1]) {
                    sport = htons(tcp->spts[0]);
                } else { continue; }
                if (tcp->dpts[0] == 0 && tcp->dpts[1] == 0xFFFF) {
                    lport = 0;
                } else if (tcp->dpts[0] == tcp->dpts[1]) {
                    lport = htons(tcp->dpts[0]);
                } else { continue; }
			} else {
				udp = (void *) m->data;
                if (udp->spts[0] == 0 && udp->spts[1] == 0xFFFF) {
                    sport = 0;
                } else if (udp->spts[0] == udp->spts[1]) {
                    sport = htons(udp->spts[0]);
                } else { continue; }
                if (udp->dpts[0] == 0 && udp->dpts[1] == 0xFFFF) {
                    lport = 0;
                } else if (udp->dpts[0] == udp->dpts[1]) {
                    lport = htons(udp->dpts[0]);
                } else { continue; }
			}

			// get the target of the rule
			t = ipt_get_target((struct ipt_entry *)entry);
			info = (void *)t;
			range = info->mr.range[0];
			if (range.min_ip != range.max_ip || range.min.all != range.max.all) {
				continue;
			}
            sip = entry->ip.src.s_addr;
			lip = entry->ip.dst.s_addr;
            dip = range.min_ip;
            dport = range.min.all;

			tRule = malloc(sizeof(struct fwRule));
			if (tRule == NULL) {
				sprintf(errMsg, "malloc failed: %s", strerror(errno));
				return -1;
			}
            tRule->rmt_ip = sip;
            tRule->rmt_port = sport;
            tRule->local_ip = lip;
            tRule->local_port = lport;
            tRule->priv_ip = dip;
            tRule->priv_port = dport;
			tRule->next = *rules;
			*rules = tRule;
		}
	}
	if (!iptc_commit(&handle)) {
		printf("iptc_commit failed\n");
		return -1;
	}
	return 0;
}

// Append or delete a SNAT rule. A rule is denoted by
// "{sip:sport->dip:dport}-->{lip:lport->dip:dport}", meaning that packets from
// sip:sport to dip:dport are rewritten as if from lip:lport to dip:dport.
// lip:lport must be a valid public address in this machine and sip:sport be a valid
// address in a machine behind this machine
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - command: CMD_ADD or CMD_DEL
//  - sip, sport, lip, lport, dip, dport: in network byte order
int
snatRuleSet(int protocol, int command,
        uint32_t sip, uint16_t sport,
        uint32_t lip, uint16_t lport,
        uint32_t dip, uint16_t dport)
{
	iptc_handle_t handle;
	struct iptables_target *target;
	struct iptables_match *match;
	struct ipt_entry fw, *e = NULL;
	struct ipt_natinfo info;
	struct ip_nat_range range;
	size_t size;
	struct ipt_tcp *tcpinfo;
	struct ipt_udp *udpinfo;
	unsigned char *mask;

	// read the snapshot of the table
	handle = iptc_init(NAT_TABLE);
	if (!handle) {
		sprintf(errMsg, "can't read the snapshot of iptables");
		return -1;
	}

    // set which parts of a packet should be examined
    memset(&fw, 0, sizeof(fw));
    fw.nfcache |= NFC_UNKNOWN;
    fw.nfcache |= NFC_IP_PROTO;
    // with SNAT, source ip must be given
    fw.nfcache |= NFC_IP_SRC;
    fw.ip.smsk.s_addr = 0xFFFFFFFF;
    if (sport != 0) {
        fw.nfcache |= NFC_IP_SRC_PT;
    }
    if (dip != htonl(INADDR_ANY)) {
        fw.nfcache |= NFC_IP_DST;
        fw.ip.dmsk.s_addr = 0xFFFFFFFF;
    }
    if (dport != 0) {
        fw.nfcache |= NFC_IP_DST_PT;
    }

    if(protocol == SOCK_STREAM) {
        fw.ip.proto = IPPROTO_TCP;
    } else if(protocol == SOCK_DGRAM) {
        fw.ip.proto = IPPROTO_UDP;
    } else {
        sprintf(errMsg, "invalid protocol passed");
        return -1;
    }

	/* setup target */
	target = &snat;
	size = IPT_ALIGN(sizeof(struct ipt_entry_target)) + target->size;
	memset(&info, 0, sizeof(info));
	target->t = (void *)&info;
	target->t->u.target_size = size;
	strcpy(target->t->u.user.name, "SNAT");
	target->tflags = 1;
	// --to-source lip:lport
	memset(&range, 0, sizeof(range));
	range.flags |= IP_NAT_RANGE_MAP_IPS;
	range.max_ip = range.min_ip = lip;
    if (lport != 0) {
	    range.flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
	    range.min.tcp.port = range.max.tcp.port = lport;
    }
	size = IPT_ALIGN(sizeof(info) + info.mr.rangesize * sizeof(range));
	//info.t.u.target_size = size;
	info.mr.range[info.mr.rangesize] = range;
	info.mr.rangesize++;

	/* setup match */
	if(protocol == SOCK_STREAM) {
		match = &tcp;
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + match->size;
		if ( (match->m = calloc(1, size)) == NULL ) {
			sprintf (errMsg, "calloc failed: %s", strerror(errno));
			return -1;
		}
		match->m->u.match_size = size;
		strcpy(match->m->u.user.name, match->name);
		// You should be aware that enough space has been allocated above
		tcpinfo = (struct ipt_tcp *) match->m->data;
        if (sport != 0) {
            // src port = (sport, sport)
		    tcpinfo->spts[0] = tcpinfo->spts[1] = ntohs(sport);
		    match->mflags |= TCP_SRC_PORTS;
        } else {
            // src port = (0, 65535)
            tcpinfo->spts[0] = 0;
            tcpinfo->spts[1] = 0xFFFF;
        }
        if (dport != 0) {
            // dst port = (dport, dport)
		    tcpinfo->dpts[0] = tcpinfo->dpts[1] = ntohs(dport);
		    match->mflags |= TCP_DST_PORTS;
        } else {
            // dst port = (0, 65535)
            tcpinfo->dpts[0] = 0;
            tcpinfo->dpts[1] = 0xFFFF;
        }
	} else if(protocol == SOCK_DGRAM) {
		match = &udp;
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + match->size;
		if ( (match->m = calloc(1, size)) == NULL ) {
			sprintf (errMsg, "calloc failed: %s", strerror(errno));
			return -1;
		}
		match->m->u.match_size = size;
		strcpy(match->m->u.user.name, match->name);
		// You should be aware that enough space has been allocated above
		udpinfo = (struct ipt_udp *) match->m->data;
        if (sport != 0) {
            // src port = (sport, sport)
		    udpinfo->spts[0] = udpinfo->spts[1] = ntohs(sport);
		    match->mflags |= TCP_SRC_PORTS;
        } else {
            // src port = (0, 65535)
            udpinfo->spts[0] = 0;
            udpinfo->spts[1] = 0xFFFF;
        }
        if (dport != 0) {
            // dst port = (dport, dport)
		    udpinfo->dpts[0] = udpinfo->dpts[1] = ntohs(dport);
		    match->mflags |= TCP_DST_PORTS;
        } else {
            // dst port = (0, 65535)
            udpinfo->dpts[0] = 0;
            udpinfo->dpts[1] = 0xFFFF;
        }
	} else {
        return -1;
    }

	e = generate_entry(&fw, match, target->t);
	if (e == NULL) {
        free(match->m);
		return -1;
	}

	switch(command) {
		case CMD_ADD:
			e->ip.src.s_addr = sip;
			e->ip.dst.s_addr = dip;
			if(!iptc_append_entry(POSTROUTING, e, &handle)) {
				printf("iptc_append_entry failed\n");
                free(match->m);
                free(e);
				return -1;
			}
			break;
		case CMD_DEL:
			mask = make_delete_mask(e, match, target);
			if (mask == NULL) {
                free(match->m);
                free(e);
				return -1;
			}
			e->ip.src.s_addr = sip;
			e->ip.dst.s_addr = dip;
			if (!iptc_delete_entry(POSTROUTING, e, mask, &handle)) {
				printf("iptc_delete_entry failed\n");
                free(match->m);
                free(e);
                free(mask);
				return -1;
			}
			free(mask);
			break;
		default:
			sprintf(errMsg, "invalid command passed: %d\n", command);
            free(match->m);
            free(e);
			return -1;
	}

	if (!iptc_commit(&handle)) {
		printf("iptc_commit failed\n");
        free(match->m);
        free(e);
		return -1;
	}

	// free up memory
	free(match->m);
	free(e);

	return 0;
}

// List SNAT rules. A rule is denoted by
// "{sip:sport->dip:dport}-->{lip:lport->dip:dport}", meaning that packets from
// sip:sport to dip:dport are rewritten as if from lip:lport to dip:dport.
// lip:lport must be a valid public address in this machine and sip:sport be a valid
// address in a machine behind this machine
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - rules: the list is returned via this arg
int
snatRuleList(int protocol, struct fwRule **rules)
{
	iptc_handle_t handle;
	char *cName;
	const struct ipt_entry *entry;
	const struct ipt_entry_target *t;
	struct ipt_natinfo *info;
	struct ipt_entry_match *m;
	struct ipt_tcp *tcp;
	struct ipt_udp *udp;
	int proto;
	struct ip_nat_range range;
    uint32_t sip, lip, dip;
    uint16_t sport, lport, dport;
	struct fwRule *tRule;

	if (protocol == SOCK_STREAM) {
		proto = IPPROTO_TCP;
	} else if (protocol == SOCK_DGRAM) {
		proto = IPPROTO_UDP;
	} else {
        return -1;
    }

	// read the snapshot of the table
	handle = iptc_init(NAT_TABLE);
	if (!handle) {
		sprintf(errMsg, "can't read the snapshot of iptables\n");
		return -1;
	}

	*rules = NULL;
	for (cName = iptc_first_chain(&handle); cName; cName = iptc_next_chain(&handle)) {

		if (strcmp(cName, POSTROUTING) != 0)
			continue;

		for (entry = iptc_first_rule(cName, &handle); entry; entry = iptc_next_rule(entry, &handle)) {
			if (entry->ip.proto != proto) {
				continue;
			}

			// get the match of the rule: simple because we know that our rules have only one match
			m = (void *)entry + sizeof(struct ipt_entry);
			if (proto == IPPROTO_TCP) {
				tcp = (void *) m->data;
                if (tcp->dpts[0] == tcp->dpts[1]) {
                    dport = htons(tcp->dpts[0]);
                } else if (tcp->dpts[0] == 0 && tcp->dpts[1] == 0xFFFF) {
                    dport = 0;
                } else { continue; }
                if (tcp->spts[0] == tcp->spts[1]) {
                    sport = htons(tcp->spts[0]);
                } else if (tcp->spts[0] == 0 && tcp->spts[1] == 0xFFFF) {
                    sport = 0;
                } else { continue; }
			} else {
				udp = (void *) m->data;
                if (udp->dpts[0] == udp->dpts[1]) {
                    dport = htons(udp->dpts[0]);
                } else if (udp->dpts[0] == 0 && udp->dpts[1] == 0xFFFF) {
                    dport = 0;
                } else { continue; }
                if (udp->spts[0] == udp->spts[1]) {
                    sport = htons(udp->spts[0]);
                } else if (udp->spts[0] == 0 && udp->spts[1] == 0xFFFF) {
                    sport = 0;
                } else { continue; }
			}

			// get the target of the rule
			t = ipt_get_target((struct ipt_entry *)entry);
			info = (void *)t;
			range = info->mr.range[0];
			if (range.min_ip != range.max_ip || range.min.all != range.max.all) {
				continue;
			}

            sip = entry->ip.src.s_addr;
            lip = range.min_ip;
            lport = range.min.all;
            dip = entry->ip.dst.s_addr; 

			tRule = malloc(sizeof(struct fwRule));
			if (tRule == NULL) {
				sprintf(errMsg, "malloc failed: %s", strerror(errno));
				return -1;
			}
            tRule->priv_ip = sip;
            tRule->priv_port = sport;
            tRule->local_ip = lip;
            tRule->local_port = lport;
            tRule->rmt_ip = dip;
            tRule->rmt_port = dport;
			tRule->next = *rules;
			*rules = tRule;
		}
	}
	if (!iptc_commit(&handle)) {
		printf("iptc_commit failed\n");
		return -1;
	}
	return 0;
}

// Append or delete a filter rule. A rule is denoted by
// "{sip:sport}-->{dip:dport}", meaning that packets from
// sip:sport to dip:dport must be accepted or rejected
// @args:
//  - protocol: SOCK_STREAM or SOCK_DGRAM
//  - command: CMD_ADD or CMD_DEL
//  - sip, sport, dip, dport: in network byte order
int
filterRuleSet(int protocol, int command,
        uint32_t sip, uint16_t sport,
        uint32_t dip, uint16_t dport)
{
	iptc_handle_t handle;
	struct iptables_target *target;
    struct ipt_entry_target info;
	struct iptables_match *match;
	struct ipt_entry fw, *e = NULL;
	size_t size;
	struct ipt_tcp *tcpinfo;
	struct ipt_udp *udpinfo;
	unsigned char *mask;

	// read the snapshot of the table
	handle = iptc_init(FILTER_TABLE);
	if (!handle) {
		sprintf(errMsg, "can't read the snapshot of iptables");
		return -1;
	}

    // set which parts of a packet should be examined
	memset(&fw, 0, sizeof(fw));
	fw.nfcache |= NFC_UNKNOWN;
	fw.nfcache |= NFC_IP_PROTO;
    // source
    if (sip != htonl(INADDR_ANY)) {
        fw.nfcache |= NFC_IP_SRC;
	    fw.ip.smsk.s_addr = 0xFFFFFFFF;
    }
    // source port
    if (sport != 0) {
        fw.nfcache |= NFC_IP_SRC_PT;
    }
    // destination ip
    if (dip != htonl(INADDR_ANY)) {
	    fw.nfcache |= NFC_IP_DST;
	    fw.ip.dmsk.s_addr = 0xFFFFFFFF;
    }
    // destination port
    if (dport != 0) {
	    fw.nfcache |= NFC_IP_DST_PT;
    }

    // protocol
	if(protocol == SOCK_STREAM) {
		fw.ip.proto = IPPROTO_TCP;
	} else if(protocol == SOCK_DGRAM) {
		fw.ip.proto = IPPROTO_UDP;
	} else {
		sprintf(errMsg, "invalid protocol passed");
		return -1;
	}

	/* setup target */
	target = &standard;
	size = IPT_ALIGN(sizeof(struct ipt_entry_target)) + target->size;
    memset(&info, 0, sizeof(info));
    target->t = (void *)&info;
    target->t->u.target_size = size;
    strcpy(target->t->u.user.name, "ACCEPT");
	target->tflags = 1;

	/* setup match */
	if(protocol == SOCK_STREAM) {
		match = &tcp;
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + match->size;
		if ( (match->m = calloc(1, size)) == NULL ) {
			sprintf (errMsg, "calloc failed: %s", strerror(errno));
			return -1;
		}
		match->m->u.match_size = size;
		strcpy(match->m->u.user.name, match->name);
		// You should be aware that enough space has been allocated above
		tcpinfo = (struct ipt_tcp *) match->m->data;
        if (sport == 0) {
            // src port = (0, 65535)
            tcpinfo->spts[0] = 0;
            tcpinfo->spts[1] = 0xFFFF;
        } else {
            tcpinfo->spts[0] = tcpinfo->spts[1] = ntohs(sport);
		    match->mflags |= TCP_SRC_PORTS;
        }
        if (dport == 0) {
            // dst port = (0, 65535)
            tcpinfo->dpts[0] = 0;
            tcpinfo->dpts[1] = 0xFFFF;
        } else {
		    // dst port = (dport, dport)
		    tcpinfo->dpts[0] = tcpinfo->dpts[1] = ntohs(dport);
		    match->mflags |= TCP_DST_PORTS;
        }
	} else if(protocol == SOCK_DGRAM) {
		match = &udp;
		size = IPT_ALIGN(sizeof(struct ipt_entry_match)) + match->size;
		if ( (match->m = calloc(1, size)) == NULL ) {
			sprintf (errMsg, "calloc failed: %s", strerror(errno));
			return -1;
		}
		match->m->u.match_size = size;
		strcpy(match->m->u.user.name, match->name);
		// You should be aware that enough space has been allocated above
		udpinfo = (struct ipt_udp *) match->m->data;
        if (sport == 0) {
            // src port = (0, 65535)
            udpinfo->spts[0] = 0;
            udpinfo->spts[1] = 0xFFFF;
        } else {
		    // src port = (sport, sport)
		    udpinfo->spts[0] = udpinfo->spts[1] = ntohs(sport);
		    match->mflags |= UDP_SRC_PORTS;
        }
        if (dport == 0) {
            // dst port = (0, 65535)
            udpinfo->dpts[0] = 0;
            udpinfo->dpts[1] = 0xFFFF;
        } else {
		    // dst port = (dport, dport)
		    udpinfo->dpts[0] = udpinfo->dpts[1] = ntohs(dport);
		    match->mflags |= UDP_DST_PORTS;
        }
	} else {
        return -1;
    }

	e = generate_entry(&fw, match, target->t);
	if (e == NULL) {
        free(match->m);
		return -1;
	}

	switch(command) {
		case CMD_ADD:
			e->ip.src.s_addr = sip;
			e->ip.dst.s_addr = dip;
			if(!iptc_append_entry(FORWARD, e, &handle)) {
				printf("iptc_append_entry failed\n");
                free(match->m);
                free(e);
				return -1;
			}
			break;
		case CMD_DEL:
			mask = make_delete_mask(e, match, target);
			if (mask == NULL) {
                free(match->m);
                free(e);
				return -1;
			}
			e->ip.src.s_addr = sip;
			e->ip.dst.s_addr = dip;
			if (!iptc_delete_entry(FORWARD, e, mask, &handle)) {
				printf("iptc_delete_entry failed\n");
                free(match->m);
                free(e);
                free(mask);
				return -1;
			}
			free(mask);
			break;
		default:
			sprintf(errMsg, "invalid command passed: %d\n", command);
            free(match->m);
            free(e);
			return -1;
	}

	if (!iptc_commit(&handle)) {
		printf("iptc_commit failed\n");
        free(match->m);
        free(e);
		return -1;
	}

	// free up memory
	free(match->m);
	free(e);

	return 0;
}

// List filter rules. A rule is denoted by
// "{sip:sport}-->{dip:dport}", meaning that packets from
// sip:sport to dip:dport are accepted or rejected
int
filterRuleList(int protocol, struct fwRule **rules)
{
	iptc_handle_t handle;
	char *cName;
	const struct ipt_entry *entry;
	struct ipt_entry_match *m;
	struct ipt_tcp *tcp;
	struct ipt_udp *udp;
	int proto;
    uint32_t sip, dip;
	uint16_t sport, dport;
	struct fwRule *tRule;

	if (protocol == SOCK_STREAM) {
		proto = IPPROTO_TCP;
	} else if (protocol == SOCK_DGRAM) {
		proto = IPPROTO_UDP;
	} else {
        return -1;
    }

	// read the snapshot of the table
	handle = iptc_init(FILTER_TABLE);
	if (!handle) {
		sprintf(errMsg, "can't read the snapshot of iptables\n");
		return -1;
	}

	*rules = NULL;
	for (cName = iptc_first_chain(&handle); cName; cName = iptc_next_chain(&handle)) {

		if (strcmp(cName, FORWARD) != 0)
			continue;

		for (entry = iptc_first_rule(cName, &handle); entry; entry = iptc_next_rule(entry, &handle)) {
			if (entry->ip.proto != proto) {
				continue;
			}

			// get the match of the rule: simple because we know that our rules have only one match
			m = (void *)entry + sizeof(struct ipt_entry);
			if (proto == IPPROTO_TCP) {
				tcp = (void *) m->data;
                if (tcp->spts[0] == 0 && tcp->spts[1] == 0xFFFF) {
                    sport = 0;
                } else if (tcp->spts[0] == tcp->spts[1]) {
                    sport = htons(tcp->spts[0]);
                } else { continue; }
                if (tcp->dpts[0] == 0 && tcp->dpts[1] == 0xFFFF) {
                    dport = 0;
                } else if (tcp->dpts[0] == tcp->dpts[1]) {
                    dport = htons(tcp->dpts[0]);
                } else { continue; }
			} else {
				udp = (void *) m->data;
                if (udp->spts[0] == 0 && udp->spts[1] == 0xFFFF) {
                    sport = 0;
                } else if (udp->spts[0] == udp->spts[1]) {
                    sport = htons(udp->spts[0]);
                } else { continue; }
                if (udp->dpts[0] == 0 && udp->dpts[1] == 0xFFFF) {
                    dport = 0;
                } else if (udp->dpts[0] == udp->dpts[1]) {
                    dport = htons(udp->dpts[0]);
                } else { continue; }
			}

			// get sip and dip
            sip = entry->ip.src.s_addr;
			dip = entry->ip.dst.s_addr;

			tRule = malloc(sizeof(struct fwRule));
			if (tRule == NULL) {
				sprintf(errMsg, "malloc failed: %s", strerror(errno));
				return -1;
			}
            tRule->rmt_ip = dip;
            tRule->rmt_port = dport;
            tRule->local_ip = htonl(INADDR_ANY);
            tRule->local_port = htons(0);
            tRule->priv_ip = sip;
            tRule->priv_port = sport;
			tRule->next = *rules;
			*rules = tRule;
		}
	}
	if (!iptc_commit(&handle)) {
		printf("iptc_commit failed\n");
		return -1;
	}
	return 0;
}

char *strError(void)
{
	return errMsg;
}
