#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "portfw.h"
extern "C" {
#include "fwkernel.h"
}
#include "nat_mnger.h"
#include "log.h"
#include "sock.h"
#include "internet.h"

#ifndef IPPROTO_NONE
#define IPPROTO_NONE    59
#endif
#define IP_PORTFW_DEF_PREF 10

/*
 *  *  Nice glibc2 kLuDge
 *   */
#if defined(__GLIBC__) || (__GLIBC__ > 2)
#define _LINUX_IN_H
#define _LINUX_IP_H
#define _LINUX_ICMP_H
#define _LINUX_UDP_H
#define _LINUX_TCP_H
#define _LINUX_BYTEORDER_GENERIC_H
#endif

NAT_Mnger::NAT_Mnger(char proto, char *logfile)
{
	for (int i=0; i<HASH_SIZE; i++) {
		_rules[i] = NULL;
	}

    if (proto == TCP) {
		_protocol = SOCK_STREAM;
	} else if (proto == UDP) {
		_protocol = SOCK_DGRAM;
	} else {
		ABORT ("NAT_Mnger::NAT_Mnger - invalid protocol");
	}

    strcpy(_logFileName, logfile);
}

// Add a rule to the internal data structure. A rule is internally represented by
// (private address, local address, remote address), where pviate address is ip:port
// of a socket behind the NAT box and in a private network, local address is ip:port
// that a NAT binding will be created for the private address, and remote address is
// ip:port of a socket outside the private network. Depending on whether the rule is
// a DNAT or SNAT rule, each address means source or destination.
// @args:
//    - target: DNAT or SNAT
//    - priv_ip, priv_port: private address
//    - local_ip, local_port: local address
//    - rmt_ip, rmt_port: remote address
// @return: true, if success
//          false if error
// @error:
//    - EINVAL: the rule with the same value already exists
bool
NAT_Mnger::addInternal(char target,
        uint32_t priv_ip, uint16_t priv_port,
        uint32_t local_ip, uint16_t local_port,
        uint32_t rmt_ip, uint16_t rmt_port)
{
    char str1[30], str2[30], str3[30];
    sprintf(str1, "priv=%s", _ipport2str(priv_ip, priv_port));
    sprintf(str2, "local=%s", _ipport2str(local_ip, local_port));
    sprintf(str3, "remote=%s", _ipport2str(rmt_ip, rmt_port));
    LOG(D_FULL, "[addInternal] %s %s %s\n", str1, str2, str3);
    // Hash with private ip
    struct fwRule *rule = _rules[priv_ip % HASH_SIZE];
	while (rule) {
		if (rule->target == target &&
            rule->priv_ip == priv_ip && rule->priv_port == priv_port &&
            rule->local_ip == local_ip && rule->local_port == local_port &&
            rule->rmt_ip == rmt_ip && rule->rmt_port == rmt_port)
        { // found
            LOG(D_FULL, "addInternal: rule with the same value exist\n");
            errno = EINVAL;
            return false;
		} else {
			rule = rule->next;
		}
	}

    rule = (struct fwRule *)malloc(sizeof(struct fwRule));
    if (!rule) {
        ABORT("addInternal: malloc failed");
    }
    rule->target = target;
	rule->priv_ip = priv_ip;
	rule->priv_port = priv_port;
	rule->local_ip = local_ip;
	rule->local_port = local_port;
	rule->rmt_ip = rmt_ip;
	rule->rmt_port = rmt_port;
    rule->pre = NULL;
	struct fwRule *rt = _rules[priv_ip % HASH_SIZE];
	if (rt) {
		rt->pre = rule;
	}
    rule->next = rt;
	_rules[priv_ip % HASH_SIZE] = rule;

	return true;
}

// Delete a rule from the internal representation. See the comments on addInternal
// for how rules are internally represented
// @args
//    - target: DNAT or SNAT
//    - priv_ip, priv_port: private address
//    - local_ip, local_port: local address. These are value-result argument. If
//      local_ip:local_port is not 0.0.0.0:0, then the rule with the same (private,
//      local, remote) will be deleted. Otherwise the rule with the same (private,
//      remote) value will be deleted and the local address of the deleted rule will
//      be returned via these args
//    - rmt_ip, rmt_port: remote address
// @return: true, if success
//          false if error
// @error:
//    - EINVAL: the rule with the given value does not exist
bool
NAT_Mnger::deleteInternal(char target,
        uint32_t priv_ip, uint16_t priv_port,
        uint32_t *local_ip, uint16_t *local_port,
        uint32_t rmt_ip, uint16_t rmt_port)
{
    char str1[30], str2[30], str3[30];
    sprintf(str1, "priv=%s", _ipport2str(priv_ip, priv_port));
    sprintf(str2, "local=%s", _ipport2str(*local_ip, *local_port));
    sprintf(str3, "remote=%s", _ipport2str(rmt_ip, rmt_port));
    LOG(D_FULL, "[deleteInternal] %s %s %s\n", str1, str2, str3);
	struct fwRule *rptr;
	rptr = _rules[priv_ip % HASH_SIZE];
	while (rptr) {
		if (rptr->target == target &&
            rptr->priv_ip == priv_ip &&
            rptr->priv_port == priv_port &&
            rptr->rmt_ip == rmt_ip &&
            rptr->rmt_port == rmt_port)
        {
            if (*local_ip == htonl(INADDR_ANY) && *local_port == htons(0)) { 
                *local_ip = rptr->local_ip;
                *local_port = rptr->local_port;
            } else if (*local_ip != rptr->local_ip || *local_port != rptr->local_port) {
		        rptr = rptr->next;
                continue;
            }
            if (rptr->pre) {
                rptr->pre->next = rptr->next;
            } else {
                _rules[priv_ip % HASH_SIZE] = rptr->next;
            }
            if (rptr->next) {
                rptr->next->pre = rptr->pre;
            }
            free(rptr);
            return true;
		}
		rptr = rptr->next;
	}

    LOG(D_FULL, "deleteInternal: rule not found\n");
    errno = EINVAL;
	return false;
}

// Delete rules remained in the log file and start with a fresh log file.
// This routine must be called at the startup time
void
NAT_Mnger::cleanup()
{
    /* Recover the status where the NAT_Mnger crashed */
    // open the log file
    int fd = open(_logFileName, O_RDWR | O_SYNC, S_IRWXU);
    if (fd < 0) {
        if (errno == ENOENT) {
            LOG(D_FULL, "NAT_Mnger::cleanup - logfile does not exist\n");
            fd = open(_logFileName, O_RDWR | O_CREAT | O_SYNC, S_IRWXU);
            if (fd < 0) { ABORT("failed to create a fresh log file"); }
            _logFD = fdopen(fd, "w+");
            if (!_logFD) { ABORT("fdopen failed"); }
            return;
        } else {
            ABORT("failed to open the log file");
        }
    }
    _logFD = fdopen(fd, "r+");
    if (!_logFD) { ABORT("fdopen failed"); }
    // Since the rules that the NAT_Mnger had at the time of crash are still remained
    // in the kernel, we just need to recreate internal representation
    char buf[100];
    while (fgets(buf, 100, _logFD) != NULL) {
        // read an entry from _persistFile file: ip and port are in network order
        char target_str[10], cmd[10], addr1[50], addr2[50], addr3[50];
        if (sscanf(buf, "%s %s %s %s %s", target_str, cmd, addr1, addr2, addr3) != 5) {
            ABORT("NAT_Mnger::cleanup - sscanf() failed");
        }
        char target;
        struct sockaddr_in priv, local, remote;
        uint32_t priv_ip, local_ip, rmt_ip;
        uint16_t priv_port, local_port, rmt_port;
        if (!strncasecmp(target_str, "DNAT", 4)) {
            // dnat rule is recorded as DNAT ADD/DEL remote local private
            target = DNAT;
            if (_str2sin(addr1, &remote) < 0) {
                ABORT("NAT_Mnger::cleanup - remote address is in a bad format");
            }
            rmt_ip = remote.sin_addr.s_addr;
            rmt_port = remote.sin_port;
            if (_str2sin(addr2, &local) < 0) {
                ABORT("NAT_Mnger::cleanup - local address is in a bad format");
            }
            local_ip = local.sin_addr.s_addr;
            local_port = local.sin_port;
            if (_str2sin(addr3, &priv) < 0) {
                ABORT("NAT_Mnger::cleanup - private address is in a bad format");
            }
            priv_ip = priv.sin_addr.s_addr;
            priv_port = priv.sin_port;
        } else if (!strncasecmp(target_str, "SNAT", 4)) {
            // snat rule is recorded as SNAT ADD/DEL private local remote
            target = SNAT;
            if (_str2sin(addr1, &priv) < 0) {
                ABORT("NAT_Mnger::cleanup - private address is in a bad format");
            }
            priv_ip = priv.sin_addr.s_addr;
            priv_port = priv.sin_port;
            if (_str2sin(addr2, &local) < 0) {
                ABORT("NAT_Mnger::cleanup - local address is in a bad format");
            }
            local_ip = local.sin_addr.s_addr;
            local_port = local.sin_port;
            if (_str2sin(addr3, &remote) < 0) {
                ABORT("NAT_Mnger::cleanup - remote address is in a bad format");
            }
            rmt_ip = remote.sin_addr.s_addr;
            rmt_port = remote.sin_port;
        }
        if (!strncasecmp(cmd, "ADD", 3)) {
            (void) addInternal(target, priv_ip, priv_port, local_ip, local_port, rmt_ip, rmt_port);
        } else if (!strncasecmp(cmd, "DEL", 3)) {
            (void) deleteInternal(target, priv_ip, priv_port, &local_ip, &local_port, rmt_ip, rmt_port);
        } else {
            ABORT("NAT_Mnger::cleanup - strange command found in the log file");
        }
    }
    if ( !feof(_logFD) ) {
        ABORT ("NAT_Mnger::sync - fgets(c_fp) failed: ");
    }
    fclose(_logFD);

    /* Delete rules in the internal representation from the kernel table */
    struct fwRule *rules = getRules();
    struct fwRule *rule = rules;
    while (rule) {
        struct fwRule *temp = rule;
        rule = rule->next;
        // delete the rule from the kernel
        if (temp->target == DNAT) {
            dnatRuleSet(_protocol, CMD_DEL,
                    temp->rmt_ip, temp->rmt_port,
                    temp->local_ip, temp->local_port,
                    temp->priv_ip, temp->priv_port);
        } else {
            snatRuleSet(_protocol, CMD_DEL,
                    temp->priv_ip, temp->priv_port,
                    temp->local_ip, temp->local_port,
                    temp->rmt_ip, temp->rmt_port);
        }
        free(temp);
    }

    /* flush the internal representation */
    flush_internal();

    /* Delete the log file */
    fd = open(_logFileName, O_RDWR | O_TRUNC | O_SYNC, S_IRWXU);
    if (fd < 0) { ABORT("failed to create a new log file"); }
    _logFD = fdopen(fd, "w+");
}

// Create NAT binding at "proxy" so that connections from "src" to "priv" may be
// allowed. If "src" is NULL, then NAT will be opened so that anybody in the public
// can connect to "priv" through the NAT binding at "proxy"
bool
NAT_Mnger::enable_inbound(
        struct sockaddr_in *src,
        struct sockaddr_in *proxy,
        struct sockaddr_in *priv)
{
    char str1[50], str2[50], buf[100];
    uint32_t priv_ip = priv->sin_addr.s_addr;
    uint16_t priv_port = priv->sin_port;
    uint32_t local_ip = proxy->sin_addr.s_addr;
    uint16_t local_port = proxy->sin_port;
    uint32_t rmt_ip;
    uint16_t rmt_port;
    if (src) {
        rmt_ip = src->sin_addr.s_addr;
        rmt_port = src->sin_port;
        sprintf(str1, "%s", _sin2str(src));
    } else {
        rmt_ip = htonl(INADDR_ANY);
        rmt_port = htons(0);
        sprintf(str1, "%s", _ipport2str(rmt_ip, rmt_port));
    }
    sprintf(str2, "%s", _sin2str(proxy));
    sprintf(buf, "DNAT ADD %s %s %s\n", str1, str2, _sin2str(priv));
	LOG(D_FULL, "\t%s", buf);

	// update internal data structure
    if (!addInternal(DNAT, priv_ip, priv_port, local_ip, local_port, rmt_ip, rmt_port)) {
        LOG(D_BASIC, "addInternal failed\n");
        return false;
    }
	LOG(D_FULL, "\t\tInternal: done\n");

	// write the rule to the log file: we write the forwarding rule to the
	// persist file before seting up actual forwarding rule, which will
	// write the forwarding rule to kernel. At some bad situation, we will have some
	// garbage rules remained in our condor persistent file, but this causes no harm
	// and the garbage will be cleaned at the next startup time
	fputs(buf, _logFD);
    fflush(_logFD);
	LOG(D_FULL, "\t\tLog: done\n");

	// now setup the port forwarding rule
	if (dnatRuleSet(_protocol, CMD_ADD, rmt_ip, rmt_port, local_ip, local_port, priv_ip, priv_port ) < 0) {
		LOG(D_BASIC, "NAT_Mnger::addRule - failed to add the rule to the kernel\n");
		ABORT("\t%s\n", strerror(errno));
	}
	LOG(D_FULL, "\t\tKernel: done\n");
	return true;
}

// Create NAT binding at "proxy" so that connections from "priv" to "dest" may be
// allowed. If "dest" is NULL, then NAT will be opened so that any connection from
// "priv" toward public network may be allowed through the NAT binding at "proxy"
bool
NAT_Mnger::enable_outbound(struct sockaddr_in *priv, struct sockaddr_in *proxy, struct sockaddr_in *dest)
{
    char str1[50], str2[50], buf[100];
    uint32_t priv_ip = priv->sin_addr.s_addr;
    uint16_t priv_port = priv->sin_port;
    sprintf(str1, "%s", _sin2str(priv));
    uint32_t local_ip = proxy->sin_addr.s_addr;
    uint16_t local_port = proxy->sin_port;
    sprintf(str2, "%s", _sin2str(proxy));
    uint32_t rmt_ip;
    uint16_t rmt_port;
    if (dest) {
        rmt_ip = dest->sin_addr.s_addr;
        rmt_port = dest->sin_port;
    } else {
        rmt_ip = htonl(INADDR_ANY);
        rmt_port = htons(0);
    }
    sprintf(buf, "SNAT ADD %s %s %s\n", str1, str2, _ipport2str(rmt_ip, rmt_port));
    LOG(D_FULL, "\t%s", buf);

    // update internal data structure
    if (!addInternal(SNAT, priv_ip, priv_port, local_ip, local_port, rmt_ip, rmt_port)) {
        LOG(D_BASIC, "addInternal failed\n");
        return false;
    }
    LOG(D_FULL, "\t\tInternal: done\n");

    // write the rule to the log file: we write the forwarding rule to the
    // persist file before seting up actual forwarding rule, which will
    // write the forwarding rule to kernel. At some bad situation, we will have some
    // garbage rules remained in our condor persistent file, but this causes no harm
    // and the garbage will be cleaned at the next startup time
    fputs(buf, _logFD);
    fflush(_logFD);
    LOG(D_FULL, "\t\tLog: done\n");

    // now setup the port forwarding rule
    if (snatRuleSet(_protocol, CMD_ADD, priv_ip, priv_port, local_ip, local_port, rmt_ip, rmt_port) < 0) {
        LOG(D_BASIC, "NAT_Mnger::addRule - failed to add the rule to the kernel\n");
        ABORT("\t%s\n", strerror(errno));
    }
    LOG(D_FULL, "\t\tKernel: done\n");
    return true;
}

// Delete NAT binding at "proxy" that was opened to make connections from "src" to
// "priv" possible. If "src" is NULL, then the NAT binding at "proxy" that enables
// all inbound connections to "priv" will be deleted
bool
NAT_Mnger::disable_inbound(
        struct sockaddr_in *src,
        struct sockaddr_in *proxy,
        struct sockaddr_in *priv)
{
    char str1[50], str2[50];
    uint32_t rmt_ip;
    uint16_t rmt_port;
    if (src) {
        rmt_ip = src->sin_addr.s_addr;
        rmt_port = src->sin_port;
        sprintf(str1, "%s", _sin2str(src));
    } else {
        rmt_ip = htonl(INADDR_ANY);
        rmt_port = htons(0);
        sprintf(str1, "%s", _ipport2str(rmt_ip, rmt_port));
    }
    uint32_t local_ip = proxy->sin_addr.s_addr;
	uint16_t local_port = proxy->sin_port;
    sprintf(str2, "%s", _sin2str(proxy));
	uint32_t priv_ip = priv->sin_addr.s_addr;
	uint16_t priv_port = priv->sin_port;
	LOG(D_FULL, "\tDNAT DEL %s %s %s\n", str1, str2, _sin2str(priv));

	// delete internal data structure first
    if (!deleteInternal(DNAT, priv_ip, priv_port, &local_ip, &local_port, rmt_ip, rmt_port)) {
		LOG(D_BASIC, "NAT_Mnger::deleteRule - deleteInternal failed\n");
        return false;
	}
    char buf[100];
    sprintf(str2, "%s", _ipport2str(local_ip, local_port));
    sprintf(buf, "DNAT DEL %s %s %s\n", str1, str2, _ipport2str(priv_ip, priv_port));
	LOG(D_FULL, "\tInternal done: %s", buf);

	// unset forwarding rule here
	if (dnatRuleSet(_protocol, CMD_DEL, rmt_ip, rmt_port, local_ip, local_port, priv_ip, priv_port ) < 0) {
		LOG(D_BASIC, "NAT_Mnger::deleteRule - failed to delete the rule to the kernel\n");
		ABORT("\t%s\n", strError());
	}
	LOG(D_FULL, "\tKernel: done\n");

    // update log
    fputs(buf, _logFD);
    fflush(_logFD);
    LOG(D_FULL, "\tLog: done\n");

    return true;
}

// Delete NAT binding at "proxy" that was opened to make connections from "priv" to
// "dest" possible. If "dest" is NULL, then the NAT binding at "proxy" that enables
// all outbound connections from "priv" will be deleted. If "proxy" is NULL, then the
// first rule with the matching (priv, dest)
bool
NAT_Mnger::disable_outbound(
        struct sockaddr_in *priv,
        struct sockaddr_in *proxy,
        struct sockaddr_in *dest)
{
    char str1[50];
    sprintf(str1, "%s", _sin2str(priv));
    LOG(D_FULL, "\tDNAT DEL %s\n", str1);

    uint32_t priv_ip = priv->sin_addr.s_addr;
    uint16_t priv_port = priv->sin_port;
    uint32_t local_ip;
    uint16_t local_port;
    if (proxy) {
        local_ip = proxy->sin_addr.s_addr;
        local_port = proxy->sin_port;
    } else {
        local_ip = htonl(INADDR_ANY);
        local_port = htons(0);
    }
    uint32_t rmt_ip;
    uint16_t rmt_port;
    if (dest) {
        rmt_ip = dest->sin_addr.s_addr;
        rmt_port = dest->sin_port;
    } else {
        rmt_ip = htonl(INADDR_ANY);
        rmt_port = htons(0);
    }

    // delete internal data structure first
    if (!deleteInternal(SNAT, priv_ip, priv_port, &local_ip, &local_port, rmt_ip, rmt_port)) {
        LOG(D_BASIC, "NAT_Mnger::deleteRule - deleteInternal failed\n");
        return false;
    }
    char str2[50];
    sprintf(str2, "%s", _ipport2str(local_ip, local_port));
    char buf[100];
    sprintf(buf, "SNAT DEL %s %s %s\n", str1, str2, _ipport2str(rmt_ip, rmt_port));
    LOG(D_FULL, "\tInternal done: %s", buf);

    // unset forwarding rule here
    if (snatRuleSet(_protocol, CMD_DEL, priv_ip, priv_port, local_ip, local_port, rmt_ip, rmt_port) < 0) {
        LOG(D_BASIC, "NAT_Mnger::deleteRule - failed to delete the rule to the kernel\n");
        ABORT("\t%s\n", strError());
    }
    LOG(D_FULL, "\tKernel: done\n");

    // update log
    fputs(buf, _logFD);
    fflush(_logFD);
    LOG(D_FULL, "\tLog: done\n");

    return true;
}

void
NAT_Mnger::compress_log()
{
    // close the current log file
    fclose(_logFD);

    // open a temporary log file
    char tempname[30];
    sprintf(tempname, ".%s.tmp", _logFileName);
	FILE *temp_log = fopen(tempname, "w+");
	if (!temp_log) {
		ABORT("NAT_Mnger::compress_log - faile to fopen(temp file)");
	}

    // dump the snapshot of internal representation to a temporary file
	struct fwRule *rules = getRules();
    struct fwRule *rule = rules;
    while (rule) {
        struct fwRule *temp = rule;
        rule = rule->next;
        char buf[100], str1[50], str2[50];
        if (temp->target == DNAT) {
            sprintf(str1, "%s", _ipport2str(temp->rmt_ip, temp->rmt_port));
            sprintf(str2, "%s", _ipport2str(temp->local_ip, temp->local_port));
            sprintf(buf, "DNAT ADD %s %s %s\n",
                    str1, str2, _ipport2str(temp->priv_ip, temp->priv_port));
        } else {
            sprintf(str1, "%s", _ipport2str(temp->priv_ip, temp->priv_port));
            sprintf(str2, "%s", _ipport2str(temp->local_ip, temp->local_port));
            sprintf(buf, "SNAT ADD %s %s %s\n",
                    str1, str2, _ipport2str(temp->rmt_ip, temp->rmt_port));
        }
        fputs(buf, temp_log); 
        free(temp);
    }
    fclose(temp_log);

    // replace the log file with the temporary log
    if (rename(tempname, _logFileName)) {
        ABORT("NAT_Mnger::compress_log - rename failed: %s", strerror(errno));
    }

    // open the log file
    _logFD = fopen(_logFileName, "a+");
    if (!_logFD) {
        ABORT("NAT_Mnger::compress_log - faile to fopen(compressed log file)");
    }
}

struct fwRule *
NAT_Mnger::getRules()
{
	struct fwRule *rules = NULL; 
	for (int i=0; i<HASH_SIZE; i++) {
	    struct fwRule *rptr = _rules[i];
		while (rptr) {
			struct fwRule * temp = (struct fwRule *)malloc(sizeof(struct fwRule));
            temp->target = rptr->target;
            temp->priv_ip = rptr->priv_ip;
            temp->priv_port = rptr->priv_port;
            temp->local_ip = rptr->local_ip;
            temp->local_port = rptr->local_port;
            temp->rmt_ip = rptr->rmt_ip;
            temp->rmt_port = rptr->rmt_port;
			rptr = rptr->next;
			temp->next = rules;
			rules = temp;
		}
	}
	return rules;
}

void
NAT_Mnger::flush_internal()
{
    for (int i=0; i<HASH_SIZE; i++) {
        struct fwRule *rptr = _rules[i];
        while (rptr) {
            struct fwRule *temp = rptr;
            rptr = rptr->next;
            free(temp);
        }
        _rules[i] = NULL;
    }
}
