#ifndef __NAT_MNGER
#define __NAT_MNGER

#include <stdio.h>
#include <netinet/in.h>

class NAT_Mnger {
	public:
		// @args: proto: TCP or UDP
        NAT_Mnger(char proto, char *logfile);

        // Create NAT binding at "proxy" so that connections from "src" to "priv" may be
        // allowed. If "src" is NULL, then NAT will be opened so that anybody in the public
        // can connect to "priv" through the NAT binding at "proxy"
        bool enable_inbound(
                struct sockaddr_in *src,
                struct sockaddr_in *proxy,
                struct sockaddr_in *priv);

        // Delete NAT binding at "proxy" that was opened to make connections from "src" to
        // "priv" possible. If "src" is NULL, then the NAT binding at "proxy" that enables
        // all inbound connections to "priv" will be deleted
        bool disable_inbound(
                struct sockaddr_in *src,
                struct sockaddr_in *proxy,
                struct sockaddr_in *priv);

        // Create NAT binding at "proxy" so that connections from "priv" to "dest" may be
        // allowed. If "dest" is NULL, then NAT will be opened so that any connection from
        // "priv" toward public network may be allowed through the NAT binding at "proxy"
        bool enable_outbound(
                struct sockaddr_in *priv,
                struct sockaddr_in *proxy,
                struct sockaddr_in *dest);

        // Delete NAT binding at "proxy" that was opened to make connections from "priv" to
        // "dest" possible. If "dest" is NULL, then the NAT binding at "proxy" that enables
        // all outbound connections from "priv" will be deleted. If "proxy" is NULL, then the
        // first rule with the matching (priv, dest)
        bool disable_outbound(
                struct sockaddr_in *priv,
                struct sockaddr_in *proxy,
                struct sockaddr_in *dest);

        // At the startup time, the log file is refered to delete rules that are left
        // due to the previous crash. To maintain the log file not too big, the log
        // file must be periodically (once a day at least) compressed. Actually no
        // compression occurrs here. This routine makes the log file compact by
        // deleting unnecessary entries in the file
        void compress_log();

        // Delete rules remained in the log file and start with a fresh log file.
        // This routine must be called at the startup time
        void cleanup();

    private:
        static const char TCP = 't';
        static const char UDP = 'u';
        static const char SNAT = 's';
        static const char DNAT = 'd';
        static const int HASH_SIZE = 173;

	protected:
		int _protocol;
		char _logFileName[30];
        FILE *_logFD;
        /* Local representation of address translation rules. Each rule is
         * represented by (src ip, src port, local ip, local port, dest ip, dest port),
         * where (local ip, local port) is a ip:port pair of this machine.
         */
		struct fwRule *_rules[HASH_SIZE];

		/* Method */
        // Add a rule to the internal data structure. A rule is internally represented by
        // (private address, local address, remote address), where pviate address is ip:port
        // of a socket behind the NAT box and in a private network, local address is ip:port
        // that a NAT binding will be created for the private address, and remote address is
        // ip:port of a socket outside the private network. Depending on whether the rule is
        // a DNAT or SNAT rule, each address means source or destination.
        // @args:
        //    - target: DNAT or SNAT
        //    - priv_ip, priv_port: private address
        //    - local_ip, local_port: local address
        //    - rmt_ip, rmt_port: remote address
        // @return: true, if success
        //          false if error
        // @error:
        //    - EINVAL: the rule with the same value already exists
		bool addInternal(char target,
                uint32_t priv_ip, uint16_t priv_port,
                uint32_t local_ip, uint16_t local_port,
                uint32_t rmt_ip, uint16_t rmt_port);
        // Delete a rule from the internal representation. See the comments on addInternal
        // for how rules are internally represented
        // @args
        //    - target: DNAT or SNAT
        //    - priv_ip, priv_port: private address
        //    - local_ip, local_port: local address. These are value-result argument. If
        //      local_ip:local_port is not 0.0.0.0:0, then the rule with the same (private,
        //      local, remote) will be deleted. Otherwise the rule with the same (private,
        //      remote) value will be deleted and the local address of the deleted rule will
        //      be returned via these args
        //    - rmt_ip, rmt_port: remote address
        // @return: true, if success
        //          false if error
        // @error:
        //    - EINVAL: the rule with the given value does not exist
        bool deleteInternal(char target,
                uint32_t priv_ip, uint16_t priv_port,
                uint32_t *local_ip, uint16_t *local_port,
                uint32_t rmt_ip, uint16_t rmt_port);
		// Get the list of rules which are in the local representation.
		// Note: The list of rules should be free by the caller.
		struct fwRule *getRules();
        // Delete all the rules from the local representation
        void flush_internal();
};

#endif
