#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include "log.h"
#include "internet.h"

char *
_prtfds(int maxfd, fd_set *fds)
{
    static char buf[50];
    int i, size;

    if (fds == NULL) {
        sprintf(buf, "NULL");
        return buf;
    }
    sprintf(buf, "<");
    for(i=0; i<maxfd; i++) {
        if (fds && FD_ISSET(i, fds)) {
            if ((size = strlen(buf)) > 40) {
                strcat(buf, "...>");
                return buf;
            }
            sprintf(&buf[strlen(buf)], "%d ", i);
        }
    }
    strcat(buf, ">");
    return buf;
}

unsigned short
_hash_sin(struct sockaddr_in *addr, unsigned short size)
{
    unsigned long index;
    index = (addr->sin_addr.s_addr + addr->sin_port) % size;
    return (unsigned short)index;
}

/* Convert a string of the form "xx.xx.xx.xx:port", "<xx.xx.xx.xx:pppp>",
   or "<hostname:port>" to a sockaddr_in */
int
_str2sin(const char *addr, struct sockaddr_in *sin)
{
    char*   addrCpy;
    char*   string;
    char*   colon = 0;
    struct  hostent *hostptr;

    if(!addr) {
        return -1;
    }
    addrCpy = strdup(addr);
    string = addrCpy;
    /* remove the leading '<'  and the trailing '>' */
    if (*string == '<') {
        string++;
        string[strlen(string)-1] = '\0';
    }

    /* Not string must be the format of either "hostname:pppp" or "dotted-ip:port" */
    if( !(colon = strchr(string, ':')) ) {
        // Not a valid sinful string, return failure
        free(addrCpy);
        ERROR("could not find ':' in the string passed");
        return -1;
    }
    *colon = '\0';
    if (!inet_aton(string, &(sin->sin_addr))) { // if invalid dotted-ip notation
        if ((hostptr=gethostbyname(string)) != NULL && hostptr->h_addrtype==AF_INET) {
            // the string is format of "<hostname:port>" */
            sin->sin_addr = *(struct in_addr *)(hostptr->h_addr_list[0]);
        } else {
            ERROR("invalid hostname passed");
            return -1;
        }
    }
    // process port part
    sin->sin_port = htons(atoi(++colon));
    sin->sin_family = AF_INET;
    free(addrCpy);
    return 0;
}

char *
_sin2str(const struct sockaddr_in *sin)
{
    static char buf[50];
    if (!sin) {
        buf[0] = '\0';
        return NULL;
    }
    sprintf(buf, "<%s:%d>", inet_ntoa(sin->sin_addr), ntohs(sin->sin_port));
    return buf;
}

char *
_fd2str(int sockd)
{
    struct sockaddr_in  addr;
    int         addr_len;
    static char *mynull = "\0";

    addr_len = sizeof(addr);

    if (getsockname(sockd, (struct sockaddr *)&addr, &addr_len) < 0) 
        return mynull;

    return (_sin2str(&addr));
}

char *
_sin2host(const struct sockaddr_in *from, char ***aliases)
{
    struct hostent  *hp;
#ifndef WIN32
    struct hostent  *gethostbyaddr();
#endif

    if(!from) {
        // make certain from is not NULL before derefencing it
        return NULL;
    }

    if( (hp=gethostbyaddr((char *)&from->sin_addr,
                sizeof(struct in_addr), AF_INET)) == NULL ) {
        // could not find a name for this address
        return NULL;
    } else {
        // CAREFULL: we are returning a staic buffer from gethostbyaddr.
        // The caller had better use the result immediately or copy it.
        // Also note this is not thread safe.  (as are lots of things in internet.c).
        if(aliases) {
            *aliases = hp->h_aliases;
        }
        return hp->h_name;
    }
}

void
_display_from(struct sockaddr_in  *from)
{
    struct hostent  *hp;
#ifndef WIN32
    struct hostent  *gethostbyaddr();
#endif

    if( !from ) {
        ERROR("from NULL source" );
        return;
    }

    if( (hp=gethostbyaddr((char *)&from->sin_addr,
                sizeof(struct in_addr), AF_INET)) == NULL ) {
        LOG(D_BASIC, "from (%s), port %d\n",
                inet_ntoa(from->sin_addr), ntohs(from->sin_port));
    } else {
        LOG(D_BASIC, "from %s, port %d\n", hp->h_name, ntohs(from->sin_port));
    }
}

char *
_subnet(char* host)
{
    char            subnetname[50];
    char            *subnet_ptr;
    char            *host_addr_string;
    int             subnet_length;
    struct          in_addr in;
    uint32_t        ipaddr;

    if (_myIP(&ipaddr) < 0) {
        return strdup("");
    }

    memcpy((char *) &in,(char *)&ipaddr, sizeof(uint32_t));
    host_addr_string = inet_ntoa( in );
    if( host_addr_string ) {
        subnet_ptr = (char *) strrchr(host_addr_string, '.');
        if(subnet_ptr == NULL) {
            return strdup("");
        }
        subnet_length = subnet_ptr - host_addr_string;
        strncpy(subnetname, host_addr_string, subnet_length);
        subnetname[subnet_length] = '\0';
        return (strdup(subnetname));
    }
    return strdup("");
}

int
_same_host(const char *h1, const char *h2)
{
    struct hostent *he1, *he2;
    char cn1[50];

    if (strcmp(h1, h2) == 0) {
        return 1;
    }
    
    if ((he1 = gethostbyname(h1)) == NULL) {
        return -1;
    }

    // stash h_name before our next call to gethostbyname
    strncpy(cn1, he1->h_name, 50);

    if ((he2 = gethostbyname(h2)) == NULL) {
        return -1;
    }

    return (strcmp(cn1, he2->h_name) == 0);
}

/*
  Return TRUE if the given domain contains the given hostname, FALSE
  if not.  Origionally taken from condor_starter.V5/starter_common.C. 
  Moved here on 1/18/02 by Derek Wright.
*/
int
_in_domain(const char *host, const char *domain)
{
    const char  *ptr;

    for( ptr=host; *ptr; ptr++ ) {
        if( strcmp(ptr,domain) == 0 ) {
            return 1;
        }
    }
    return 0;
}

/*
  is_ipaddr() returns TRUE if buf is an ascii IP address (like
  "144.11.11.11") and false if not (like "cs.wisc.edu").  Allow
  wildcard "*".  If we return TRUE, and we were passed in a non-NULL 
  sin_addr, it's filled in with the integer version of the ip address. 
*/
int
_valid_ip(const char *inbuf, struct in_addr *sin_addr)
{
    int len;
    char buf[17];
    int part = 0;
    int i,j,x;
    char save_char;
    unsigned char *cur_byte = NULL;
    if( sin_addr ) {
        cur_byte = (unsigned char *) sin_addr;
    }

    len = strlen(inbuf);
    if ( len < 3 || len > 16 ) 
        return 0;   // shortest possible IP addr is "1.*" - 3 chars
    
    // copy to our local buf
    strncpy( buf, inbuf, 17 );

    // on IP addresses, wildcards only permitted at the end, 
    // i.e. 144.92.* , _not_ *.92.11
    if ( buf[0] == '*' ) 
        return 0;

    // strip off any trailing wild card or '.'
    if ( buf[len-1] == '*' || buf[len-1] == '.' ) {
        if ( buf[len-2] == '.' )
            buf[len-2] = '\0';
        else
            buf[len-1] = '\0';
    }

    // Make certain we have a valid IP address, and count the parts,
    // and fill in sin_addr
    i = 0;
    for(;;) {
        
        j = i;
        while (buf[i] >= '0' && buf[i] <= '9') i++;
        // make certain a number was here
        if ( i == j )
            return 0;   
        // now that we know there was a number, check it is between 0 & 255
        save_char = buf[i];
        buf[i] = '\0';
        x = atoi( &(buf[j]) );
        if( x < 0 || x > 255 ) {
            return 0;
        }
        if( cur_byte ) {
            *cur_byte = x;  /* save into sin_addr */
            cur_byte++;
        }
        buf[i] = save_char;

        part++;
        
        if ( buf[i] == '\0' ) 
            break;
        
        if ( buf[i] == '.' )
            i++;
        else
            return 0;

        if ( part >= 4 )
            return 0;
    }
    
    if( cur_byte ) {
        for (i=0; i < 4 - part; i++) {
            *cur_byte = (unsigned char) 255;
            cur_byte++;
        }
    }
    return 1;
}

int
_valid_sin(const char *sinful)
{
    char* tmp;
    if( !sinful ) return 0;
    if( !(sinful[0] == '<') ) return 0;
    if( !(tmp = strchr(sinful, ':')) ) return 0;
    if( !(tmp = strrchr(sinful, '>')) ) return 0;
    return 1;
}

int
_str2port(const char* addr)
{
    char *sinful, *tmp;
    int port = 0;

    if( ! (addr && _valid_sin(addr)) ) {
        return -1;
    }

    sinful = strdup( addr );
    if( (tmp = strrchr(sinful, '>')) ) {
        *tmp = '\0';
        tmp = strchr( sinful, ':' );
        if( tmp && tmp[1] ) {
            port = atoi( &tmp[1] );
        } 
    }
    free( sinful );
    return port;
}

uint32_t
_str2ip(const char* addr)
{
    char *sinful, *tmp;
    unsigned int ip = 0;
    struct in_addr sin_addr;

    if(!(addr && _valid_sin(addr))) {
        return -1;
    }

    sinful = strdup(addr);
    if((tmp = strchr(sinful, ':'))) {
        *tmp = '\0';
        if( _valid_ip(&sinful[1], &sin_addr) ) {
            ip = sin_addr.s_addr;
        }
    } else {
        ERROR("_valid_ip(\"%s\") is true, but can't find ':'");
    }
    free(sinful);
    return ip;
}

char*
_str2ipstr(const char* addr) 
{
    char *tmp;
    static char result[50];
    char sinful[50];

    if( ! (addr && _valid_sin(addr)) ) {
        return NULL;
    }

    strncpy( sinful, addr, 50 );
    tmp = strchr( sinful, ':' );
    if( tmp ) {
        *tmp = '\0';
    } else {
        return NULL;
    }
    if( _valid_ip(&sinful[1], NULL) ) {
        strncpy( result, &sinful[1], 50 );
        return result;
    }
    return NULL;
}

char*
_str2host(const char* addr) 
{
    char *tmp;
    static char result[50];
    struct sockaddr_in sin;

    if( ! (addr && _valid_sin(addr)) ) {
        return NULL;
    }

    _str2sin( addr, &sin );
    if( (tmp = _sin2host(&sin, NULL)) ) {
        strncpy( result, tmp, 50 );
    } else {
        return NULL;
    }
    return result;
}

char *
_ip2host(unsigned int ip)
{
    struct hostent *ent;
    ent = gethostbyaddr((char *) &ip, sizeof(int), AF_INET);
    if (!ent) {
        return NULL;
    }
    return ent->h_name;
}

/* Check if the ip is in private ip address space */
/* ip: in host byte order */
int
_priv_net(uint32_t ip)
{
    return ((ip & 0xFF000000) == 0x0A000000 ||      // 10/8
            (ip & 0xFFF00000) == 0xAC100000 ||      // 172.16/12
            (ip & 0xFFFF0000) == 0xC0A80000);       // 192.168/16
}

/* Check if two ip addresses, given in network byte, are in the same network */
int
_same_net(uint32_t ipA, uint32_t ipB)
{
    unsigned char *byteA, *fA, *byteB;
    int i, index;

    fA = byteA = (char *)&ipA;
    byteB = (char *)&ipB;

    if (*fA < 128) { // A class
        index = 1;
    } else if(*fA < 192) { // B class
        index = 2;
    } else {    // C class
        index = 3;
    }

    for (i = 0; i < index; i++) {
        if (*byteA != *byteB) {
            return 0;
        }
        byteA++;
        byteB++;
    }

    return 1;
}

// ip: network-byte order
// port: network-byte order
char *
_ipport2str(const unsigned int ip, const unsigned short port)
{
    int             i;
    static  char    buf[24];
    char    tmp_buf[10];
    char    *cur_byte;
    unsigned char   this_byte;

    buf[0] = '<';
    buf[1] = '\0';
    cur_byte = (char *) &ip;
    for (i = 0; i < sizeof(ip); i++) {
        this_byte = (unsigned char) *cur_byte;
        sprintf(tmp_buf, "%u.", this_byte);
        cur_byte++;
        strcat(buf, tmp_buf);
    }
    buf[strlen(buf) - 1] = ':';
    sprintf(tmp_buf, "%d>", ntohs(port));
    strcat(buf, tmp_buf);
    return buf;
}

// ip: in network order
char *
_ip2str(const uint32_t ip)
{
    int             i;
    static  char    buf[24];
    char    *cur_byte;
    unsigned char   this_byte;

    buf[0] = '<';
    buf[1] = '\0';
    cur_byte = (char *) &ip;
    for (i = 0; i < sizeof(ip); i++) {
        this_byte = (unsigned char) *cur_byte;
        sprintf(&buf[strlen(buf)], "%u.", this_byte);
        cur_byte++;
    }
    sprintf(&buf[strlen(buf)-1], ">");
    return buf;
}

/* Get the ip address of this host. If the user set the environment variable MY_IP,
 * then use that ip address. If not, this routine tries to find a public ip address
 * that this machine has. If it can't find any public ip address, then it returns a
 * private ip address found */
// @args:   ipaddr - ip address found will be returned via this arg
//                   in network byte order
// @return: -1, failed
//          0, succeed
int
_myIP(uint32_t *ipaddr)
{
    int sockfd, len, lastlen;
    char *ptr, *buf, lastname[IFNAMSIZ], *cptr;
    struct ifconf ifc;
    struct ifreq *ifr, ifrcopy;
    struct sockaddr_in *sinptr;
    char *env = NULL;
    struct in_addr inaddr;

    // Search environment variable
    env = getenv("MY_IP");
    if (env) {
        if (!inet_aton(env, &inaddr)) { // if invalid dotted-ip notation
            ERROR("MY_IP is not valid ip str");
        }
        *ipaddr = inaddr.s_addr;
        return 0;
    }

    // Create a socket to do ioctl
    sockfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sockfd < 0) {
        ERROR("_myIP: socket creation");
        return -1;
    }

    // Get the list of interfaces in the system
    lastlen = 0;
    len = 100 * sizeof(struct ifreq); // initial buffer size guess
    while(1) {
        if ((buf = (char *)malloc(len)) == NULL) {
            ERROR("_myIP: malloc failed");
            close(sockfd);
            return -1;
        }
        ifc.ifc_len = len;
        ifc.ifc_buf = buf;
        // The problem of ioctl is that some implementations of OS does not
        // return an error when buf is not large enough to hold the result.
        // Instead the result is truncated and success is returned. Hence
        // we try multiple ioctl with increasing buf size, until we get the
        // same length returned from two consecutive ioctl calls
        if (ioctl(sockfd, SIOCGIFCONF, &ifc) < 0) {
            if (errno != EINVAL || lastlen != 0) {
                ERROR("_myIP: ioctl failed");
            }
        } else {
            if (ifc.ifc_len == lastlen) { // success
                break;
            }
            lastlen = ifc.ifc_len;
        }
        len += 10 * sizeof(struct ifreq);
        free(buf);
    }

    // Skip "lo" and its alias interfaces. Also skip interfaces not up.
    // Keep getting attributes of interfaces until an interface with
    // a public ip address is found.
    lastname[0] = 0;
    for(ptr = buf; ptr < buf + ifc.ifc_len; ) {
        ifr = (struct ifreq *) ptr;
        // Figure out the size of ifreq structure
        // Difficulty is that some systems provide length field but others not
#ifdef HAVE_SOCKADDR_SA_LEN
        len = max(sizeof(struct sockaddr), ifr->ifr_addr.sa_len);
#else
        switch(ifr->ifr_addr.sa_family) {
#ifdef IPV6
            case AF_INET6:
                len = sizeof(struct sockaddr_in6);
                break;
#endif
            case AF_INET:
            default:
                len = sizeof(struct sockaddr_in);
                break;
        }
#endif
        // Move ptr to the next interface
        ptr += sizeof(ifr->ifr_name) + len;

        // For now, ignore non ipv4 addresses
        if (ifr->ifr_addr.sa_family != AF_INET) {
            continue;
        }

        // In solaris, alias ip has interface name associated with it
        // in the form of "primary_interface_name:number", while in BSD
        // alias ip's share the same interface name. Hence we must check
        // only the primary interface name part
        if ((cptr = strchr(ifr->ifr_name, ':')) != NULL) {
            *cptr = 0;
        }
        // Check if the interface is loopback
        if (!strcmp(ifr->ifr_name, "lo")) {
            continue;
        }
        // Check if the interface is the same as the previous one
        if (!strncmp(ifr->ifr_name, lastname, IFNAMSIZ)) {
            continue;
        }
        memcpy(lastname, ifr->ifr_name, IFNAMSIZ);

        // Now, get the attributes of the current interface
        ifrcopy = *ifr;
        if (ioctl(sockfd, SIOCGIFFLAGS, &ifrcopy) < 0) {
            ERROR("_myIP: ioctl failed");
        }

        // If not up, ignore
        if ( !(ifrcopy.ifr_flags & IFF_UP) ) {
            continue;
        }

        // Check if ip address is in private range
        sinptr = (struct sockaddr_in *) &ifr->ifr_addr;
        *ipaddr = sinptr->sin_addr.s_addr;
        if (_priv_net(ntohl(*ipaddr))) {
            continue;
        }
        close(sockfd);
        return 0;
    }
    close(sockfd);
    return 0;
}
