#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "log.h"
#include "nb_wbuf.h"

NB_wbuf::NB_wbuf(int sock)
{
    fd = sock;
    bytes = size = 0;
}

bool
NB_wbuf::empty()
{
    return (size == 0);
}

bool
NB_wbuf::flush()
{
    int sent;

    sent = ::send(fd, &data[bytes], size - bytes, 0);
    if (sent < 0) { // error
        return false;
    } else if (sent < size - bytes) { // not totally flushed
        bytes += sent;
        errno = EAGAIN;
        LOG(D_FULL, "sent: %d, bytes: %d, size: %d\n", sent, bytes, size);
        return false;
    } else { // flushed
        bytes = size = 0;
        LOG(D_FULL, "sent: %d, bytes: %d, size: %d\n", sent, bytes, size);
        return true;
    }
}

bool
NB_wbuf::send(char *record, int len)
{
    // check if the record is too big
    if (len > MAX) {
        LOG(D_BASIC, "NB_wbuf::send - too big a record\n");
        errno = EINVAL;
        return false;
    }

    // try to send the record and copy whatever remained into this buffer
    int sent = ::send(fd, record, len, 0);
    if (sent < 0) {
        if (errno == EAGAIN) {
            size = len;
            bytes = 0;
            memcpy(data, record, size);
            LOG(D_FULL, "sent: 0, buffered: %d\n", size);
        }
        return false;
    } else if (sent < len) {
        size = len - sent;
        bytes = 0;
        memcpy(data, &record[sent], size);
        errno = EAGAIN;
        LOG(D_FULL, "sent: %d, buffered: %d\n", sent, size);
        return false;
    } else {
        LOG(D_FULL, "sent: %d, buffered: 0\n", sent);
        return true;
    }
}

void
NB_wbuf::tostr(char *str)
{
    sprintf(str, "[fd: %d, size:%d, bytes:%d]", fd, size, bytes); 
}

int
NB_wbuf::stripe(char *buf)
{
    // the length of the serialization excluding this field(2 bytes)
    // fd: network byte order(4 bytes)
    // size of the msg buffered in 'data'(2 bytes)
    // unsigned char data[bytes ~ size-1](size bytes)

    char str[50];

    int index = 2;
    uint32_t nlong;
    uint16_t nshort;

    // fd: network byte order(4 bytes)
    nlong = htonl(fd);
    memcpy(&buf[index], &nlong, 4);
    sprintf(str, "[%d]fd=%d", index, fd);
    index += 4;
    // size of the msg buffered in 'data'(2 bytes)
    unsigned short datalen = size - bytes;
    nshort = htons(datalen);
    memcpy(&buf[index], &nshort, 2);
    sprintf(&str[strlen(str)], "[%d]data.len=%d", index, datalen);
    index += 2;
    // unsigned char data[bytes ~ size-1](size bytes)
    if (datalen > 0) {
        memcpy(&buf[index], &data[bytes], datalen);
        sprintf(&str[strlen(str)], "[%d]data", index);
        index += datalen;
    }

    // the length of the serialization excluding this field(2 bytes)
    nshort = htons(index-2);
    memcpy(buf, &nshort, 2);

    LOG(D_FULL, "Serialization of NB_wbuf: [0]len=%d%s\n", index-2, str);

    return index;
}

int
NB_wbuf::unstripe(char *buf)
{
    // the length of the serialization excluding this field(2 bytes)
    // fd: network byte order(4 bytes)
    // size of the msg buffered in 'data'(2 bytes)
    // unsigned char data[bytes ~ size-1](size bytes)

    char str[50];

    int index = 0;
    uint32_t nlong;
    uint16_t nshort;

    // the length of the serialization excluding this field(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short length = ntohs(nshort);
    sprintf(str, "[%d]len=%d", index, length);
    index += 2;
    // fd: network byte order(4 bytes)
    memcpy(&nlong, &buf[index], 4);
    fd = ntohl(nlong);
    sprintf(&str[strlen(str)], "[%d]fd=%d", index, fd);
    index += 4;
    // size of the msg buffered in 'data'(2 bytes)
    memcpy(&nshort, &buf[index], 2);
    unsigned short datalen = ntohs(nshort);
    sprintf(&str[strlen(str)], "[%d]data.len=%d", index, datalen);
    index += 2;
    // unsigned char data[bytes ~ size-1](size bytes)
    if (datalen > 0) {
        memcpy(&data, &buf[index], datalen);
        size = datalen;
        sprintf(&str[strlen(str)], "[%d]data", index);
        index += datalen;
    }
    
    LOG(D_FULL, "Unserialization of NB_wbuf: %s\n", str);

    if (length != index-2) {
        ABORT("length field does not match with the actual size of serialization");
    }

    return index;
}
