#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>

#include "log.h"
#include "internet.h"
#include "sock.h"

int
_make_nonblock(int fd)
{
    int val;
    if ((val = fcntl(fd, F_GETFL, 0)) < 0) {
        LOG(D_BASIC, "_make_nonblock: fcntl failed - %s\n", strerror(errno));
        return -1;
    }
    if (fcntl(fd, F_SETFL, val | O_NONBLOCK) < 0) {
        LOG(D_BASIC, "_make_nonblock: fcntl failed - %s\n", strerror(errno));
        return -1;
    }
    return 0;
}

int
_tcp_nonblock_socket()
{
    int sock;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
        LOG(D_BASIC, "_tcp_nonblock_socket: socket creation failed - %s", strerror(errno));
        return -1;
    }
    if (_make_nonblock(sock) < 0) {
        return -1;
    }
    return sock;
}

int
_udp_nonblock_socket()
{
    int sock;

    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        LOG(D_BASIC, "_udp_nonblock_socket: socket creation failed - %s", strerror(errno));
        return -1;
    }
    if (_make_nonblock(sock) < 0) {
        return -1;
    }
    return sock;
}

static int
get_port_range(unsigned short *low_port, unsigned short *high_port)
{
    char *low = NULL, *high = NULL;

    if ( (low = getenv("LOWPORT")) == NULL ) {
        return 0;
    }
    if ( (high = getenv("HIGHPORT")) == NULL ) {
        free(low);
        LOG(D_BASIC, "LOWPORT is defined but HIGHPORT undefined!\n");
        return 0;
    }

    *low_port = atoi(low);
    *high_port = atoi(high);

    if(*low_port < 1024 || *high_port < 1024 || *low_port > *high_port) {
        LOG(D_BASIC,
                "get_port_range - invalid LOWPORT(%d) and/or HIGHPORT(%d)\n",
                *low_port, *high_port);
        free(low);
        free(high);
        return 0;
    }

    free(low);
    free(high);
    return 1;
}

// Bind the socket to a port within [low_port, high_port]
// @args:
//  - my_addr: sin_port should be 0
static int
bindWithin(int fd, struct sockaddr_in *my_addr,
        unsigned short low_port, unsigned short high_port)
{
    unsigned short start_trial, this_trial;
    int range;
    struct timeval curTime;

    // Use hash function with pid to get the starting point
    (void) gettimeofday(&curTime, NULL);

    // int pid = (int) getpid();
    range = high_port - low_port + 1;
    // this line must be changed to use the hash function of condor
    start_trial = low_port + (curTime.tv_usec * 73/*some prime number*/ % range);

    this_trial = start_trial;
    do {
        my_addr->sin_port = htons((u_short)this_trial++);

        if (bind(fd, (struct sockaddr *)my_addr, sizeof(struct sockaddr_in)) == 0) {
            LOG(D_FULL, "bindWithin - bound to %d...\n", this_trial-1);
            return 0;
        } else {
            LOG(D_FULL, "bindWithin - failed to bind: %s\n", strerror(errno));
        }
        if ( this_trial > high_port )
            this_trial = low_port;
    } while(this_trial != start_trial);

    LOG(D_BASIC, "bindWithin - failed to bind any port within (%d ~ %d)\n",
            low_port, high_port);
    errno = EADDRINUSE;
    return -1;
}

int
_local_bind(int fd, struct sockaddr_in *my_addr, socklen_t addrlen)
{
    struct sockaddr_in sinaddr;

    memset(&sinaddr, 0, sizeof(sinaddr));
    sinaddr.sin_family = AF_INET;
    if (my_addr->sin_addr.s_addr == htonl(INADDR_ANY)) {
        if (_myIP(&(sinaddr.sin_addr.s_addr)) < 0) {
            LOG(D_BASIC, "can't do _myIP\n");
            return -1;
        }
    } else {
        sinaddr.sin_addr.s_addr = my_addr->sin_addr.s_addr;
    }
    sinaddr.sin_port = my_addr->sin_port;

    unsigned short lowPort, highPort;
    if ( get_port_range(&lowPort, &highPort) ) {
        unsigned short port = ntohs(sinaddr.sin_port);
        if (port != 0) {
            if (port < lowPort || port > highPort) {
                LOG(D_BASIC, "_local_bind: requested port is out of range\n");
                errno = EINVAL;
                return -1;
            }
            return bind(fd, (struct sockaddr *)&sinaddr, addrlen);
        }
        return bindWithin(fd, &sinaddr, lowPort, highPort);
    } else {
        return bind(fd, (struct sockaddr *)&sinaddr, addrlen);
    }
}

// bind the socket pointed by fd to ipstr:port
// @args:
//   - ipstr: dotted notation of ip address. If NULL is passed, the socket will be
//            bound to INADDR_ANY
int
_inet_bind(int fd, char *ipstr, uint16_t port)
{
    struct sockaddr_in sin;

    sin.sin_family = AF_INET;
    if (ipstr) {
        if (inet_aton(ipstr, &(sin.sin_addr)) < 0) {
            return -1;
        }
        /* solaris
        sin.sin_addr.s_addr = inet_addr(ipstr);
        // solaris */
    } else {
        sin.sin_addr.s_addr = INADDR_ANY;
    }
    sin.sin_port = port;

    return _local_bind(fd, &sin, sizeof(struct sockaddr_in));
}

int
_tcp_connect(int fd, uint32_t ip, uint16_t port)
{
    struct sockaddr_in sin;

    bzero(&sin, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = ip;
    sin.sin_port = port;
    return connect(fd, (const struct sockaddr *)&sin, sizeof(sin));
}

int
_tcp_send(int fd, char *buf, size_t size)
{
    int sent, total = 0;
    while(total < size) {
        sent = send(fd, &buf[total], size-total, 0);
        if (sent < 0) {
            if (errno != EAGAIN) { return 0; }
        }
        total += sent;
    }
    return 1;
}

/*
int
_inet_connect(int fd, uint32_t ip, uint16_t port)
{
    struct sockaddr_in sin;
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = ip;
    sin.sin_port = port;
    return connect(fd, (struct sockaddr *)&sin, sizeof(sin));
}
*/

/*
int
tcp_recv(int fd, char *buf, size_t size)
{
    int recvd, total = 0;
    while(total < size) {
        recvd = recv(fd, &buf[total], size-total, 0);
        if (recvd < 0) {
            if (errno != EAGAIN) { return 0; }
        }
        total += recvd;
    }
    return 1;
}
*/
